//************************************************************************
// Definitions of methods for all classes declared in arrange.h.
//------------------------------------------------------------------------
// Copyright (C) 1997 by James MacKay
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// The author can be contacted via e-mail at: jmackay@sentex.net
//
//************************************************************************

#include "arrange.h"

#ifndef ENABLE_INLINES
   #pragma hdrstop         // prevent attempts to precompile .inl file
   #include "arrange.inl"
#endif

//************************************************************************
// Utility functions.
//************************************************************************

template<class T>
boolean elementsNonNull(T **array, ArrangeInt size);

//------------------------------------------------------------------------
// Returns true iff the first size elements of array and array itself are 
// all non-NULL.
//------------------------------------------------------------------------
template<class T>
boolean elementsNonNull(T **array, ArrangeInt size)
   {
   if (array == 0)
      return false;
   for (ArrangeInt i = 0; i < size; i++)
      if (array[i] == 0)
         return false;
   return true;
   }

//************************************************************************
// Class Arrange method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Constructs an Arrange filter from an argument list.
//------------------------------------------------------------------------
Arrange::Arrange(int argc, char **argv): 
   SingleFilter(), numPairs(0), outputBuffer(uninit)
   {
   // construct an argument string from the argument list
   StringBuffer arguments;
   for (int i = 0; i < argc; i++)
      arguments.append(argv[i]).append(' ');
   
   String errorMessage;
   if (buildPairs(arguments.toString(), &errorMessage) == false)
      throw InvalidFilterArgumentException(errorMessage);

   ensure(numPairs >= 0);
   ensure(elementsNonNull(extractor, numPairs));
   ensure(elementsNonNull(overlayer, numPairs));
   }

//------------------------------------------------------------------------
// Constructs an Arrange filter from an argument string.
//------------------------------------------------------------------------
Arrange::Arrange(const String &arguments):
   SingleFilter(), numPairs(0), outputBuffer(uninit)
   {
   String errorMessage;
   if (buildPairs(arguments, &errorMessage) == false)
      throw InvalidFilterArgumentException(errorMessage);

   ensure(numPairs >= 0);
   ensure(elementsNonNull(extractor, numPairs));
   ensure(elementsNonNull(overlayer, numPairs));
   }
 
//------------------------------------------------------------------------
// Copy constructor.
//------------------------------------------------------------------------
Arrange::Arrange(const Arrange &f): 
   SingleFilter(f), outputBuffer(f.outputBuffer)
   {
   require(f.numPairs >= 0);
   require(elementsNonNull(f.extractor, f.numPairs));
   require(elementsNonNull(f.overlayer, f.numPairs));

   numPairs = 0;
   allocatePairArrays(f.numPairs);
   for (ArrangeInt i = 0; i < f.numPairs; i++)
      {
      extractor[i] = f.extractor[i]->getCopy();
      overlayer[i] = f.overlayer[i]->getCopy();
      }

   ensure(numPairs >= 0);
   ensure(elementsNonNull(extractor, numPairs));
   ensure(elementsNonNull(overlayer, numPairs));
   }  // Arrange::Arrange(const Arrange &)

//------------------------------------------------------------------------
// Arranges the contents of line and outputs the result.
//------------------------------------------------------------------------
void Arrange::processLine(const String &line)
   {
   String newLine = line.stripTrailing(EOL);
   check(line.length() - newLine.length() <= 1);

   // build the rearranged line in outputBuffer:
   String extracted;
   StringInt nextPos = 1;
   outputBuffer.clear();
   for (ArrangeInt i = 0; i < numPairs; i++)
      {
      extracted = extractor[i]->extract(newLine);
      nextPos   = overlayer[i]->overlay(outputBuffer, extracted, nextPos);
      }

   // the output line will end in an EOL iff the input line did:
   if (line.length() > newLine.length())
      outputBuffer.append(EOL);
   outputLine(outputBuffer);
   }

//------------------------------------------------------------------------
// Constructs extractor-overlayer pairs based on the argument string args.
// Returns true if successful; otherwise errMsg is set to a message
// describing the error and false is returned.
//------------------------------------------------------------------------
boolean Arrange::buildPairs(const String &args, String *errMsg)
   {
   PairFactory fact;

   allocatePairArrays(fact.countPairs(args));

   boolean noErrors = true;
   String consumableArgs = args;
   for (ArrangeInt i = 0; noErrors && i < numPairs; i++)
      {
      extractor[i] = fact.createExtractor(&consumableArgs, errMsg);
      if (extractor[i] == 0)
         noErrors = false;
      else
         {
         overlayer[i] = fact.createOverlayer(&consumableArgs, errMsg);
         if (overlayer[i] == 0)
            noErrors = false;
         }
      }  // for

   if (noErrors == false)
      deallocatePairArrays(numPairs);

   ensure(numPairs >= 0);
   ensure(noErrors == false || elementsNonNull(extractor, numPairs));
   ensure(noErrors == false || elementsNonNull(overlayer, numPairs));
   return noErrors;

   }  // Arrange::buildPairs(const String &, String *)

//------------------------------------------------------------------------
// Allocate the extractor and overlayer arrays to hold pairCount elements
// each, initializes each element to 0, and sets numPairs to pairCount.
//------------------------------------------------------------------------
void Arrange::allocatePairArrays(ArrangeInt pairCount)
   {
   require(pairCount >= 0);
 
   extractor = new Extractor *[pairCount];
   overlayer = new Overlayer *[pairCount];
   numPairs = pairCount;

   for (ArrangeInt i = 0; i < numPairs; i++)
      {
      extractor[i] = 0;
      overlayer[i] = 0;
      }

   ensure(numPairs >= 0);
   ensure(extractor != 0);
   ensure(overlayer != 0);
   }  // Arrange::allocatePairArrays(ArrangeInt)

//------------------------------------------------------------------------
// Delete the pairCount elements of each of the extractor and overlayer
// arrays, then deletes the arrays themselves and sets numPairs to zero.
//------------------------------------------------------------------------
void Arrange::deallocatePairArrays(ArrangeInt pairCount)
   {
   for (ArrangeInt i = 0; i < pairCount; i++)
      {
      delete extractor[i];
      delete overlayer[i];
      }
   delete[] extractor;
   delete[] overlayer;
   extractor = 0;
   overlayer = 0;
   numPairs = 0;
   }

