//************************************************************************
// Declarations of all Extractor classes, and definitions of all of their
// methods.
//------------------------------------------------------------------------
// Copyright (C) 1997 by James MacKay
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// The author can be contacted via e-mail at: jmackay@sentex.net
//
//************************************************************************
#ifndef EXTRACT_H_
#define EXTRACT_H_

#include "jstring.h"
#include "release.h"

#define DEFINE_EXTRACTOR_COPIER(extractorClassName)      \
   Extractor *getCopy() const                            \
      {                                                  \
      return new extractorClassName(*this);              \
      }

//************************************************************************
// Class Extractor - abstract base class for all extractors.
//************************************************************************
class Extractor
   {
   public:
      virtual String extract(const String &str) = 0;
      virtual Extractor *getCopy() const = 0;
      virtual ~Extractor()
         {
         // empty
         }
   private:
      Extractor &operator =(const Extractor &);
   }; // class Extractor

//************************************************************************
// Class SingleItemExtractor - parent class for all extractors that 
//                             extract a single item from a String.
//************************************************************************
class SingleItemExtractor: public Extractor
   {
   public:
      SingleItemExtractor(StringInt pos): Extractor()
         {
         require(pos > 0);
         index = pos - 1;
         }
      SingleItemExtractor(const SingleItemExtractor &e): Extractor(e)
         {
         index = e.index;
         }
   protected:
      StringInt index;
   private:
      SingleItemExtractor &operator =(const SingleItemExtractor &);
   }; // class SingleItemExtractor

//************************************************************************
// Class ToEndExtractor - parent class for all extractors that extract to 
//                        the end of a String.
//************************************************************************
class ToEndExtractor: public Extractor
   {
   public:
      ToEndExtractor(StringInt fromPos): Extractor()
         {
         require(fromPos > 0);
         index = fromPos - 1;
         }
      ToEndExtractor(const ToEndExtractor &e): Extractor(e)
         {
         index = e.index;
         }
   protected:
      StringInt index;
   private:
      ToEndExtractor &operator =(const ToEndExtractor &);
   }; // class ToEndExtractor

//************************************************************************
// Class RangeExtractor - parent class for all extractors that extract a 
//                        substring of a String.
//************************************************************************
class RangeExtractor: public Extractor
   {
   public:
      RangeExtractor(StringInt fromPos, StringInt forLength): Extractor()
         {
         require(fromPos > 0);
         require(forLength > 0);

         index  = fromPos - 1;
         length = forLength;
         }
      RangeExtractor(const RangeExtractor &e): Extractor(e)
         {
         index  = e.index;
         length = e.length;
         }
   protected:
      StringInt index;
      StringInt length;
   private:
      RangeExtractor &operator =(const RangeExtractor &);
   }; // class RangeExtractor

//************************************************************************
// Class CharExtractor - extracts the pos`th character from a String.
//************************************************************************
class CharExtractor: public SingleItemExtractor
   {
   public:
      CharExtractor(StringInt pos): SingleItemExtractor(pos)
         {
         // empty
         }
      CharExtractor(const CharExtractor &e): SingleItemExtractor(e)
         {
         // empty
         }
      String extract(const String &str)
         {
         if (index < str.length())
            return String::valueOf(str.charAt(index));
         return String();
         }
      DEFINE_EXTRACTOR_COPIER(CharExtractor);
   private:
      CharExtractor &operator =(const CharExtractor &);
   }; // class CharExtractor

//************************************************************************
// Class WordExtractor - extracts the pos`th word from a String.
//************************************************************************
class WordExtractor: public SingleItemExtractor
   {
   public:
      WordExtractor(StringInt pos): SingleItemExtractor(pos)
         {
         // empty
         }
      WordExtractor(const WordExtractor &e): SingleItemExtractor(e)
         {
         // empty
         }
      String extract(const String &str)
         {
         return str.word(index);
         }
      DEFINE_EXTRACTOR_COPIER(WordExtractor);
   private:
      WordExtractor &operator =(const WordExtractor &);
   }; // class WordExtractor

//************************************************************************
// Class CharToEndExtractor - extracts the end of a String starting at the 
//                            fromPos`th character.
//************************************************************************
class CharToEndExtractor: public ToEndExtractor
   {
   public:
      CharToEndExtractor(StringInt fromPos): ToEndExtractor(fromPos)
         {
         // empty
         }
      CharToEndExtractor(const CharToEndExtractor &e): ToEndExtractor(e)
         {
         // empty
         }
      String extract(const String &str)
         {
         if (index < str.length())
            return str.substring(index);
         return String();
         }
      DEFINE_EXTRACTOR_COPIER(CharToEndExtractor);
   private:
      CharToEndExtractor &operator =(const CharToEndExtractor &);
   }; // class CharToEndExtractor

//************************************************************************
// Class WordToEndExtractor - extracts the last words of a String starting 
//                            at the fromPos`th word.
//************************************************************************
class WordToEndExtractor: public ToEndExtractor
   {
   public:
      WordToEndExtractor(StringInt fromPos): ToEndExtractor(fromPos)
         {
         // empty
         }
      WordToEndExtractor(const WordToEndExtractor &e): ToEndExtractor(e)
         {
         // empty
         }
      String extract(const String &str)
         {
         return str.subword(index);
         }
      DEFINE_EXTRACTOR_COPIER(WordToEndExtractor);
   private:
      WordToEndExtractor &operator =(const WordToEndExtractor &);
   }; // class WordToEndExtractor

//************************************************************************
// Class CharRangeExtractor - extracts a range of forLength characters from
//                            a String starting at the fromPos`th character.
//************************************************************************
class CharRangeExtractor: public RangeExtractor
   {
   public:
      CharRangeExtractor(StringInt fromPos, StringInt forLength):
         RangeExtractor(fromPos, forLength)
         {
         // empty
         }
      CharRangeExtractor(const CharRangeExtractor &e): RangeExtractor(e)
         {
         // empty
         }
      String extract(const String &str)
         {
         if (index + length <= str.length())
            return str.substring(index, index + length - 1);
         else if (index < str.length())
            return str.substring(index);
         return String();
         }
      DEFINE_EXTRACTOR_COPIER(CharRangeExtractor);
   private:
      CharRangeExtractor &operator =(const CharRangeExtractor &);
   }; // class CharRangeExtractor

//************************************************************************
// Class WordRangeExtractor - extracts forLength words from a String 
//                            starting with the fromPos`th word.
//************************************************************************
class WordRangeExtractor: public RangeExtractor
   {
   public:
      WordRangeExtractor(StringInt fromPos, StringInt forLength):
         RangeExtractor(fromPos, forLength)
         {
         // empty
         }
      WordRangeExtractor(const WordRangeExtractor &e): RangeExtractor(e)
         {
         // empty
         }
      String extract(const String &str)
         {
         return str.subword(index, index + length - 1);
         }
      DEFINE_EXTRACTOR_COPIER(WordRangeExtractor);
   private:
      WordRangeExtractor &operator =(const WordRangeExtractor &);
   }; // class WordRangeExtractor

/* there is no inlines file associated with this header file
#ifdef ENABLE_INLINES
   #include "extract.inl"
#endif
*/

#endif   // #ifndef EXTRACT_H_
