//************************************************************************
// Declarations of all Overlayer classes, and definitions of all of their
// methods.
//------------------------------------------------------------------------
// Copyright (C) 1997 by James MacKay
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// The author can be contacted via e-mail at: jmackay@sentex.net
//
//************************************************************************
#ifndef OVERLAY_H_
#define OVERLAY_H_

#include "jstring.h"
#include "release.h"

#define DEFINE_OVERLAYER_COPIER(overlayerClassName)      \
   Overlayer *getCopy() const                            \
      {                                                  \
      return new overlayerClassName(*this);              \
      }

//************************************************************************
// Class Overlayer - base class for all overlayers.
//************************************************************************
class Overlayer
   {
   public:
      virtual StringInt overlay(StringBuffer &buf, 
                                const String &str, StringInt nextPos) = 0;
      virtual Overlayer *getCopy() const = 0;
      virtual ~Overlayer()
         {
         // empty
         }
   protected:
      // overlays str onto buf at the atPos`th position. Returns the
      // position after str in buf.
      StringInt overlayStringAtPos(StringBuffer &buf, const String &str, 
                                   StringInt atPos)
         {
         require(atPos > 0);

         if (str.length() > 0)
            {
            if (atPos > buf.length())
               buf.setLength(atPos);
            buf.overlay(atPos - 1, str);
            }
         return atPos + str.length();
         }
   private:
      Overlayer &operator =(const Overlayer &);
   }; // class Overlayer

//************************************************************************
// Class SizedOverlayer - parent for all overlayers that size the String 
//                        to a specific length before overlaying it onto
//                        the StringBuffer.
//************************************************************************
class SizedOverlayer: public Overlayer
   {
   public:
      SizedOverlayer(StringInt overlaySize): Overlayer()
         {
         require(overlaySize > 0);
         size = overlaySize;
         }
      SizedOverlayer(const SizedOverlayer &ov): Overlayer(ov)
         {
         size = ov.size;
         }
   protected:
      StringInt size;
   private:
      SizedOverlayer &operator =(const SizedOverlayer &);
   }; // class SizedOverlayer

//************************************************************************
// Class PositionOverlayer - overlays starting at the pos`th position.
//************************************************************************
class PositionOverlayer: public Overlayer
   {
   public:
      PositionOverlayer(StringInt pos): Overlayer()
         {
         require(pos > 0);
         index = pos - 1;
         }
      PositionOverlayer(const PositionOverlayer &ov): Overlayer(ov)
         {
         index = ov.index;
         }
      // overlays str onto buf at the (index+1)th position. Returns the
      // position after str in buf.
      StringInt overlay(StringBuffer &buf, 
                        const String &str, StringInt)
         {
         return overlayStringAtPos(buf, str, index + 1);
         }
      DEFINE_OVERLAYER_COPIER(PositionOverlayer);
   protected:
      StringInt index;
   private:
      PositionOverlayer &operator =(const PositionOverlayer &);
   }; // class PositionOverlayer

//************************************************************************
// Class SizedPositionOverlayer - pads/chops the String to length size, 
//                                then overlays starting at the pos`th 
//                                position.
//************************************************************************
class SizedPositionOverlayer: public SizedOverlayer
   {
   public:
      SizedPositionOverlayer(StringInt pos, StringInt size):
         SizedOverlayer(size)
         {
         require(pos > 0);
         index = pos - 1;
         }
      SizedPositionOverlayer(const SizedPositionOverlayer &ov):
         SizedOverlayer(ov)
         {
         index = ov.index;
         }
      // overlays str, sized to size size, onto buf at the (index+1)th
      // position. Returns the position after the sized str in buf.
      StringInt overlay(StringBuffer &buf, 
                        const String &str, StringInt)
         {
         return overlayStringAtPos(buf, str.left(size), index + 1);
         }
      DEFINE_OVERLAYER_COPIER(SizedPositionOverlayer);
   protected:
      StringInt index;
   private:
      SizedPositionOverlayer &operator =(const SizedPositionOverlayer &);
   }; // class SizedPositionOverlayer

//************************************************************************
// Class NextOverlayer - overlays starting at the nextPos`th position.
//************************************************************************
class NextOverlayer: public Overlayer
   {
   public:
      NextOverlayer(): Overlayer()
         {
         // empty
         }
      NextOverlayer(const NextOverlayer &ov): Overlayer(ov)
         {
         // empty
         }
      // overlays str onto buf at the nextPos`th position. Returns the
      // position after str in buf.
      StringInt overlay(StringBuffer &buf, 
                        const String &str, StringInt nextPos)
         {
         require(nextPos > 0);
         return overlayStringAtPos(buf, str, nextPos);
         }
      DEFINE_OVERLAYER_COPIER(NextOverlayer);
   private:
      NextOverlayer &operator =(const NextOverlayer &);
   }; // class NextOverlayer

//************************************************************************
// Class SizedNextOverlayer - pads/chops the String to length size, then 
//                            overlays starting at the nextPos`th position.
//************************************************************************
class SizedNextOverlayer: public SizedOverlayer
   {
   public:
      SizedNextOverlayer(StringInt size): SizedOverlayer(size)
         {
         // empty
         }
      SizedNextOverlayer(const SizedNextOverlayer &ov): SizedOverlayer(ov)
         {
         // empty
         }
      // overlays str, sized to size size, onto buf at the nextPos`th
      // position. Returns the position after the sized str in buf.
      StringInt overlay(StringBuffer &buf, 
                        const String &str, StringInt nextPos)
         {
         require(nextPos > 0);
         return overlayStringAtPos(buf, str.left(size), nextPos);
         }
      DEFINE_OVERLAYER_COPIER(SizedNextOverlayer);
   private:
      SizedNextOverlayer &operator =(const SizedNextOverlayer &);
   }; // class SizedNextOverlayer

//************************************************************************
// Class NextWordOverlayer - prepends a space to the String, then overlays
//                           starting at the nextPos`th position.
//************************************************************************
class NextWordOverlayer: public Overlayer
   {
   public:
      NextWordOverlayer(): Overlayer()
         {
         // empty
         }
      NextWordOverlayer(const NextWordOverlayer &ov): Overlayer(ov)
         {
         // empty
         }
      // overlays str, prefixed with a space, onto buf at the nextPos`th 
      // position. Returns the position after the modified str in buf.
      StringInt overlay(StringBuffer &buf, 
                        const String &str, StringInt nextPos)
         {
         require(nextPos > 0);
         return overlayStringAtPos(buf, String::valueOf(' ') + str, nextPos);
         }
      DEFINE_OVERLAYER_COPIER(NextWordOverlayer);
   private:
      NextWordOverlayer &operator =(const NextWordOverlayer &);
   }; // class NextWordOverlayer

//************************************************************************
// Class SizedNextWordOverlayer - pads/chops the String to length size,
//                                prepends a space to it, then overlays
//                                starting at the nextPos`th position.
//************************************************************************
class SizedNextWordOverlayer: public SizedOverlayer
   {
   public:
      SizedNextWordOverlayer(StringInt size): SizedOverlayer(size)
         {
         // empty
         }
      SizedNextWordOverlayer(const SizedNextWordOverlayer &ov):
         SizedOverlayer(ov)
         {
         // empty
         }
      // overlays str, sized to size size then prefixed with a space, onto 
      // buf at the nextPos`th position. Returns the position after the 
      // modified str in buf.
      StringInt overlay(StringBuffer &buf, 
                        const String &str, StringInt nextPos)
         {
         require(nextPos > 0);
         return overlayStringAtPos(buf, String::valueOf(' ') + str.left(size),
                                   nextPos);
         }
      DEFINE_OVERLAYER_COPIER(SizedNextWordOverlayer);
   private:
      SizedNextWordOverlayer &operator =(const SizedNextWordOverlayer &);
   }; // class SizedNextWordOverlayer

/* there is no inlines file associated with this header file
#ifdef ENABLE_INLINES
   #include "overlay.inl"
#endif
*/

#endif   // #ifndef OVERLAY_H_
