//************************************************************************
// Definitions of class methods for all classes declared in pairfact.h.
//------------------------------------------------------------------------
// Copyright (C) 1997 by James MacKay
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// The author can be contacted via e-mail at: jmackay@sentex.net
//
//************************************************************************

#include "pairfact.h"

#ifndef ENABLE_INLINES
   #pragma hdrstop         // prevent attempts to precompile .inl file
   #include "pairfact.inl"
#endif

//************************************************************************
// String constants.
//************************************************************************

#define MISSING_BOTH_RANGES_MSG \
   "an input range and an output range must follow the last argument."
#define IS_INVALID_INPUT_RANGE_MSG \
   " is an invalid input range."
#define MISSING_OUTPUT_RANGE_MSG    \
   "an output range must follow the last argument."
#define IS_INVALID_OUTPUT_RANGE_MSG \
   " is an invalid output range."

//************************************************************************
// Initialization of PairFactory`s static data.
//************************************************************************

const String PairFactory::rangeSeparator = ".";
const String PairFactory::toEndLength    = "-";

const String PairFactory::wordsTag             = "words";
const StringInt PairFactory::minWordsTagLength = 1;

const String PairFactory::nextWordTag = "nextword";
const StringInt PairFactory::minNextWordTagLength = 5;

const int PairFactory::MAX_RANGE_DIGITS = 4;

//************************************************************************
// Class PairFactory`s method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Given a String args of arguments, attempts to build an Extractor from
// the first n arguments (for some n > 0). If successful, a pointer to
// the Extractor is returned; otherwise errMsg will point to a String
// explaining why the Extractor couldn`t be built and 0 is returned.
// In either case args will point to a String equal to the original args 
// with the first n arguments removed.
//------------------------------------------------------------------------
Extractor *PairFactory::createExtractor(String *args, String *errMsg)
   {
   require(args != 0);
   require(errMsg != 0);

   Extractor *ex = 0;

   // remove any wordsTag and set isWordRange accordingly:
   String inputRange = args->word(0);
   *args = args->subword(1);
   boolean isWordRange = false;
   if ( inputRange.toLowerCase().isAbbreviationOf(wordsTag.toLowerCase(), 
                                                  minWordsTagLength) )
      {
      isWordRange = true;
      inputRange = args->word(0);
      *args = args->subword(1);
      }
 
   if (inputRange.length() == 0)
      {
      *errMsg = MISSING_BOTH_RANGES_MSG;
      return ex;
      }

   // point ex at an Extractor specified by isWordRange and inputRange:
   StringInt sepIndex = inputRange.indexOf(rangeSeparator);
   if (sepIndex < 0 && isValidRangePos(inputRange))               // n
      {
      if (isWordRange)
         ex = new WordExtractor(stringToInt(inputRange));
      else
         ex = new CharExtractor(stringToInt(inputRange));
      }
   else if (sepIndex > 0)                                         // ?.?
      {
      String startPos = inputRange.substring(0, sepIndex - 1);
      String length   = inputRange.substring(sepIndex + 1);

      if (isValidRangePos(startPos))                              // n.?
         {
         if (isValidRangeLength(length))                          // n.m
            {
            if (isWordRange)
               ex = new WordRangeExtractor(stringToInt(startPos),
                                           stringToInt(length));
            else
               ex = new CharRangeExtractor(stringToInt(startPos),
                                           stringToInt(length)); 
            }
         else if (length.equals(toEndLength))                     // n.-
            {
            if (isWordRange)
               ex = new WordToEndExtractor(stringToInt(startPos));
            else
               ex = new CharToEndExtractor(stringToInt(startPos));
            }
         }  // if (isValidRangePos(startPos))
      }  // else if (sepIndex > 0)

   if (ex == 0)
      *errMsg = inputRange + IS_INVALID_INPUT_RANGE_MSG;

   ensure(args != 0);
   ensure(errMsg != 0);
   return ex;

   }  // PairFactory::createExtractor(String *, String *)

//------------------------------------------------------------------------
// Given a String args of arguments, attempts to build an Overlayer from
// the first n arguments (for some n > 0). If successful, a pointer to
// the Overlayer is returned; otherwise errMsg will point to a String
// explaining why the Overlayer couldn`t be built and 0 is returned.
// In either case args will point to a String equal to the original args
// with the first n arguments removed.
//------------------------------------------------------------------------
Overlayer *PairFactory::createOverlayer(String *args, String *errMsg)
   {
   require(args != 0);
   require(errMsg != 0);

   Overlayer *ov = 0;

   String outputRange = args->word(0);
   *args = args->subword(1);
   if (outputRange.length() == 0)
      {
      *errMsg = MISSING_OUTPUT_RANGE_MSG;
      return ov;
      }

   // build the Overlayer specified by outputRange:
   StringInt sepIndex = outputRange.indexOf(rangeSeparator);
   if (sepIndex < 0)                                           // ?
      {
      if (isValidRangePos(outputRange))                        // n
         ov = new PositionOverlayer(stringToInt(outputRange));
		                                                         // n[extword]
      else if (outputRange.toLowerCase().
                  isAbbreviationOf(nextWordTag.toLowerCase(), 1))
         {
         if (outputRange.length() >= minNextWordTagLength)     // nextword
            ov = new NextWordOverlayer();
         else                                                  // next
            ov = new NextOverlayer();
         }
      }
   else if (sepIndex > 0)                                      // ?.?
      {
      String startPos = outputRange.substring(0, sepIndex - 1);
      String length   = outputRange.substring(sepIndex + 1);

      if (isValidRangePos(startPos) && isValidRangeLength(length)) // m.n
         ov = new SizedPositionOverlayer(stringToInt(startPos),
                                         stringToInt(length));
                                                             // n[extword].m
      else if (startPos.toLowerCase().
                  isAbbreviationOf(nextWordTag.toLowerCase(), 1) &&
               isValidRangeLength(length))
         {
         StringInt len = stringToInt(length);
         if (startPos.length() >= minNextWordTagLength)     // nextword.m
            ov = new SizedNextWordOverlayer(len);
         else                                               // next.m
            ov = new SizedNextOverlayer(len);
         }
      }  // else if sepIndex > 0

   if (ov == 0)
      *errMsg = outputRange + IS_INVALID_OUTPUT_RANGE_MSG;

   ensure(args != 0);
   ensure(errMsg != 0);
   return ov;

   }  // PairFactory::createOverlayer(String *, String *)

//------------------------------------------------------------------------
// Returns the number of partial or complete Extractor-Overlayer pairs
// that the argument string args specifies, assuming the arguments are
// all valid.
//------------------------------------------------------------------------
int PairFactory::countPairs(const String &args)
   {
   int count = 0;
   WordIterator iter = args;
   String lowWordsTag = wordsTag.toLowerCase();

   while (iter.hasMoreWords())
      {
      count++;
 
      // discard the arguments specifiying the Extractor:
      if (iter.nextWord().toLowerCase().
                          isAbbreviationOf(lowWordsTag, minWordsTagLength))
         iter.nextWord();

      // discard the arguments specifying the corresponding Overlayer:
      iter.nextWord();
      }
   return count;
   }

//------------------------------------------------------------------------
// Returns true iff pos is a valid range position.
//------------------------------------------------------------------------
boolean PairFactory::isValidRangePos(const String &pos) const
   {
   boolean isValid = false;
   if (pos.length() > 0 && pos.length() <= MAX_RANGE_DIGITS &&
       pos.nonDigitIndex() < 0)
      {
      long temp = atol(pos);
      if (temp > 0 && temp <= maxStringInt)
         isValid = true;
      }
   return isValid;
   }

