//************************************************************************
// Definitions of methods for all classes declared in append.h; and
// declarations and method definitions of all implementation filter
// classes: AppendData, AppendString, AppendFile and AppendStream.
//------------------------------------------------------------------------
// Copyright (C) 1997 by James MacKay
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// The author can be contacted via e-mail at: jmackay@sentex.net
//
//************************************************************************

#include "append.h"

#ifndef ENABLE_INLINES
   #pragma hdrstop         // prevent attempts to precompile .inl file
   #include "append.inl"
#endif

//************************************************************************
// Abstract class AppendData`s declaration and method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Abstract base class for filters that append data to their input.
//------------------------------------------------------------------------
class AppendData: public SingleFilter
   {
   public:
      AppendData();
      AppendData(const AppendData &f);

      void processLine(const String &line);
   protected:
      void beforeInput();
      void afterInput();

      virtual void outputData(StringBuffer &buf) = 0;
   private:
      AppendData &operator =(const AppendData &f);

      String previousLine;

   }; // class AppendData

//------------------------------------------------------------------------
// Default constructor.
//------------------------------------------------------------------------
AppendData::AppendData(): SingleFilter(), previousLine()
   {
   // empty
   }

//------------------------------------------------------------------------
// Copy constructor.
//------------------------------------------------------------------------
AppendData::AppendData(const AppendData &f):
   SingleFilter(f), previousLine(f.previousLine)
   {
   // empty
   }

//------------------------------------------------------------------------
// Empty previousLine.
//------------------------------------------------------------------------
void AppendData::beforeInput()
   {
   previousLine = String();
   }

//------------------------------------------------------------------------
// Output the previousLine then store the current line as the previousLine.
//------------------------------------------------------------------------
INLINE void AppendData::processLine(const String &line)
   {
	// output the previous line iff there has been one:
   if (previousLine.length() > 0)
      outputLine(previousLine);
   previousLine = line;
   }

//------------------------------------------------------------------------
// Output the last line with the data to be appended appended to it.
//------------------------------------------------------------------------
void AppendData::afterInput()
   {
   StringBuffer buf;
   if (previousLine.endsWith(EOLString))
      outputLine(previousLine);
   else
      buf = previousLine;

   outputData(buf);
   if (buf.length() > 0)
      outputLine(buf);

   previousLine = String();
   }

//************************************************************************
// Class AppendString`s declaration and method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Filter that appends the specified string to its input.
//------------------------------------------------------------------------
class AppendString: public AppendData
   {
   public:
      AppendString(const String &str);
      AppendString(const AppendString &f);

      DEFINE_FILTER_COPIER(AppendString);
   protected:
      void outputData(StringBuffer &buf);
   private:
      AppendString &operator =(const AppendString &f);

      const String theString;
 
   }; // class AppendString

//------------------------------------------------------------------------
// Constructs an AppendString filter from the string to append.
//------------------------------------------------------------------------
AppendString::AppendString(const String &str): AppendData(), theString(str)
   {
   // empty
   }

//------------------------------------------------------------------------
// Copy constructor.
//------------------------------------------------------------------------
AppendString::AppendString(const AppendString &f):
   AppendData(f), theString(f.theString)
   {
   // empty
   }

//------------------------------------------------------------------------
// Outputs the String to be appended, prepended with the contents of buf,
// and leaving the last partial line (if any) in buf.
//------------------------------------------------------------------------
INLINE void AppendString::outputData(StringBuffer &buf)
   {
   outputString(theString, buf);
   }

//************************************************************************
// Class AppendFile`s declaration and method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Filter that appends the contents of the specified file to its input.
//------------------------------------------------------------------------
class AppendFile: public AppendData
   {
   public:
      AppendFile(const String &pathname);
      AppendFile(const AppendFile &f);

      DEFINE_FILTER_COPIER(AppendFile);
   protected:
      void outputData(StringBuffer &buf);
   private:
      AppendFile &operator =(const AppendFile &f);

      const String theFile;
 
   }; // class AppendFile

//------------------------------------------------------------------------
// Constructs an AppendFile filter from the pathname of the file to append.
//------------------------------------------------------------------------
AppendFile::AppendFile(const String &pathname): 
   AppendData(), theFile(pathname)
   {
   // empty
   }

//------------------------------------------------------------------------
// Copy constructor.
//------------------------------------------------------------------------
AppendFile::AppendFile(const AppendFile &f):
   AppendData(f), theFile(f.theFile)
   {
   // empty
   }

//------------------------------------------------------------------------
// Outputs the contents of the file to be appended, prepended with the
// contents of buf, and leaving the last partial line (if any) in buf.
//------------------------------------------------------------------------
INLINE void AppendFile::outputData(StringBuffer &buf)
   {
   outputFile(theFile, buf);
   }

//************************************************************************
// Class AppendStream`s declaration and method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Filter that appends all of the data on the specified input stream to
// its input.
//------------------------------------------------------------------------
class AppendStream: public AppendData
   {
   public:
      AppendStream(istream &in);
      AppendStream(const AppendStream &f);

      DEFINE_FILTER_COPIER(AppendStream);
   protected:
      void outputData(StringBuffer &buf);
   private:
      AppendStream &operator =(const AppendStream &f);

      istream *theStream;
 
   }; // class AppendStream

//------------------------------------------------------------------------
// Constructs an AppendStream filter from the istream whose contents are
// to be appended.
//------------------------------------------------------------------------
AppendStream::AppendStream(istream &in):
   AppendData(), theStream(&in)
   {
   ensure(theStream != 0);
   }

//------------------------------------------------------------------------
// Copy constructor.
//------------------------------------------------------------------------
AppendStream::AppendStream(const AppendStream &f):
   AppendData(f), theStream(f.theStream)
   {
   ensure(theStream != 0);
   }

//------------------------------------------------------------------------
// Outputs the contents of the stream to be appended, prepended with the 
// contents of buf, and leaving the last partial line (if any) in buf.
//------------------------------------------------------------------------
INLINE void AppendStream::outputData(StringBuffer &buf)
   {
   outputStream(*theStream, buf);
   }

//************************************************************************
// Class Append method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Constructs an Append filter from the string to append.
//------------------------------------------------------------------------
Append Append::string(const String &str)
   {
   return Append(AppendString(str), 0);
   }

//------------------------------------------------------------------------
// Constructs an Append filter from the pathname of the file to append.
//------------------------------------------------------------------------
Append Append::file(const String &pathname)
   {
   return Append(AppendFile(pathname), 0);
   }

//------------------------------------------------------------------------
// Constructs an Append filter from the stream whose contents are to
// be appended.
//------------------------------------------------------------------------
Append Append::stream(istream &in)
   {
   return Append(AppendStream(in), 0);
   }

//------------------------------------------------------------------------
// Copy constructor.
//------------------------------------------------------------------------
Append::Append(const Append &f): ShellFilter(f)
   {
   // empty
   }

