//************************************************************************
// Does a full-coverage test of all of the code in each of the following
// example filter classes.
//------------------------------------------------------------------------
// Copyright (C) 1997 by James MacKay
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// The author can be contacted via e-mail at: jmackay@sentex.net
//
//************************************************************************

#include <fstream.h>
#include <new.h>

#include "passthru.h"
#include "take.h"
#include "drop.h"
#include "append.h"
#include "prepend.h"
#include "writeto.h"
#include "generror.h"
#include "outerror.h"
#include "translat.h"
#include "include.h"

#include "jstring.h"
#include "release.h"

#if TARGET_PLATFORM == LINUX_PLATFORM
   #include <strstream.h>
#else
   #include <strstrea.h>
#endif

//------------------------------------------------------------------------
// String constants.
//------------------------------------------------------------------------

// the next two must be as defined in include.cpp:
#define EOL_IN_TAG_MSG \
   "there is an EOL character in the tag "
#define EMPTY_TAG_MSG \
   "an empty tag was passed to IncludeTagged`s constructor"

//------------------------------------------------------------------------
// Note: the correct output for test4() in egtest.val contains the string
//       BAD_LINE_NUM_MSG as defined in generror.cpp. So if that message
//       changes then egtest.val must be updated accordingly.
//------------------------------------------------------------------------

//------------------------------------------------------------------------
// Function prototypes.
//------------------------------------------------------------------------
void testFilter(const ComponentFilter &filterToTest, const String &input,
                ostream &outputStream);
void stringToFile(const String &filename, const String &fileContents);
void test1();
void test2();
void test3();
void test4();
void test5();
void test6();
void test7();
void test8();
void test9();
void test10();

const char outputFile[] = "output.tst";
ofstream outStream(outputFile);

//------------------------------------------------------------------------
// Tests the filterToTest by passing it the data in input and outputting
// the results to the outputStream. The filterToTest`s copy constructor
// is also executed.
//------------------------------------------------------------------------
void testFilter(const ComponentFilter &filterToTest, const String &input,
                ostream &outputStream)
   {
   String output;

   Append  in  = Append::string(input);
   WriteTo out = WriteTo::string(output, WriteTo::OVERWRITE);

   // Using a Filter here allows us to test the filterToTest`s copy
   // constructor:
   Filter f = filterToTest;
   f.connect(in, out);
   outputStream << output;
   }

//------------------------------------------------------------------------
// Creates a file named filename with contents fileContents.
//------------------------------------------------------------------------
void stringToFile(const String &filename, const String fileContents)
   {
   Append in = Append::string(fileContents);
   WriteTo out = WriteTo::file(filename, WriteTo::OVERWRITE);   
   WriteTo zip = WriteTo::nothing();

   out.connect(in, zip);
   }

//------------------------------------------------------------------------
// Function called when new fails.
//------------------------------------------------------------------------
void onNewFailure()
   {
	outStream << "Out of memory - exiting." << endl;
   exit(-1);
	}

//------------------------------------------------------------------------
// main().
//------------------------------------------------------------------------
int main()
   {
   set_new_handler(&onNewFailure);

   test1();
   test2();
   test3();
   test4();
   test5();
   test6();
   test7();
   test8();
   test9();
   test10();

   return 0;
   }

//------------------------------------------------------------------------
// #1. Tests the PassThrough filter.
//------------------------------------------------------------------------
void test1()
   {
   String input1 =  "If all the world were made of cheese,\n"
                    "And all the seas of ink ...\n";
   String input2 = "";

   outStream << "#1" << endl;

   PassThrough pt;
   testFilter(pt, input1, outStream);   
   testFilter(pt, input2, outStream);
   } // test1()

//------------------------------------------------------------------------
// #2. Tests the Take filter.
//------------------------------------------------------------------------
void test2()
   {
   // 7 lines, the last of which does NOT end with a newline:
   String input1 = "Principal characters:\n"
                   "  ARTHUR, King of Britain\n  QUEEN GUINEVERE, his wife\n"
                   "  SIR MORDRED, his natural son\n  SIR LAUNCELOT\n"
                   "  SIR TRISTRAM, and\n  SIR GALAHAD, knights of the Round Table";

   outStream << "#2" << endl;

   Take take1(1);
   testFilter(take1, input1, outStream);      // outputs first line
   testFilter(take1, input1, outStream);      // re-outputs first line

   Take nothing(0);
   testFilter(nothing, input1, outStream);    // outputs nothing

   Take all(8);
   testFilter(all, input1, outStream);        // outputs all of input1
   outStream << endl;                         // and a newline
   } // test2()

//------------------------------------------------------------------------
// #3. tests the Drop filter.
//------------------------------------------------------------------------
void test3()
   {
   // 7 lines, the last of which does NOT end with a newline:
   String input1 = "Principal characters:\n"
                   "  ARTHUR, King of Britain\n  QUEEN GUINEVERE, his wife\n"
                   "  SIR MORDRED, his natural son\n  SIR LAUNCELOT\n"
                   "  SIR TRISTRAM, and\n  SIR GALAHAD, knights of the Round Table";

   outStream << "#3" << endl;

   Drop drop1(1);
   testFilter(drop1, input1, outStream);   // outputs all but the first line
   testFilter(drop1, input1, outStream);   // re-outputs all but the first line

   Drop nothing(0);
   testFilter(nothing, input1, outStream); // outputs all of input1

   Drop all(7);
   testFilter(all, input1, outStream);     // outputs nothing
   outStream << endl;                      // and a newline
   } // test3()

//------------------------------------------------------------------------
// #4. tests the GenerateError and OutputError filters.
//------------------------------------------------------------------------
void test4()
   {
	// 3 lines, no newline at end of last line.
   String input1 = "Once upon a time\nit was a dark and stormy\nnight.";

   String beforeErrMsg = "Before error generated.";
   String afterErrMsg  = "After error ...\ngenerated.";
   String forLineErrMsg0 = "Error generated ...\nat line 0\n";
   String forLineErrMsg1 = "Error generated ...\nat line 1\n";
   String forLineErrMsg3 = "Error generated ...\nat line 3\n";
   String forLineErrMsg4 = "Error generated ...\nat line 4\n";

   GenerateError before = GenerateError::before(beforeErrMsg);
   GenerateError after  = GenerateError::after(afterErrMsg);
   GenerateError at1    = GenerateError::forLine(1, forLineErrMsg1);
   GenerateError at3    = GenerateError::forLine(3, forLineErrMsg3);
   GenerateError at4    = GenerateError::forLine(4, forLineErrMsg4);
   GenerateError at0    = GenerateError::forLine(0, forLineErrMsg0);

   OutputError errOut = OutputError::toNextFilter();

   outStream << "#4" << endl;

   // Note that OutputError filters, such as errOut, cause their last line of
   // input to always end with a newline

   Filter f = before | errOut;
   testFilter(f, input1, outStream);   // beforeErrMsg + input1 + newline

   f = after | errOut;
   testFilter(f, input1, outStream);   // input1 + afterErrMsg + newline

   f = at1 | errOut;
   testFilter(f, input1, outStream);   // "Once upon a time\n" + forLineErrMsg1
                                       // + "it was [...] night." + newline

   testFilter(f, input1, outStream);   // [same output as previous]

   f = at3 | errOut;
   testFilter(f, input1, outStream);   // "Once upon a time\nit was [...] stormy\n"
                                       // "night." + forLineErrMsg3 + newline

   f = at4 | errOut;
   testFilter(f, input1, outStream);   // input1 + newline

   // tests the case where GenerateError::forLine() is passed an invalid
   // line number:
   f = at0 | errOut;
   testFilter(f, input1, outStream);   // bad line # msg + input1 + newline

   // tests OutputError::toStream():
   OutputError errOut2 = OutputError::toStream(outStream);
   f = before | errOut2;
   testFilter(f, input1, outStream);   // beforeErrMsg + newline + input1
   outStream << endl;                  // + newline

   } // test4()

//------------------------------------------------------------------------
// #5. Tests Append and WriteTo filters.
//------------------------------------------------------------------------
// Note: since a newline may be appended to the last line of data
//       written to a file, all inputs written to a file end with a
//       newline already.
//------------------------------------------------------------------------
void test5()
   {
   String input1 = "Once upon a time\nthere lived a man from\nNantucket\n";
   String input2 = "... and they lived happily ever after\nTHE END.\n";
   String output = "Initial data that should be overwritten.";
   String testFile1 = "test5-1.tst";

   outStream << "#5" << endl;

   // Note: assignments to Filters tests Append and WriteTo copy constructors

   Filter in = Append::string(input1);

   Append appendInput2 = Append::string(input2);
   WriteTo writeToTestFile1 = WriteTo::file(testFile1, WriteTo::OVERWRITE);
   Filter out = writeToTestFile1;
   appendInput2.connect(in, out);   // writes input1 + input2 to testFile1

   Append appendTestFile1 = Append::file(testFile1);
   WriteTo writeToOutputStream = WriteTo::stream(outStream);
   out = writeToOutputStream;
   appendTestFile1.connect(in, out);  //   input1 + contents of testFile1
                                      // = input1 + input1 + input2

   istrstream input2Stream(input2);
   Append appendInput2Stream = Append::stream(input2Stream);
   WriteTo writeToOutputString = WriteTo::string(output, WriteTo::OVERWRITE);
   out = writeToOutputString;
   appendInput2Stream.connect(in, out);  // input1 + input2
   outStream << output;

   out = WriteTo::nothing();
   appendTestFile1.connect(in, out);     // outputs nothing

   WriteTo appendToTestFile1 = WriteTo::file(testFile1, WriteTo::APPEND);
   out = appendToTestFile1;
   appendInput2.connect(in, out);   // appends input1 + input2 to testFile1, so
                                    // testFile1 = input1 + input2 + input1 + input2

   // the following appends input1 + the contents of testFile1 to output, so
   // we`ll have output = output + input1 + the contents of testFile1
   //                   = (input1 + input2) + input1 +
   //                     (input1 + input2 + input1 + input2)
   WriteTo appendToOutputString = WriteTo::string(output, WriteTo::APPEND);
   out = appendToOutputString;
   appendTestFile1.connect(in, out);   // see above comment for output
   outStream << output << endl;        // + newline

   // test the minimize() methods of those Append and WriteTo filters
   // that have overridden it:
   appendToOutputString.minimize();
   } // test5()

//------------------------------------------------------------------------
// #6. Tests Prepend and WriteTo filters.
//------------------------------------------------------------------------
// Note: since a newline may be appended to the last line of data
//       written to a file, all inputs written to a file end with a
//       newline already.
//------------------------------------------------------------------------
void test6()
   {
   String input1 = "Once upon a time\nthere lived a man from\nNantucket\n";
   String input2 = "... and they lived happily ever after\nTHE END.\n";
   String output = "Initial data that should be overwritten.";
   String testFile1 = "test6-1.tst";

   outStream << "#6" << endl;

   Filter in = Prepend::string(input1);

   Prepend prependInput2 = Prepend::string(input2);
   WriteTo writeToTestFile1 = WriteTo::file(testFile1, WriteTo::OVERWRITE);
   Filter out = writeToTestFile1;
   prependInput2.connect(in, out);   // writes input2 + input1 to testFile1

   Prepend prependTestFile1 = Prepend::file(testFile1);
   WriteTo writeToOutputStream = WriteTo::stream(outStream);
   out = writeToOutputStream;
   prependTestFile1.connect(in, out);  //   contents of testFile1 + input1
                                       // = input2 + input1 + input1

   istrstream input2Stream(input2);
   Prepend prependInput2Stream = Prepend::stream(input2Stream);
   WriteTo writeToOutputString = WriteTo::string(output, WriteTo::OVERWRITE);
   out = writeToOutputString;
   prependInput2Stream.connect(in, out);  // input2 + input1
   outStream << output;

   out = WriteTo::nothing();
   prependTestFile1.connect(in, out);     // outputs nothing

   WriteTo appendToTestFile1 = WriteTo::file(testFile1, WriteTo::APPEND);
   out = appendToTestFile1;
   prependInput2.connect(in, out);   // prepends testFile1 with input2 + input1, so
                                     // testFile1 = input2 + input1 + input2 + input1

   // the following appends the contents of testFile1 + input1 to output, so
   // we`ll have output = output + the contents of testFile1 + input1
   //                   = (input2 + input1) + (input2 + input1 + input2 + input1) +
   //                     input1
   WriteTo appendToOutputString = WriteTo::string(output, WriteTo::APPEND);
   out = appendToOutputString;
   prependTestFile1.connect(in, out);   // see above comment for output 
   outStream << output << endl;         // + newline

   // test the PrependData.minimize() method:
   prependTestFile1.minimize();   

   } // test6()

//------------------------------------------------------------------------
// #7. Tests code in Append, Prepend and WriteTo filters that isn`t
//     exercised by the previous tests - mostly error handling code.
//------------------------------------------------------------------------
void test7()
   {



   outStream << "#7" << endl;

   }

//------------------------------------------------------------------------
// #8. Tests Translate and CharacterFilter filters.
//------------------------------------------------------------------------
void test8()
   {
   String input1 = "LineOne LineTwo\nWordTwo LineThree LineFour LineFive\n"
                   "WordTwo\nWordThree\nWordFour\n";

   outStream << "#8" << endl;

   Translate tr1(" \n", "\n ");            // switches newlines and spaces
   testFilter(tr1, input1, outStream);
   outStream << '/' << endl;               // + '/' + newline 

   Translate upper(Translate::TO_UPPER_CASE);
   testFilter(upper, input1, outStream);   // input1 all in upper case

   Translate lower(Translate::TO_LOWER_CASE);
   testFilter(lower, input1, outStream);   // input1 all in lower case

   tr1.minimize();

   // test building Translate filters with maps of unequal size:
   Translate tr2("LINEline", "WORD");
   Translate tr3("line", "wordWORD");
   testFilter(tr2, input1, outStream);  // input1 with 'L' -> 'W'
   testFilter(tr3, input1, outStream);  // input1 with 'l'->'w', 'i'->'o',
                                        // 'n'->'r' and 'e'->'d'
   } // test8()

//------------------------------------------------------------------------
// #9. Tests the Include filter.
//------------------------------------------------------------------------
void test9()
   {
   String testFile1 = "test9-1.tst";
   String testFile2 = "test9-2.tst";

   String input1 = String(":") + testFile1 + ": <-- the first file, and " +
                   "the second file is here --> :\t" + testFile2 + 
                   " :, but the first file appears again here --> \t:  " + 
                   testFile1 + "\t\t:";
   String input2 = String("<include> \t </include>");
   String input3 = String("This is a line --> : ") + testFile1 +
                   " : <-- with a file in the middle of it.";

   // build the testFiles:
   String fileContents1 = "And then there were three\n"
                          "No, wait a moment ... two\n";
   String fileContents2 = "... and there goes the last one.\n";
   stringToFile(testFile1, fileContents1);
   stringToFile(testFile2, fileContents2);

   outStream << "#9" << endl;

   Include inc1 = Include::tagged(":", ":");
   testFilter(inc1, input1, outStream);
   outStream << endl;                          // + newline

   Include inc2 = Include::tagged("<include>", "</include>");
   testFilter(inc2, input2, outStream);        // ""
   outStream << endl;                          // + newline

   inc1.minimize();
   testFilter(inc1, input3, outStream);
   outStream << endl;                          // + newline

   // test validation of tags:
   String emptyTag = "";
   String EOLTag = "two\nparter";
   String validTag = "::";
   try
      {
      Include inc3 = Include::tagged(emptyTag, validTag);
      }
   catch (InvalidFilterArgumentException e)
      {
      if (e.getMessage().indexOf(EMPTY_TAG_MSG) == 0)
         outStream << "Empty tag InvalidFilterArgumentException "
                   << "correctly thrown" << endl;
      else
         outStream << "Incorrect InvalidFilterArgumentException thrown" << endl;
      }
   try
      {
      Include inc3 = Include::tagged(validTag, EOLTag);
      }
   catch (InvalidFilterArgumentException e)
      {
      if (e.getMessage().indexOf(EOL_IN_TAG_MSG) == 0)
         outStream << "EOL in tag InvalidFilterArgumentException "
                   << "correctly thrown" << endl;
      else
         outStream << "Incorrect InvalidFilterArgumentException thrown" << endl;
      }

   } // test9()

//------------------------------------------------------------------------
// #10. 
//------------------------------------------------------------------------
void test10()
   {

   outStream << "#10" << endl;

   } // test10()

