//************************************************************************
// Definitions of methods for all classes declared in generror.h.
//------------------------------------------------------------------------
// Copyright (C) 1997 by James MacKay
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// The author can be contacted via e-mail at: jmackay@sentex.net
//
//************************************************************************

#include "generror.h"

#ifndef ENABLE_INLINES
   #pragma hdrstop         // prevent attempts to precompile .inl file
   #include "generror.inl"
#endif

#define BAD_LINE_NUM_MSG   \
   "The line number passed to GenerateError::forLine() must be > 0; " \
   "it was " 

//************************************************************************
// Static data initialization.
//************************************************************************

const int GenerateError::BEFORE = 0;
const int GenerateError::AFTER  = -1;

//************************************************************************
// Class GenerateError method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Constructs a GenerateError filter from the point at which the error
// should be generated and the message from which to build the error.
//------------------------------------------------------------------------
GenerateError::GenerateError(int whenToGenerate, const String &errMsg):
   SingleFilter(), errorMsg(errMsg), errorLineNum(whenToGenerate),
   currLineNum(-1)
   {
   ensure(errorLineNum > 0 || 
          errorLineNum == BEFORE || errorLineNum == AFTER);
   }

//------------------------------------------------------------------------
// Constructs a GenerateError filter that will generate an error with
// message errMsg when its beforeInput() method is called.
//------------------------------------------------------------------------
GenerateError GenerateError::before(const String &errMsg)
   {
   return GenerateError(BEFORE, errMsg);
   }

//------------------------------------------------------------------------
// Constructs a GenerateError filter that will generate an error with
// message errMsg immediately after the lineNum`th line is output. If
// lineNum is invalid then it instead constructs a GenerateError filter 
// that generates an error indicating this when its beforeInput() method 
// is called.
//------------------------------------------------------------------------
GenerateError GenerateError::forLine(int lineNum, const String &errMsg)
   {
   if (lineNum <= 0)   // => lineNum is invalid
      {
      String lineNumInvalidMsg = String(BAD_LINE_NUM_MSG) + 
                                 String::valueOf(lineNum) + ".";
      return GenerateError(BEFORE, lineNumInvalidMsg);
      }
   return GenerateError(lineNum, errMsg);
   }

//------------------------------------------------------------------------
// Constructs a GenerateError filter that will generate an error with
// message errMsg when its afterInput() method is called.
//------------------------------------------------------------------------
GenerateError GenerateError::after(const String &errMsg)
   {
   return GenerateError(AFTER, errMsg);
   }

//------------------------------------------------------------------------
// Copy constructor.
//------------------------------------------------------------------------
GenerateError::GenerateError(const GenerateError &f):
   SingleFilter(f), errorMsg(f.errorMsg), errorLineNum(f.errorLineNum),
   currLineNum(f.currLineNum)
   {
   ensure(errorLineNum > 0 || 
          errorLineNum == BEFORE || errorLineNum == AFTER);
   }

//------------------------------------------------------------------------
// Initialize the current line number and generate the error now if that 
// is what is indicated.
//------------------------------------------------------------------------
void GenerateError::beforeInput()
   {
   currLineNum = 0;
   if (errorLineNum == BEFORE)
      outputError(errorMsg);
   }

//------------------------------------------------------------------------
// Output the line, and if this is the line at which to output an error
// then do so.
//------------------------------------------------------------------------
void GenerateError::processLine(const String &line)
   {
   // if this fails, currLineNum wasn`t set to zero in beforeInput():
   ensure(currLineNum >= 0);

   outputLine(line);

   // if currLineNum hasn`t already equalled or exceeded the errorLineNum ...
   if (currLineNum < errorLineNum)
      {
      // ... set currLineNum to the line # of the current line and check
      // whether it is the line at which to generate an error:
      currLineNum++;
      if (currLineNum == errorLineNum)
         outputError(errorMsg);
      }
   }

//------------------------------------------------------------------------
// Generate the error now if that is what is indicated.
//------------------------------------------------------------------------
void GenerateError::afterInput()
   {
   if (errorLineNum == AFTER)
      outputError(errorMsg);
   }

