//************************************************************************
// Definitions of methods for all classes declared in include.h
//------------------------------------------------------------------------
// Copyright (C) 1997 by James MacKay
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// The author can be contacted via e-mail at: jmackay@sentex.net
//
//************************************************************************

#include "include.h"

#ifndef ENABLE_INLINES
   #pragma hdrstop         // prevent attempts to precompile .inl file
   #include "include.inl"
#endif

//************************************************************************
// String constants.
//************************************************************************
#define EOL_IN_TAG_MSG \
   "there is an EOL character in the tag "
#define EMPTY_TAG_MSG \
   "an empty tag was passed to IncludeTagged`s constructor"

//************************************************************************
// Class IncludeTagged`s declaration and method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Filter that replaces each file name preceded by beginTag and followed
// by endTag with the contents of that file.
//------------------------------------------------------------------------
class IncludeTagged: public SingleFilter
   {
   public:
      // throws an InvalidFilterArgumentException iff a tag is invalid:
      IncludeTagged(const String &beginTag, const String &endTag);
      IncludeTagged(const IncludeTagged &f);

      void processLine(const String &line);
      void minimize();

      DEFINE_FILTER_COPIER(IncludeTagged);
   protected:
      void afterInput();
   private:
      // throws an InvalidFilterArgumentException iff tag is invalid:
      static void validateTag(const String &tag);
      IncludeTagged &operator =(const IncludeTagged &f);

      String beginTag, endTag;
      StringInt beginTagLength, endTagLength;
      StringBuffer outputBuffer;
      String includeFileName;

   }; // class IncludeTagged

//------------------------------------------------------------------------
// Constructs an IncludeTagged filter from the tags that must enclose a
// file name in order for that file to be included. Throws an Invalid-
// FilterArgumentException if either of the tags are invalid tags.
//------------------------------------------------------------------------
IncludeTagged::IncludeTagged(const String &beginTag, const String &endTag):
   SingleFilter(), beginTag(beginTag), endTag(endTag), outputBuffer(), 
   includeFileName()
   {
   validateTag(this->beginTag);
   validateTag(this->endTag);

   beginTagLength = this->beginTag.length();
   endTagLength   = this->endTag.length();
   }

//------------------------------------------------------------------------
// Copy constructor.
//------------------------------------------------------------------------
IncludeTagged::IncludeTagged(const IncludeTagged &f):
   SingleFilter(f), beginTag(f.beginTag), endTag(f.endTag), 
   beginTagLength(f.beginTagLength), endTagLength(f.endTagLength), 
   outputBuffer(f.outputBuffer), includeFileName(f.includeFileName)
   {
   ensure(beginTag.length() == beginTagLength);
   ensure(endTag.length() == endTagLength);
   }

//------------------------------------------------------------------------
// Replaces any tagged file name with the contents of that file.
//------------------------------------------------------------------------
void IncludeTagged::processLine(const String &line)
   {
   StringInt beginIndex, endIndex;
   StringInt startIndex = 0;   // index to start next tag search at

   // while line contains a beginTag followed by an endTag at or after startIndex
   outputBuffer.clear();
   while ( (beginIndex = line.indexOf(beginTag, startIndex)) >= 0 &&
           (endIndex = line.indexOf(endTag, beginIndex + beginTagLength)) >= 0 )
      {
      // output the part of line between the last endTag and this beginTag:
      if (beginIndex > startIndex)
         outputString(line.substring(startIndex, beginIndex - 1), outputBuffer);

      // attempt to output the tagged file:
      includeFileName = line.substring(beginIndex + beginTagLength,
                                       endIndex - 1).strip();
      outputFile(includeFileName, outputBuffer);

      startIndex = endIndex + endTagLength;
           //    = the index of the char after the endTag
      }

   // output the rest of the file and/or line:
   if (startIndex < line.length() || outputBuffer.length() > 0)
      outputLine(outputBuffer.toString() + line.substring(startIndex));

   } // IncludeTagged::processLine(const String &)

//------------------------------------------------------------------------
// Minimizes outputBuffer.
//------------------------------------------------------------------------
void IncludeTagged::minimize()
   {
   SingleFilter::minimize();
   minimizeBuffer(outputBuffer);
   }

//------------------------------------------------------------------------
// Clears the outputBuffer, allowing minimize() to shrink it.
//------------------------------------------------------------------------
void IncludeTagged::afterInput()
   {
   outputBuffer.clear();
   }

//------------------------------------------------------------------------
// Throws an InvalidFilterArgumentxception if tag is not a valid tag
// (that is, if it is empty or if it contains an EOL character).
//------------------------------------------------------------------------
void IncludeTagged::validateTag(const String &tag)
   {
   if (tag.length() == 0)
      throw InvalidFilterArgumentException(EMPTY_TAG_MSG);
   else if (tag.indexOf(EOL) >= 0)
      throw InvalidFilterArgumentException(EOL_IN_TAG_MSG);
   }

//************************************************************************
// Class Include method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Constructs an include filter that includes files whose names are
// prefixed with beginTag and suffixed with endTag. Throws an Invalid-
// FilterArgumentException if either tag is invalid.
//------------------------------------------------------------------------
Include Include::tagged(const String &beginTag, const String &endTag)
   {
   return Include(IncludeTagged(beginTag, endTag), 0);
   }

//------------------------------------------------------------------------
// Copy constructor.
//------------------------------------------------------------------------
Include::Include(const Include &f): ShellFilter(f)
   {
   // empty
   }
