//************************************************************************
// Definitions of methods for all classes declared in outerror.h.
//------------------------------------------------------------------------
// Copyright (C) 1997 by James MacKay
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// The author can be contacted via e-mail at: jmackay@sentex.net
//
//************************************************************************

#include "outerror.h"

#ifndef ENABLE_INLINES
   #pragma hdrstop         // prevent attempts to precompile .inl file
   #include "outerror.inl"
#endif

//************************************************************************
// Class OutputErrorToNextFilter`s declaration and method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Filter that outputs the error message associated with any FilterError
// passed to it to the next filter (as input).
//------------------------------------------------------------------------
class OutputErrorToNextFilter: public SingleFilter
   {
   public:
      OutputErrorToNextFilter();
      OutputErrorToNextFilter(const OutputErrorToNextFilter &f);

      void processLine(const String &line);
      void handleError(const FilterError &fe);

      DEFINE_FILTER_COPIER(OutputErrorToNextFilter);
   private:
      OutputErrorToNextFilter &operator =(const OutputErrorToNextFilter &f);
   }; // class OutputErrorToNextFilter

//------------------------------------------------------------------------
// Default constructor.
//------------------------------------------------------------------------
OutputErrorToNextFilter::OutputErrorToNextFilter(): SingleFilter()
   {
   // empty
   }

//------------------------------------------------------------------------
// Copy constructor.
//------------------------------------------------------------------------
OutputErrorToNextFilter::
   OutputErrorToNextFilter(const OutputErrorToNextFilter &f): 
   SingleFilter(f)
   {
   // empty
   }

//------------------------------------------------------------------------
// Output each line, terminated with an EOL if it isn`t already (in case
// handleError() outputs more lines after the last line processed here).
//------------------------------------------------------------------------
void OutputErrorToNextFilter::processLine(const String &line)
   {
   if (line.endsWith(EOLString))
      outputLine(line);
   else
      outputLine(line + EOLString);
   }

//------------------------------------------------------------------------
// Output fe`s message string as input to the next filter, terminated with
// a newline if it isn`t already, then pass fe on to the next filter.
//------------------------------------------------------------------------
void OutputErrorToNextFilter::handleError(const FilterError &fe)
   {
   StringBuffer buf;
   outputString(fe.getMessage(), buf);
   if (buf.length() > 0)
      outputLine(buf.append(EOL));

   outputError(fe);
   }

//************************************************************************
// Class OutputErrorToStream`s declaration and method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Filter that outputs the error message associated with any FilterError
// passed to it to the specified output stream. Its input is output
// unchanged.
//------------------------------------------------------------------------
class OutputErrorToStream: public SingleFilter
   {
   public:
      OutputErrorToStream(ostream &outputStream);
      OutputErrorToStream(const OutputErrorToStream &f);

      void processLine(const String &line);
      void handleError(const FilterError &fe);

      DEFINE_FILTER_COPIER(OutputErrorToStream);
   private:
      OutputErrorToStream &operator =(const OutputErrorToStream &f);

      ostream &out;

   }; // class OutputErrorToStream

//------------------------------------------------------------------------
// Constructs an OutputErrorToStream from the ostream to which errors
// are to be output.
//------------------------------------------------------------------------
OutputErrorToStream::OutputErrorToStream(ostream &outputStream):
   SingleFilter(), out(outputStream)
   {
   // empty
   }

//------------------------------------------------------------------------
// Copy constructor.
//------------------------------------------------------------------------
OutputErrorToStream::OutputErrorToStream(const OutputErrorToStream &f):
   SingleFilter(f), out(f.out)
   {
   // empty
   }

//------------------------------------------------------------------------
// Pass through input lines unchanged.
//------------------------------------------------------------------------
INLINE void OutputErrorToStream::processLine(const String &line)
   {
   outputLine(line);
   }

//------------------------------------------------------------------------
// Outputs the error message associated with fe to the stream out, 
// followed by an EOL iff the message doesn`t end in one. Then fe is
// passed to the next filter.
//------------------------------------------------------------------------
void OutputErrorToStream::handleError(const FilterError &fe)
   {
   String msg = fe.getMessage();
   out << msg;
   if (msg.endsWith(EOLString) == false)
      out << '\n';
   outputError(fe);
   }

//************************************************************************
// Class OutputError method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Constructs an OutputError filter that outputs the message associated
// with each FilterError it receives to the next filter as input.
//------------------------------------------------------------------------
OutputError OutputError::toNextFilter()
   {
   return OutputError(OutputErrorToNextFilter(), 0);
   }

//------------------------------------------------------------------------
// Constructs an OutputError filter that outputs the message associated
// with each FilterError it receives to the outputStream.
//------------------------------------------------------------------------
OutputError OutputError::toStream(ostream &outputStream)
   {
   return OutputError(OutputErrorToStream(outputStream), 0);
   }

//------------------------------------------------------------------------
// Copy constructor.
//------------------------------------------------------------------------
OutputError::OutputError(const OutputError &f): ShellFilter(f)
   {
   // empty
   }
