//************************************************************************
// Definitions of methods for all classes declared in prepend.h; and
// declarations and method definitions of all implementation filter
// classes: PrependData, PrependString, PrependFile and PrependStream.
//------------------------------------------------------------------------
// Copyright (C) 1997 by James MacKay
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// The author can be contacted via e-mail at: jmackay@sentex.net
//
//************************************************************************

#include "prepend.h"

#ifndef ENABLE_INLINES
   #pragma hdrstop         // prevent attempts to precompile .inl file
   #include "prepend.inl"
#endif

//************************************************************************
// Abstract class PrependData`s declaration and method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Abstract base class for filters that prepend data to their input.
//------------------------------------------------------------------------
class PrependData: public SingleFilter
   {
   public:
      PrependData();
      PrependData(const PrependData &f);

      void processLine(const String &line);
      void minimize();
   protected:
      void beforeInput();
      void afterInput();

      virtual void outputData(StringBuffer &buf) = 0;
   private:
      PrependData &operator =(const PrependData &f);

      StringBuffer theRest;

   }; // class PrependData

//------------------------------------------------------------------------
// Default constructor.
//------------------------------------------------------------------------
PrependData::PrependData(): SingleFilter(), theRest()
   {
   // empty
   }

//------------------------------------------------------------------------
// Copy constructor.
//------------------------------------------------------------------------
PrependData::PrependData(const PrependData &f):
   SingleFilter(f), theRest(f.theRest)
   {
   // empty
   }

//------------------------------------------------------------------------
// Output the data to be prepended.
//------------------------------------------------------------------------
void PrependData::beforeInput()
   {
   theRest.clear();
   outputData(theRest);
   }

//------------------------------------------------------------------------
// Output line prepended by the contents of theRest.
//------------------------------------------------------------------------
void PrependData::processLine(const String &line)
   {
   require(theRest.toString().indexOf(EOL) < 0);

   outputLine(theRest.append(line));
   theRest.clear();
   }

//------------------------------------------------------------------------
// Output any remaining data, then empty theRest.
//------------------------------------------------------------------------
void PrependData::afterInput()
   {
   require(theRest.toString().indexOf(EOL) < 0);

   if (theRest.length() > 0)
      outputLine(theRest);
   theRest.clear();
   }

//------------------------------------------------------------------------
// Minimize theRest.
//------------------------------------------------------------------------
INLINE void PrependData::minimize()
   {
   SingleFilter::minimize();
   minimizeBuffer(theRest);
   }

//************************************************************************
// Class PrependString`s declaration and method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Filter that prepends the specified string to its input.
//------------------------------------------------------------------------
class PrependString: public PrependData
   {
   public:
      PrependString(const String &str);
      PrependString(const PrependString &f);

      DEFINE_FILTER_COPIER(PrependString);
   protected:
      void outputData(StringBuffer &buf);
   private:
      PrependString &operator =(const PrependString &f);

      const String theString;
 
   }; // class PrependString

//------------------------------------------------------------------------
// Constructs an PrependString filter from the string to prepend.
//------------------------------------------------------------------------
PrependString::PrependString(const String &str): 
   PrependData(), theString(str)
   {
   // empty
   }

//------------------------------------------------------------------------
// Copy constructor.
//------------------------------------------------------------------------
PrependString::PrependString(const PrependString &f):
   PrependData(f), theString(f.theString)
   {
   // empty
   }

//------------------------------------------------------------------------
// Outputs the String to be preended, prepended with the contents of buf,
// and leaving the last partial line (if any) in buf.
//------------------------------------------------------------------------
INLINE void PrependString::outputData(StringBuffer &buf)
   {
   outputString(theString, buf);
   }

//************************************************************************
// Class PrependFile`s declaration and method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Filter that prepends the contents of the specified file to its input.
//------------------------------------------------------------------------
class PrependFile: public PrependData
   {
   public:
      PrependFile(const String &pathname);
      PrependFile(const PrependFile &f);

      DEFINE_FILTER_COPIER(PrependFile);
   protected:
      void outputData(StringBuffer &buf);
   private:
      PrependFile &operator =(const PrependFile &f);

      const String theFile;
 
   }; // class PrependFile

//------------------------------------------------------------------------
// Constructs an PrependFile filter from the pathname of the file to prepend.
//------------------------------------------------------------------------
PrependFile::PrependFile(const String &pathname): 
   PrependData(), theFile(pathname)
   {
   // empty
   }

//------------------------------------------------------------------------
// Copy constructor.
//------------------------------------------------------------------------
PrependFile::PrependFile(const PrependFile &f):
   PrependData(f), theFile(f.theFile)
   {
   // empty
   }

//------------------------------------------------------------------------
// Outputs the contents of the file to be prepended, prepended with the
// contents of buf, and leaving the last partial line (if any) in buf.
//------------------------------------------------------------------------
INLINE void PrependFile::outputData(StringBuffer &buf)
   {
   outputFile(theFile, buf);
   }

//************************************************************************
// Class PrependStream`s declaration and method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Filter that prepends all of the data on the specified input stream to
// its input.
//------------------------------------------------------------------------
class PrependStream: public PrependData
   {
   public:
      PrependStream(istream &in);
      PrependStream(const PrependStream &f);

      DEFINE_FILTER_COPIER(PrependStream);
   protected:
      void outputData(StringBuffer &buf);
   private:
      PrependStream &operator =(const PrependStream &f);

      istream *theStream;
 
   }; // class PrependStream

//------------------------------------------------------------------------
// Constructs an PrependStream filter from the istream whose contents are
// to be prepended.
//------------------------------------------------------------------------
PrependStream::PrependStream(istream &in):
   PrependData(), theStream(&in)
   {
   ensure(theStream != 0);
   }

//------------------------------------------------------------------------
// Copy constructor.
//------------------------------------------------------------------------
PrependStream::PrependStream(const PrependStream &f):
   PrependData(f), theStream(f.theStream)
   {
   // empty
   }

//------------------------------------------------------------------------
// Outputs the contents of the stream to be prepended, prepended with the
// contents of buf, and leaving the last partial line (if any) in buf.
//------------------------------------------------------------------------
INLINE void PrependStream::outputData(StringBuffer &buf)
   {
   outputStream(*theStream, buf);
   }

//************************************************************************
// Class Prepend method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Constructs an Prepend filter from the string to prepend.
//------------------------------------------------------------------------
Prepend Prepend::string(const String &str)
   {
   return Prepend(PrependString(str), 0);
   }

//------------------------------------------------------------------------
// Constructs an Prepend filter from the pathname of the file to prepend.
//------------------------------------------------------------------------
Prepend Prepend::file(const String &pathname)
   {
   return Prepend(PrependFile(pathname), 0);
   }

//------------------------------------------------------------------------
// Constructs an Prepend filter from the stream whose contents are to
// be prepended.
//------------------------------------------------------------------------
Prepend Prepend::stream(istream &in)
   {
   return Prepend(PrependStream(in), 0);
   }

//------------------------------------------------------------------------
// Copy constructor.
//------------------------------------------------------------------------
Prepend::Prepend(const Prepend &f): ShellFilter(f)
   {
   // empty
   }

