//************************************************************************
// Definitions of methods for all classes declared in translat.h, and
// implementation filter declarations and method definitions:
// MappedTranslate, ToUpperCase and ToLowerCase.
//------------------------------------------------------------------------
// Copyright (C) 1997 by James MacKay
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// The author can be contacted via e-mail at: jmackay@sentex.net
//
//************************************************************************

#include "translat.h"

#ifndef ENABLE_INLINES
   #pragma hdrstop         // prevent attempts to precompile .inl file
   #include "translat.inl"
#endif

//************************************************************************
// Class MappedTranslate`s declaration and method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Filter that replaces each occurrence in its input of the n`th
// character in inMap with the n`th character in outMap.
//------------------------------------------------------------------------
class MappedTranslate: public CharacterFilter
   {
   public:
      MappedTranslate(const String &inMap, const String &outMap);
      MappedTranslate(const MappedTranslate &f);

      void processChar(char ch);
 
      DEFINE_FILTER_COPIER(MappedTranslate);
   private:
      MappedTranslate &operator =(const MappedTranslate &f);

      String inputMap;
      String outputMap;

   }; // class MappedTranslate

//------------------------------------------------------------------------
// Constructs a MappedTranslate filter from a list of characters and the
// list of characters to map them to. If the two map strings aren`t the
// same length, the extra characters in the longer map are ignored.
//------------------------------------------------------------------------
MappedTranslate::MappedTranslate(const String &inMap, const String &outMap):
   CharacterFilter()
   {
   StringInt mapLength = String::min(inMap.length(), outMap.length());
   inputMap  = inMap.left(mapLength);
   outputMap = outMap.left(mapLength);

   ensure(inputMap.length() == outputMap.length());
   }

//------------------------------------------------------------------------
// Copy constructor.
//------------------------------------------------------------------------
MappedTranslate::MappedTranslate(const MappedTranslate &f):
   CharacterFilter(f), inputMap(f.inputMap), outputMap(f.outputMap)
   {
   ensure(inputMap.length() == outputMap.length());
   }

//------------------------------------------------------------------------
// Outputs the translated value of ch.
//------------------------------------------------------------------------
INLINE void MappedTranslate::processChar(char ch)
   {
   require(inputMap.length() == outputMap.length());

   StringInt mapIndex = inputMap.indexOf(ch);
   outputChar( (mapIndex < 0) ? ch : outputMap.charAt(mapIndex) );
   }

//************************************************************************
// Class ToUpperCase`s declaration and method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Filter that translates each character in its input to upper case.
//------------------------------------------------------------------------
class ToUpperCase: public SingleFilter
   {
   public:
      ToUpperCase();
      ToUpperCase(const ToUpperCase &f);

      void processLine(const String &line);
 
      DEFINE_FILTER_COPIER(ToUpperCase);
   private:
      ToUpperCase &operator =(const ToUpperCase &f);
   }; // class ToUpperCase

//------------------------------------------------------------------------
// Default constructor.
//------------------------------------------------------------------------
INLINE ToUpperCase::ToUpperCase(): SingleFilter()
   {
   // empty
   }

//------------------------------------------------------------------------
// Copy constructor.
//------------------------------------------------------------------------
ToUpperCase::ToUpperCase(const ToUpperCase &f): SingleFilter(f)
   {
   // empty
   }

//------------------------------------------------------------------------
// Output line converted to upper case.
//------------------------------------------------------------------------
INLINE void ToUpperCase::processLine(const String &line)
   {
   outputLine(line.toUpperCase());
   }

//************************************************************************
// Class ToLowerCase`s declaration and method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Filter that translates each character in its input to lower case.
//------------------------------------------------------------------------
class ToLowerCase: public SingleFilter
   {
   public:
      ToLowerCase();
      ToLowerCase(const ToLowerCase &f);

      void processLine(const String &line);
 
      DEFINE_FILTER_COPIER(ToLowerCase);
   private:
      ToLowerCase &operator =(const ToLowerCase &f);
   }; // class ToLowerCase

//------------------------------------------------------------------------
// Default constructor.
//------------------------------------------------------------------------
INLINE ToLowerCase::ToLowerCase(): SingleFilter()
   {
   // empty
   }

//------------------------------------------------------------------------
// Copy constructor.
//------------------------------------------------------------------------
ToLowerCase::ToLowerCase(const ToLowerCase &f): SingleFilter(f)
   {
   // empty
   }

//------------------------------------------------------------------------
// Output line converted to lower case.
//------------------------------------------------------------------------
INLINE void ToLowerCase::processLine(const String &line)
   {
   outputLine(line.toLowerCase());
   }

//************************************************************************
// Class Translate`s method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Constructs a Translate filter from a list of characters and the list
// of characters to map them to.
//------------------------------------------------------------------------
Translate::Translate(const String &inMap, const String &outMap):
   ShellFilter()
   {
   select(MappedTranslate(inMap, outMap));
   }

//------------------------------------------------------------------------
// Constructs a Translate filter to do the special translation indicated
// by transType.
//------------------------------------------------------------------------
Translate::Translate(TranslationType transType):
   ShellFilter()
   {
   switch (transType)
      {
      case TO_LOWER_CASE:
         select(ToLowerCase());
         break;
      default:
         check(transType == TO_UPPER_CASE);
         select(ToUpperCase());
         break;
      }
   }

//------------------------------------------------------------------------
// Copy constructor.
//------------------------------------------------------------------------
Translate::Translate(const Translate &f): ShellFilter(f)
   {
   // empty
   }
