//************************************************************************
// Definitions of methods for all classes declared in writeto.h, and
// declarations and method definitions of implementation classes:
// WriteToString, WriteToFile and WriteToStream.
//------------------------------------------------------------------------
// Copyright (C) 1997 by James MacKay
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// The author can be contacted via e-mail at: jmackay@sentex.net
//
//************************************************************************

#include "writeto.h"

#ifndef ENABLE_INLINES
   #pragma hdrstop         // prevent attempts to precompile .inl file
   #include "writeto.inl"
#endif

//************************************************************************
// String constants.
//************************************************************************

#define FILE_OPEN_FAIL_MSG    \
   "WriteToFile: couldn`t open (for writing) the file "
#define FILE_WRITE_FAIL_MSG   \
   "WriteToFile: failed attempt to write to file "
#define STREAM_WRITE_FAIL_MSG \
   "WriteToStream: failed attempt to write to stream.\n"

//************************************************************************
// Class WriteToString`s declaration and method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Filter that outputs its input to the specified String (appending to
// or overwriting it as specified by the outputMode), as well as to the
// next filter.
//------------------------------------------------------------------------
class WriteToString: public SingleFilter
   {
   public:
      WriteToString(String &dest, WriteTo::Mode outputMode);
      WriteToString(const WriteToString &f);

      void processLine(const String &line);
      void minimize();

      DEFINE_FILTER_COPIER(WriteToString);
   protected:
      void beforeInput();
      void afterInput();
   private:
      WriteToString &operator =(const WriteToString &f);

      String        *destString;
      StringBuffer  temp;
      WriteTo::Mode outMode;

   }; // class WriteToString

//------------------------------------------------------------------------
// Constructs a WriteToString filter from the String to write to.
//------------------------------------------------------------------------
WriteToString::WriteToString(String &dest, WriteTo::Mode outputMode): 
   SingleFilter(), temp(), outMode(outputMode)
   {
   require(&dest != 0);
   require(outputMode == WriteTo::APPEND || 
           outputMode == WriteTo::OVERWRITE);

   destString = &dest;

   ensure(destString != 0);
   ensure(outMode == WriteTo::APPEND || outMode == WriteTo::OVERWRITE);
   }

//------------------------------------------------------------------------
// Copy constructor.
//------------------------------------------------------------------------
WriteToString::WriteToString(const WriteToString &f):
   SingleFilter(f), destString(f.destString), temp(f.temp), outMode(f.outMode)
   {
   ensure(destString != 0);
   ensure(outMode == WriteTo::APPEND || outMode == WriteTo::OVERWRITE);
   }
   
//------------------------------------------------------------------------
// Sets the initial contents of the temp buffer.
//------------------------------------------------------------------------
void WriteToString::beforeInput()
   {
   require(destString != 0);
   require(outMode == WriteTo::APPEND || outMode == WriteTo::OVERWRITE);

   if (outMode == WriteTo::OVERWRITE)
      temp.clear();
   else
      temp = *destString;
   }

//------------------------------------------------------------------------
// Appends line to the temp buffer and passes it to the next filter.
//------------------------------------------------------------------------
INLINE void WriteToString::processLine(const String &line)
   {
   temp.append(line);
   outputLine(line);
   }

//------------------------------------------------------------------------
// If no errors occurred, write to *destString. Also empties temp.
//------------------------------------------------------------------------
void WriteToString::afterInput()
   {
   *destString = temp;
   temp.clear();
   }

//------------------------------------------------------------------------
// Minimizes the temp buffer.
//------------------------------------------------------------------------
INLINE void WriteToString::minimize()
   {
   SingleFilter::minimize();
   minimizeBuffer(temp);
   }

//************************************************************************
// Class WriteToFile`s declaration and method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Filter that outputs its input to the file named pathname (appending to
// or overwriting it as specified by the outputMode), as well as to the
// next filter.
//------------------------------------------------------------------------
class WriteToFile: public SingleFilter
   {
   public:
      WriteToFile(const String &pathname, WriteTo::Mode outputMode);
      WriteToFile(const WriteToFile &f);

      void processLine(const String &line);

      DEFINE_FILTER_COPIER(WriteToFile);
   protected:
      void beforeInput();
      void afterInput();
   private:
      WriteToFile &operator =(const WriteToFile &f);
      void outputLineToFile(const String &line);

      const String   destFile;
      ofstream       *destOut;
      ios::open_mode outMode;

   }; // class WriteToFile

//------------------------------------------------------------------------
// Constructs a WriteToFile filter from the pathname of the file to 
// write to.
//------------------------------------------------------------------------
WriteToFile::WriteToFile(const String &pathname, WriteTo::Mode outputMode):
   SingleFilter(), destFile(pathname), destOut(0)
   {
   require(outputMode == WriteTo::APPEND || 
           outputMode == WriteTo::OVERWRITE);

   outMode = (outputMode == WriteTo::APPEND) ? ios::app : ios::out;

   ensure(outMode == ios::app || outMode == ios::out);
   }

//------------------------------------------------------------------------
// Copy constructor.
//------------------------------------------------------------------------
WriteToFile::WriteToFile(const WriteToFile &f):
   SingleFilter(), destFile(f.destFile), destOut(f.destOut), 
   outMode(f.outMode)
   {
   ensure(outMode == ios::app || outMode == ios::out);
   }

//------------------------------------------------------------------------
// Attempt to open an ofstream on the file we`re writing to.
//------------------------------------------------------------------------
void WriteToFile::beforeInput()
   {
   require(outMode == ios::app || outMode == ios::out);

   destOut = new ofstream(destFile, outMode);
   if ( !*destOut )
      {
      outputError(String(FILE_OPEN_FAIL_MSG) + destFile + ".\n");
      }
   }

//------------------------------------------------------------------------
// Write the line to the temporary file then pass it to the next filter.
//------------------------------------------------------------------------
INLINE void WriteToFile::processLine(const String &line)
   {
   outputLineToFile(line);
   outputLine(line);
   }

//------------------------------------------------------------------------
// Outputs line to the file we`re writing to.
//------------------------------------------------------------------------
void WriteToFile::outputLineToFile(const String &line)
   {
   if (destOut)
      {
      *destOut << line;
      if ( !destOut->good() )
         {
         outputError(String(FILE_WRITE_FAIL_MSG) + destFile + ".\n");
         destOut->clear();
         }
      }
   }

//------------------------------------------------------------------------
// Close and delete the ofstream on the file we`ve written to.
//------------------------------------------------------------------------
void WriteToFile::afterInput()
   {
   destOut->close();
   delete destOut;
   }

//************************************************************************
// Class WriteToStream`s declaration and method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Filter that outputs its input to the specified output stream, as well
// as to the next filter.
//------------------------------------------------------------------------
class WriteToStream: public SingleFilter
   {
   public:
      WriteToStream(ostream &outStream);
      WriteToStream(const WriteToStream &f);
 
      void processLine(const String &line);

      DEFINE_FILTER_COPIER(WriteToStream);
   private:
      WriteToStream &operator =(const WriteToStream &f);
      void outputLineToStream(const String &line);

      ostream *out;
   };

//------------------------------------------------------------------------
// Constructs a WriteToStream filter from the output stream to write to.
//------------------------------------------------------------------------
WriteToStream::WriteToStream(ostream &outStream): SingleFilter()
   {
   out = &outStream;
   ensure(out != 0);
   }

//------------------------------------------------------------------------
// Copy constructor.
//------------------------------------------------------------------------
WriteToStream::WriteToStream(const WriteToStream &f):
   SingleFilter(f), out(f.out)
   {
   ensure(out != 0);
   }

//------------------------------------------------------------------------
//  Outputs line to *out then passes it on to the next filter.
//------------------------------------------------------------------------
INLINE void WriteToStream::processLine(const String &line)
   {
   outputLineToStream(line);
   outputLine(line);
   }

//------------------------------------------------------------------------
// Outputs line to *out.
//------------------------------------------------------------------------
void WriteToStream::outputLineToStream(const String &line)
   {
   require(out != 0);

   *out << line;
   if ( !out->good() )
      {
      outputError(STREAM_WRITE_FAIL_MSG);
      out->clear();
      }
   }

//************************************************************************
// Class WriteTo`s method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Constructs a WriteTo filter from the String to write to.
//------------------------------------------------------------------------
WriteTo WriteTo::string(String &dest, WriteTo::Mode outputMode)
   {
   return WriteTo(WriteToString(dest, outputMode), 0);
   }

//------------------------------------------------------------------------
// Constructs a WriteTo filter from the pathname of the file to write to.
//------------------------------------------------------------------------
WriteTo WriteTo::file(const String &pathname, WriteTo::Mode outputMode)
   {
   return WriteTo(WriteToFile(pathname, outputMode), 0);
   }

//------------------------------------------------------------------------
// Constructs a WriteTo filter from the output stream to write to.
//------------------------------------------------------------------------
WriteTo WriteTo::stream(ostream &out)
   {
   return WriteTo(WriteToStream(out), 0);
   }

//------------------------------------------------------------------------
// Constructs a WriteTo filter that writes to nothing.
//------------------------------------------------------------------------
WriteTo WriteTo::nothing()
   {
   return WriteTo(PassThrough(), 0);
   }

//------------------------------------------------------------------------
// Copy constructor.
//------------------------------------------------------------------------
WriteTo::WriteTo(const WriteTo &f): ShellFilter(f)
   {
   // empty
   }

