//************************************************************************
// Definitions of class methods for all classes declared in filter.h.
//------------------------------------------------------------------------
// Copyright (C) 1997 by James MacKay
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// The author can be contacted via e-mail at: jmackay@sentex.net
//
//************************************************************************

#include "filter.h"

#ifndef ENABLE_INLINES
   #pragma hdrstop         // prevent attempts to precompile .inl file
   #include "filter.inl"
#endif

//************************************************************************
// String constants (all classes).
//************************************************************************

#define UNINITIALIZED_STRING_MSG \
   "!!! This is an uninitialized String !!!"

#define FILE_NOT_OPENED_MSG   \
   "outputFile(): couldn`t open (for reading) file "
#define IO_ERROR_MSG \
   "I/O error reading from the input stream.\n"
#define NO_LAST_NEWLINE_MSG   \
   "All lines but the last must end in a newline.\n"
#define INTERNAL_NEWLINE_MSG  \
   "A line cannot contain a newline anywhere except as the last " \
   "character in the line.\n"

//************************************************************************
// Class ComponentFilter static data initialization.
//************************************************************************

const char   ComponentFilter::EOL       = '\n';
const String ComponentFilter::EOLString = String::valueOf(ComponentFilter::EOL);
const String ComponentFilter::uninit    = String();

//************************************************************************
// Class ComponentFilter method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Notifies the nextFilter and then this filter that input is about to
// arrive.
//------------------------------------------------------------------------
void ComponentFilter::notifyBeforeInput()
   {  
   if (nextFilter)
      nextFilter->notifyBeforeInput();
   lastLineOutput(false);
   beforeInput();
   }

//------------------------------------------------------------------------
// Notifes this filter and then the nextFilter that the end of the input
// has been reached.
//------------------------------------------------------------------------
void ComponentFilter::notifyAfterInput()
   {
   afterInput();
   if (nextFilter)
      nextFilter->notifyAfterInput();
   }

//------------------------------------------------------------------------
// Puts the next line read from in into buf, returning false if there
// wasn`t a line to read in or an I/O error occurred, and true otherwise. 
// In either case, the previous contents of buf are lost.
//------------------------------------------------------------------------
boolean ComponentFilter::getNextLine(istream &in, StringBuffer &buf)
   {
   char ch;
   buf.clear();
   while (in.get(ch))
      {
      if (in.good())
         {
         buf.append(ch);
         if (ch == EOL)
            break;   // while
         }
      else
         {
         outputError(String("getNextLine(): ") + IO_ERROR_MSG);
         in.clear();
         return false;
         }
      }
   return boolean(buf.length() > 0);
   } 

//------------------------------------------------------------------------
// Minimizes the amount of memory that buf uses.
//------------------------------------------------------------------------
void ComponentFilter::minimizeBuffer(StringBuffer &buf) const
   {
   if (buf.length() == 0)
      buf = StringBuffer();
   }

//------------------------------------------------------------------------
// Outputs str to the next filter, prepending it with what was in buf when
// this method was called, and leaving in buf any partial line not output.
//------------------------------------------------------------------------
void ComponentFilter::outputString(const String &str, StringBuffer &buf)
   {
   require(buf.toString().indexOf(EOL) < 0);

   StringInt len = str.length();
   char ch;
   for (StringInt i = 0; i < len; i++)
      {
      ch = str.charAt(i);
      buf.append(ch);
      if (ch == EOL)
         {
         outputLine(buf);
         buf.clear();
         }
      }

   ensure(buf.toString().indexOf(EOL) < 0);
   }

//------------------------------------------------------------------------
// Outputs the contents of the file named filenamestr to the next filter, 
// prepending it with what was in buf when this method was called, and 
// leaving in buf any partial line not output.
//------------------------------------------------------------------------
void ComponentFilter::outputFile(const String &pathname, StringBuffer &buf)
   {
   ifstream in(pathname);
   if (in)
      outputStream(in, buf);
   else
      outputError(String(FILE_NOT_OPENED_MSG) + pathname + ".\n");
   in.close();
   }

//------------------------------------------------------------------------
// Outputs all data from in to the next filter, prepending it with what
// was in buf when this method was called and leaving in buf any partial
// line not output.
//------------------------------------------------------------------------
void ComponentFilter::outputStream(istream &in, StringBuffer &buf)
   {
   require(buf.toString().indexOf(EOL) < 0);

   char ch;
   while (in.get(ch))
      {
      if (in.good())
         {
         buf.append(ch);
         if (ch == EOL)
            {
            outputLine(buf);
            buf.clear();
            }
         }
      else
         {
         outputError(String("outputStream(): ") + IO_ERROR_MSG);
         in.clear();
         break;
         }
      }

   ensure(buf.toString().indexOf(EOL) < 0);
   }  // ComponentFilter::outputStream(istream &, StringBuffer &)

#if CHECK_ASSERTIONS

//------------------------------------------------------------------------
// Returns true iff outputLine is a valid line to output.
//------------------------------------------------------------------------
boolean ComponentFilter::isValidOutputLine(const String &outputLine)
   {
   StringInt EOLIndex = outputLine.indexOf(EOL);
   boolean isValid = true;

   if (outputLastLine)
      {
      debugOutput("outputLine(): "NO_LAST_NEWLINE_MSG);

      isValid = false;
      }
   else if (EOLIndex < 0)
      {
      outputLastLine = true;
      isValid = true;
      }
   else if (EOLIndex < outputLine.length() - 1)
      {
      debugOutput("outputLine(): "INTERNAL_NEWLINE_MSG);
      isValid = false;
      }
   else
      isValid = true;
 
   return isValid;
   }

#endif  // #if CHECK_ASSERTIONS

//************************************************************************
// Class InputFilter`s declaration and method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Class InputFilter.
//------------------------------------------------------------------------
class InputFilter: public ComponentFilter
   {
   public:
      InputFilter(istream &inStream);

      void processLine(const String &line);

      DEFINE_FILTER_COPIER(InputFilter);
   protected:
      void beforeInput();
   private:
      InputFilter &operator =(const InputFilter &f);

      istream *in;

   }; // class InputFilter

//------------------------------------------------------------------------
// Constructs an InputFilter from an istream.
//------------------------------------------------------------------------
InputFilter::InputFilter(istream &inStream): 
   ComponentFilter(), in(&inStream)
   {
   // empty
   }

//------------------------------------------------------------------------
// This method shouldn`t be called at all.
//------------------------------------------------------------------------
void InputFilter::processLine(const String &)
   {
   debugOutput("InputFilter::processLine() called.\n");
   }

//------------------------------------------------------------------------
// Passes each line read from in to the next filter.
//------------------------------------------------------------------------
void InputFilter::beforeInput()
   {
   StringBuffer buf;

   while (getNextLine(*in, buf))
      outputLine(buf);
   }

//************************************************************************
// Class OutputFilter`s declaration and method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Class OutputFilter.
//------------------------------------------------------------------------
class OutputFilter: public ComponentFilter
   {
   public:
      OutputFilter(ostream &outStream);

      void processLine(const String &line);

      DEFINE_FILTER_COPIER(OutputFilter);
   private:
      OutputFilter &operator =(const OutputFilter &f);

      ostream *out;

   }; // class OutputFilter

//------------------------------------------------------------------------
// Constructs an OutputFilter from an ostream.
//------------------------------------------------------------------------
OutputFilter::OutputFilter(ostream &outStream):
   ComponentFilter(), out(&outStream)
   {
   // empty
   }

//------------------------------------------------------------------------
// Outputs line to *out.
//------------------------------------------------------------------------
INLINE void OutputFilter::processLine(const String &line)
   {
   *out << line;
   }

//************************************************************************
// Class SingleFilter method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Gets data from the input filter, filters it, then passes it to the
// output filter.
//------------------------------------------------------------------------
void SingleFilter::connect(ComponentFilter &input, ComponentFilter &output)
   {
   input.linkToNextFilter(*this);
   this->linkToNextFilter(output);

   input.notifyBeforeInput();
   input.notifyAfterInput();

   this->detachFromNextFilter();
   input.detachFromNextFilter();
   }

//------------------------------------------------------------------------
// Reads data from the in stream, filters it, then outputs it to out.
//------------------------------------------------------------------------
void SingleFilter::connect(istream &in, ostream &out)
   {
   InputFilter  input(in);
   OutputFilter output(out);
   connect(input, output);
   }

//------------------------------------------------------------------------
// Gets data from the input filter, filters it, then outputs it to out.
//------------------------------------------------------------------------
void SingleFilter::connect(ComponentFilter &input, ostream &out)
   {
   OutputFilter output(out);
   connect(input, output);
   }

//------------------------------------------------------------------------
// Reads data from the in stream, filters it, then passes it to the
// output filter.
//------------------------------------------------------------------------
void SingleFilter::connect(istream &in, ComponentFilter &output)
   {
   InputFilter input(in);
   connect(input, output);
   }

//************************************************************************
// Class Filter method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Constructs a Filter from a ComponentFilter.
//------------------------------------------------------------------------
Filter::Filter(const ComponentFilter &f): SingleFilter()
   {
   constructFilter(f);
   }

//------------------------------------------------------------------------
// Copy constructor. (Identical to Filter(const ComponentFilter &).)
//------------------------------------------------------------------------
Filter::Filter(const Filter &f): SingleFilter()
   {
   constructFilter(f);
   }

//------------------------------------------------------------------------
// Constructs a Filter from a pair of ComponentFilters.
//------------------------------------------------------------------------
Filter::Filter(const ComponentFilter &first,
               const ComponentFilter &second): SingleFilter()
   {
   require(first.componentCount() + second.componentCount() > 0);

   numComponents = first.componentCount() + second.componentCount();
   component = new ComponentFilter *[numComponents];
   copyComponents(first, component);
   copyComponents(second, component + first.componentCount());
   linkComponents();

   ensure(component != 0);
   ensure(componentCount() > 0);
   ensure(componentCount() == first.componentCount() +
                              second.componentCount());
   }

//------------------------------------------------------------------------
// Assignment operator.
//------------------------------------------------------------------------
Filter &Filter::operator =(const ComponentFilter &f)
   {
   require(f.componentCount() > 0);

   if (this != &f)
      {
      ComponentFilter **temp = new ComponentFilter *[f.componentCount()];
      copyComponents(f, temp);
      deleteComponents();
      component = temp;
      numComponents = f.componentCount();
      linkComponents();
      }

   ensure(component != 0);
   ensure(componentCount() > 0);
   ensure(componentCount() == f.componentCount());
   return *this;
   }

//------------------------------------------------------------------------
// Minimizes the amount of resources that this Filter and each of its
// component filter`s uses.
//------------------------------------------------------------------------
void Filter::minimize()
   {
   SingleFilter::minimize();
   for (FilterInt i = 0; i < numComponents; i++)
      component[i]->minimize();
   }

//------------------------------------------------------------------------
// Makes this Filter a copy of c. This utility method should only be
// called from a constructor.
//------------------------------------------------------------------------
void Filter::constructFilter(const ComponentFilter &f)
   {
   require(f.componentCount() > 0);

   numComponents = f.componentCount();
   component = new ComponentFilter *[numComponents];
   copyComponents(f, component);
   linkComponents();
 
   ensure(component != 0);
   ensure(componentCount() > 0);
   ensure(componentCount() == f.componentCount());
   }

//------------------------------------------------------------------------
// Copies all of the components of f into compArray. It is assumed that
// compArray is large enough to hold all of f`s components.
//------------------------------------------------------------------------
void Filter::copyComponents(const ComponentFilter &f,
                            ComponentFilter **compArray)
   {
   require(compArray != 0);

   for (FilterInt i = f.componentCount() - 1; i >= 0; i--)
      {
      check(f.getComponent(i) != 0);
      compArray[i] = f.getComponent(i)->getCopy();
      check(compArray[i] != 0);
      }
   }

//------------------------------------------------------------------------
// Links together all of the filters in the component array.
//------------------------------------------------------------------------
void Filter::linkComponents()
   {
   for (FilterInt i = 1; i < numComponents; i++)
      component[i-1]->linkToNextFilter(*component[i]);
   }
   
//------------------------------------------------------------------------
// Deletes the filter pointed to by each element of the component array,
// then deletes the array itself.
//------------------------------------------------------------------------
void Filter::deleteComponents()
   {
   for (FilterInt i = 0; i < numComponents; i++)
      delete component[i];
   delete[] component;
   component = 0;
   }

//************************************************************************
// Class ShellFilter method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Default constructor.
//------------------------------------------------------------------------
ShellFilter::ShellFilter(): SingleFilter(), impl(0)
   {
   // empty
   }

//------------------------------------------------------------------------
// Copy constructor.
//------------------------------------------------------------------------
ShellFilter::ShellFilter(const ShellFilter &f): SingleFilter(f), impl(0)
   {
   if (f.impl)
      select(*f.impl);
   }

//------------------------------------------------------------------------
// Minimize the resources used by the ShellFilter and the implementation
// filter.
//------------------------------------------------------------------------
void ShellFilter::minimize()
   {
   require(impl);

   SingleFilter::minimize();
   impl->minimize();
   }

//------------------------------------------------------------------------
// Links the ShellFilter and the implementation filter to f.
//------------------------------------------------------------------------
void ShellFilter::linkToNextFilter(ComponentFilter &f)
   {
   require(impl);

   SingleFilter::linkToNextFilter(f);
   impl->linkToNextFilter(f);
   }

//------------------------------------------------------------------------
// Detaches the ShellFilter and the implementation filter from the next 
// filter.
//------------------------------------------------------------------------
void ShellFilter::detachFromNextFilter()
   {
   require(impl);

   impl->detachFromNextFilter();
   SingleFilter::detachFromNextFilter();
   }
