//************************************************************************
// Declaration of the filter classes (ComponentFilter, SingleFilter,
// Filter and ShellFilter), as well as the FilterError and
// FilterException classes and some of their descendants.
//------------------------------------------------------------------------
// Copyright (C) 1997 by James MacKay
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// The author can be contacted via e-mail at: jmackay@sentex.net
//
//************************************************************************
#ifndef FILTER_H_
#define FILTER_H_

#include <iostream.h>
#include <fstream.h>
#include "jstring.h"
#include "release.h"

#define DEFINE_FILTER_COPIER(filterClassName)               \
   ComponentFilter *getCopy() const                         \
      {                                                     \
      return new filterClassName(*this);                    \
      }

typedef int FilterInt;

//************************************************************************
// Class declarations.
//************************************************************************

class FilterError;

//------------------------------------------------------------------------
// Class ComponentFilter.
//------------------------------------------------------------------------
class ComponentFilter
   {
   public:
      static const char EOL;           // = the end-of-line character
      static const String EOLString;   // = the EOL char as a String
      static const String uninit;

      enum CaseUsage { IGNORE_CASE, CASE_SENSITIVE };

      ComponentFilter();
      ComponentFilter(const ComponentFilter &f);
      virtual ~ComponentFilter();

      virtual void processLine(const String &line) = 0;
      virtual void handleError(const FilterError &fe);
      virtual void minimize();
   protected:
      virtual void beforeInput();
      virtual void afterInput();
   public:
      virtual ComponentFilter *getCopy() const = 0;

      virtual void linkToNextFilter(ComponentFilter &f);
      virtual void detachFromNextFilter();

      virtual FilterInt componentCount() const;
      virtual const ComponentFilter *getComponent(FilterInt compIndex) const;

      virtual void notifyBeforeInput();
      virtual void notifyAfterInput();
   protected:      
      boolean getNextLine(istream &in, StringBuffer &buf);
      void minimizeBuffer(StringBuffer &buf) const;

      void outputLine(const String &str);
      void outputString(const String &str, StringBuffer &buf);
      void outputFile(const String &filename, StringBuffer &buf);
      void outputStream(istream &in, StringBuffer &buf);

      void outputError(const String &msg) const;
      void outputError(const FilterError &fe) const;
   protected:
      void lastLineOutput(boolean wasOutput);
   private:
      ComponentFilter &operator =(const ComponentFilter &f);

      ComponentFilter *nextFilter;

      #if CHECK_ASSERTIONS
         boolean isValidOutputLine(const String &outputLine);
         boolean outputLastLine;
      #endif

   }; // class ComponentFilter

//------------------------------------------------------------------------
// Class SingleFilter.
//------------------------------------------------------------------------
class SingleFilter: public ComponentFilter
   {
   public:
      SingleFilter();
      SingleFilter(const ComponentFilter &f);
      SingleFilter(const SingleFilter &f);

      void connect(istream &in, ostream &out);
      void connect(ComponentFilter &in, ostream &out);
      void connect(istream &in, ComponentFilter &out);
      void connect(ComponentFilter &in, ComponentFilter &out); 
   private:
      SingleFilter &operator =(const SingleFilter &f);
   }; // class SingleFilter

//------------------------------------------------------------------------
// Class Filter.
//------------------------------------------------------------------------
class Filter: public SingleFilter
   {
   public:
      Filter(const ComponentFilter &f);
      Filter(const ComponentFilter &first,
             const ComponentFilter &second);
      Filter(const Filter &f);
      ~Filter();
      Filter &operator =(const ComponentFilter &f);

      friend Filter operator |(const ComponentFilter &left,
                               const ComponentFilter &right);

      void processLine(const String &line);
      void handleError(const FilterError &fe);
      void minimize();

      void linkToNextFilter(ComponentFilter &f);
      void detachFromNextFilter();

      FilterInt componentCount() const;
      const ComponentFilter *getComponent(FilterInt compIndex) const;

      void notifyBeforeInput();
      void notifyAfterInput();

      DEFINE_FILTER_COPIER(Filter);
   private:
      void constructFilter(const ComponentFilter &f);
      void copyComponents(const ComponentFilter &f, 
                          ComponentFilter **compArray);
      void linkComponents();
      void deleteComponents();

      ComponentFilter **component;
      FilterInt numComponents;

   }; // class Filter

//------------------------------------------------------------------------
// Class ShellFilter.
//------------------------------------------------------------------------
class ShellFilter: public SingleFilter
   {
   public:
      ShellFilter();
      ShellFilter(const ShellFilter &f);
      virtual ~ShellFilter();

      void processLine(const String &line);
      void handleError(const FilterError &fe);
      void minimize();
 
      void linkToNextFilter(ComponentFilter &f);
      void detachFromNextFilter();

      void notifyBeforeInput();
      void notifyAfterInput();

      DEFINE_FILTER_COPIER(ShellFilter);
   protected:
      void select(const ComponentFilter &f);
      ComponentFilter *getImplementation() const;
   private:
      ShellFilter &operator =(const ShellFilter &f);

      ComponentFilter *impl;

   }; // class ShellFilter

//------------------------------------------------------------------------
// Class FilterError.
//------------------------------------------------------------------------
class FilterError
   {
   public:
      FilterError(const String &errorMessage);
      String getMessage() const;
   private:
      String msg;
   }; // class FilterError

//------------------------------------------------------------------------
// Class FilterException.
//------------------------------------------------------------------------
class FilterException
   {
   public:
      FilterException(const String &errorMessage);
      String getMessage() const;
   private:
      String msg;
   }; // class FilterException

//------------------------------------------------------------------------
// Class InvalidFilterArgumentException.
//------------------------------------------------------------------------
class InvalidFilterArgumentException: public FilterException
   {
   public:
      InvalidFilterArgumentException(const String &errorMessage);
   }; 


#ifdef ENABLE_INLINES
   #include "filter.inl"
#endif

#endif   // #ifndef FILTER_H_

