//************************************************************************
// Inline method definitions for all classes declared in filter.h.
//------------------------------------------------------------------------
// Copyright (C) 1997 by James MacKay
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// The author can be contacted via e-mail at: jmackay@sentex.net
//
//************************************************************************

//************************************************************************
// FilterError`s inline method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Constructs a FilterError from an error message.
//------------------------------------------------------------------------
INLINE FilterError::FilterError(const String &errorMessage):
   msg(errorMessage)
   {
   // empty
   }

//------------------------------------------------------------------------
// Return the error message.
//------------------------------------------------------------------------
INLINE String FilterError::getMessage() const
   {
   return msg;
   }

//************************************************************************
// ComponentFilter`s inline method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Outputs line to the next filter.
//------------------------------------------------------------------------
INLINE void ComponentFilter::outputLine(const String &line)
   {
   require(isValidOutputLine(line));

   if (nextFilter)
      nextFilter->processLine(line);
   }

//------------------------------------------------------------------------
// Signals the next filter that there has been an error in this or a
// previous filter.
//------------------------------------------------------------------------
INLINE void ComponentFilter::outputError(const FilterError &fe) const
   {
   if (nextFilter)
      nextFilter->handleError(fe);
   }

//------------------------------------------------------------------------
// Signals the next filter that there has been an error in this or a
// previous filter.
//------------------------------------------------------------------------
INLINE void ComponentFilter::outputError(const String &msg) const
   {
   outputError(FilterError(msg));
   }

//------------------------------------------------------------------------
// Sets the indicator indicating whether the last line has been output.
//------------------------------------------------------------------------
INLINE void ComponentFilter::lastLineOutput(boolean wasOutput)
   {
   #if CHECK_ASSERTIONS   
      outputLastLine = wasOutput;
   #endif
   }

//------------------------------------------------------------------------
// Default constructor.
//------------------------------------------------------------------------
INLINE ComponentFilter::ComponentFilter(): nextFilter(0)
   {
   lastLineOutput(false);
   }

//------------------------------------------------------------------------
// Copy constructor.
//------------------------------------------------------------------------
INLINE ComponentFilter::ComponentFilter(const ComponentFilter &f):
   nextFilter(0)
   {
   #if CHECK_ASSERTIONS
      lastLineOutput(f.outputLastLine);
   #endif
   }

//------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------
INLINE ComponentFilter::~ComponentFilter()
   {
   // empty
   }

//------------------------------------------------------------------------
// This method is called before input is sent to the filter.
//------------------------------------------------------------------------
INLINE void ComponentFilter::beforeInput()
   {
   // empty
   }

//------------------------------------------------------------------------
// This method is called after all input has been passed to the filter.
//------------------------------------------------------------------------
INLINE void ComponentFilter::afterInput()
   {
   // empty
   }

//------------------------------------------------------------------------
// This method is called if an error occurred in a previous filter.
//------------------------------------------------------------------------
INLINE void ComponentFilter::handleError(const FilterError &fe)
   {
   outputError(fe);
   }

//------------------------------------------------------------------------
// Minimizes the amount of resources the ComponentFilter uses.
//------------------------------------------------------------------------
INLINE void ComponentFilter::minimize()
   {
   // empty
   }
 
//------------------------------------------------------------------------
// Links f as the next filter after this one.
//------------------------------------------------------------------------
INLINE void ComponentFilter::linkToNextFilter(ComponentFilter &f)
   {
   nextFilter = &f;
   }

//------------------------------------------------------------------------
// Detaches from the next filter.
//------------------------------------------------------------------------
INLINE void ComponentFilter::detachFromNextFilter()
   {
   nextFilter = 0;
   }

//------------------------------------------------------------------------
// Returns the number of component filters this filter is composed of.
//------------------------------------------------------------------------
INLINE FilterInt ComponentFilter::componentCount() const
   {
   return 1;
   }

//------------------------------------------------------------------------
// Returns a pointer to the (compIndex+1)th component of this filter
// (which is the filter itself for ComponentFilters).
//------------------------------------------------------------------------
INLINE const ComponentFilter *ComponentFilter::
   getComponent(FilterInt compIndex) const
   {
   require(compIndex == 0);
   return this;
   }

//************************************************************************
// FilterException`s inline method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Constructs a FilterException from an error message.
//------------------------------------------------------------------------
INLINE FilterException::FilterException(const String &errorMessage):
   msg(errorMessage)
   {
   // empty
   }

//------------------------------------------------------------------------
// Return the error message.
//------------------------------------------------------------------------
INLINE String FilterException::getMessage() const
   {
   return msg;
   }

//************************************************************************
// InvalidFilterArgumentException`s inline method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Constructs a InvalidFilterArgumentException from an error message.
//------------------------------------------------------------------------
INLINE InvalidFilterArgumentException::
   InvalidFilterArgumentException(const String &errorMessage):
   FilterException(errorMessage)
   {
   // empty
   }

//************************************************************************
// SingleFilter`s inline method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Default constructor.
//------------------------------------------------------------------------
INLINE SingleFilter::SingleFilter(): ComponentFilter()
   {
   // empty
   }

//------------------------------------------------------------------------
// Constructs a SingleFilter from a ComponentFilter.
//------------------------------------------------------------------------
INLINE SingleFilter::SingleFilter(const ComponentFilter &f):
   ComponentFilter(f)
   {
   // empty
   }

//------------------------------------------------------------------------
// Copy constructor. Should be identical to SingleFilter::
// SingleFilter(const ComponentFilter &).
//------------------------------------------------------------------------
INLINE SingleFilter::SingleFilter(const SingleFilter &f):
   ComponentFilter(f)
   {
   // empty
   }

//************************************************************************
// Filter`s inline method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------
INLINE Filter::~Filter()
   {
   deleteComponents();
   }

//------------------------------------------------------------------------
// Filter concatenation operator.
//------------------------------------------------------------------------
INLINE Filter operator |(const ComponentFilter &left,
                         const ComponentFilter &right)
   {
   return Filter(left, right);
   }

//------------------------------------------------------------------------
// Passes line on to the first component for processing.
//------------------------------------------------------------------------
INLINE void Filter::processLine(const String &line)
   {
   require(componentCount() > 0);
   require(component[0] != 0);

   component[0]->processLine(line);
   }

//------------------------------------------------------------------------
// Calls the handleError() method of the first component.
//------------------------------------------------------------------------
INLINE void Filter::handleError(const FilterError &fe)
   {
   require(componentCount() > 0);
   require(component[0] != 0);

   component[0]->handleError(fe);
   }

//------------------------------------------------------------------------
// Links the last component filter to f.
//------------------------------------------------------------------------
INLINE void Filter::linkToNextFilter(ComponentFilter &f)
   {
   require(componentCount() > 0);
   require(component[componentCount() - 1] != 0);

   component[componentCount() - 1]->linkToNextFilter(f);
   }

//------------------------------------------------------------------------
// Detaches the last component from the filter to which it is linked.
//------------------------------------------------------------------------
INLINE void Filter::detachFromNextFilter()
   {
   require(componentCount() > 0);
   require(component[componentCount() - 1] != 0);

   component[componentCount() - 1]->detachFromNextFilter();
   }

//------------------------------------------------------------------------
// Returns the number of component filters that this filter is composed of.
//------------------------------------------------------------------------
INLINE FilterInt Filter::componentCount() const
   {
   return numComponents;
   }

//------------------------------------------------------------------------
// Returns a pointer to the (compIndex+1)th component of this filter.
//------------------------------------------------------------------------
INLINE const ComponentFilter *Filter::getComponent(FilterInt compIndex) const
   {
   require(compIndex < componentCount());
   require(component[compIndex] != 0);

   return component[compIndex];
   }

//------------------------------------------------------------------------
// Pass the notification on to the first component.
//------------------------------------------------------------------------
INLINE void Filter::notifyBeforeInput()
   {
   require(componentCount() > 0);
   require(component[0] != 0);

   component[0]->notifyBeforeInput();
   }

//------------------------------------------------------------------------
// Pass the notification on to the first component.
//------------------------------------------------------------------------
INLINE void Filter::notifyAfterInput()
   {
   require(componentCount() > 0);
   require(component[0] != 0);

   component[0]->notifyAfterInput();
   }

//************************************************************************
// ShellFilter`s inline method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------
INLINE ShellFilter::~ShellFilter()
   {
   delete impl;
   }

//------------------------------------------------------------------------
// Calls the implementation filter`s processLine() method.
//------------------------------------------------------------------------
INLINE void ShellFilter::processLine(const String &line)
   {
   require(impl);
   impl->processLine(line);
   }

//------------------------------------------------------------------------
// Calls the implementation filter`s handleError() method.
//------------------------------------------------------------------------
INLINE void ShellFilter::handleError(const FilterError &fe)
   {
   require(impl);
   impl->handleError(fe);
   }

//------------------------------------------------------------------------
// Notify the implementation filter that input is about to start, then let
// this filter (i.e. the ShellFilter) do its pre-input processing.
//------------------------------------------------------------------------
INLINE void ShellFilter::notifyBeforeInput()
   {
   require(impl);

   impl->notifyBeforeInput();
   lastLineOutput(false);
   beforeInput();   
   }

//------------------------------------------------------------------------
// Let this filter (i.e. the ShellFilter) do its post-input processing,
// then notify the implementation filter that there will be no more input.
//------------------------------------------------------------------------
INLINE void ShellFilter::notifyAfterInput()
   {
   require(impl);

   afterInput();
   impl->notifyAfterInput();
   }

//------------------------------------------------------------------------
// Returns a pointer to the implementation filter, or 0 if no 
// implementation filter has been selected.
//------------------------------------------------------------------------
INLINE ComponentFilter *ShellFilter::getImplementation() const
   {
   return impl;
   }

//------------------------------------------------------------------------
// Selects (a copy of) f as the implementation filter. Assumes that an
// implementation filter hasn`t already been selected.
//------------------------------------------------------------------------
INLINE void ShellFilter::select(const ComponentFilter &f)
   {
   require(impl == 0);  

   impl = f.getCopy();

   ensure(impl != 0);
   }
