//************************************************************************
// Tests filter classes.
//------------------------------------------------------------------------
// Copyright (C) 1997 by James MacKay
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// The author can be contacted via e-mail at: jmackay@sentex.net
//
//************************************************************************

#include <fstream.h>
#include <strstream.h>
#include <new.h>

#include "filter.h"
#include "take.h"
#include "drop.h"
#include "append.h"
#include "writeto.h"
#include "generror.h"
#include "outerror.h"
#include "include.h"

#include "jstring.h"
#include "release.h"

void test1();
void test2();
void test3();
void test4();
void test5();
void test6();
void test7();

const char outputFile[] = "output.tst";
ofstream outStream(outputFile);

//------------------------------------------------------------------------
// Function called when new fails.
//------------------------------------------------------------------------
void onNewFailure()
   {
	outStream << "Out of memory - exiting." << endl;
   exit(-1);
	}

//------------------------------------------------------------------------
// main().
//------------------------------------------------------------------------
int main()
   {
   set_new_handler(&onNewFailure);

   test1();
   test2();
   test3();
   test4();
   test5();
   test6();
   test7();

   return 0;
   }

//------------------------------------------------------------------------
// #1. Basic ComponentFilter, SingleFilter and ShellFilter class tests.
//------------------------------------------------------------------------
void test1()
   {
   String input1 = "Line 1\n\nline 3\nline 4\nline 5\nline 6";
   String input2 = "\nline 2\nline 3\nline 4\n";
   String output;

   Append  in  = Append::string(input1);
   WriteTo out = WriteTo::string(output, WriteTo::OVERWRITE);
   Take take3(3);

   outStream << "#1" << endl;

   take3.connect(in, out);
   outStream << output << endl;  // first 3 lines of input1
   
   out.minimize();
   Take(6).connect(in, out);
   outStream << output << endl;  // all of input1 (not newline-terminated)

   Append in2 = Append::string(input2);
   take3.connect(in2, out);
   outStream << output << endl;  // first 3 lines of input2
 
   Take(4).connect(in2, out);
   outStream << output << endl;  // all of input2 (incl. terminating newline)
   }

//------------------------------------------------------------------------
// #2. Basic Filter class tests.
//------------------------------------------------------------------------
void test2()
   {
   String input1 = "Line 1\n\nline 3\nline 4\nline 5\nline 6";
   String output;

   Append  in  = Append::string(input1);
   WriteTo out = WriteTo::string(output, WriteTo::OVERWRITE);
   Drop drop2(2);
   Filter  extract = Take(5) | drop2;     // extracts lines 3 - 5

   outStream << "#2" << endl;

   extract.connect(in, out);
   outStream << output << endl;           // lines 3-5 of input1

   (Drop(3) | Take(3)).connect(in, out);
   outStream << output << endl;           // lines 4-6 of input1
                                          // (not newline-terminated)

   extract = Drop(0) | Take(0);           // outputs a newline.     
   extract.minimize();
   extract.connect(in, out);
   outStream << output << endl;
   }

//------------------------------------------------------------------------
// #3. Tests handling of FilterErrors.
//------------------------------------------------------------------------
void test3()
   {
   String input1 = "Line 1\n\nline 3\nline 4\nline 5\nline 6";
   String output;

   Append  in  = Append::string(input1);
   WriteTo out = WriteTo::string(output, WriteTo::OVERWRITE);
   Drop dropOne(1);
   Filter extract = Take(5) | dropOne;     // extracts lines 2 - 5

   String errMsg = "An error occurred in beforeInput()";
   GenerateError error = GenerateError::before(errMsg);

   outStream << "#3" << endl;
   Filter f = error | dropOne | extract;  // output lines 3 - 6
   f.connect(in, out);
   outStream << output << endl;

   f = f | OutputError::toNextFilter();
   Filter f2 = f;
   f2.connect(in, out);                   // output errMsg then lines 3-6
   outStream << output << endl;           // + newline (from OutputError)
                                          // + newline (from endl)
   }

//------------------------------------------------------------------------
// #4. Tests FilterException and InvalidFilterArgumentException.
//------------------------------------------------------------------------
// Note: begin and end are defined outside of Include::tagged()`s
//       parameter list because otherwise gcc (2.7.2) seg faults. 
//       Is this a gcc bug?
//------------------------------------------------------------------------
void test4()
   {
   outStream << "#4" << endl;
   try
      {
      // tags cannot be empty
      String begin;
      String end = "end";
      Include inc = Include::tagged(begin, end);
      }
   catch (InvalidFilterArgumentException e)
      {
      e.getMessage();
      outStream << "InvalidFilterArgumentException occurred." << endl;
      }
   }

//------------------------------------------------------------------------
// #5. Tests connect() with one or two stream arguments, and 
// outputStream().
//------------------------------------------------------------------------
void test5()
   {
   String input1 = "Line 1\n\nline 3\nline 4\nline 5\nline 6";
   String output;
   char *ch1 = input1.toCharArray();

   outStream << "#5" << endl;

   // tests connect(istream, ostream):
   istrstream source1(ch1);
   Drop(1).connect(source1, outStream);   // lines 2-6
   outStream << endl;                     // + newline

   // tests connect(istream, ComponentFilter)
   istrstream source2(ch1);
   WriteTo out = WriteTo::string(output, WriteTo::OVERWRITE);
   Take(3).connect(source2, out);         // lines 1-3
   outStream << output << endl;           // + newline

   // tests connect(ComponentFilter, ostream)
   istrstream source3(ch1);
   Filter in  = Append::stream(source3);
   Filter f = Drop(2) | Take(2);
   f.connect(in, outStream);              // lines 3-4
   outStream << endl;                     // + newline

   delete ch1;
   }

//------------------------------------------------------------------------
// #6. Tests SingleFilter(const ComponentFilter &f) constructor when
//     f`s class is derived from SingleFilter.  
//------------------------------------------------------------------------
void test6()
   {
   String input1 = "Line 1\n\nline 3\nline 4\nline 5\nline 6";
   String output;
   
   Append  in  = Append::string(input1);
   WriteTo out = WriteTo::string(output, WriteTo::OVERWRITE);

   Take temp(3);
   Take take3 = temp;

   outStream << "#6" << endl;

   take3.connect(in, out);
   outStream << output << endl;     // lines 1-3
   }

//------------------------------------------------------------------------
// #7. Tests
//------------------------------------------------------------------------
void test7()
   {
   outStream << "#7" << endl;

   // write this !!!!
   }

