//************************************************************************
// Definitions of class methods for all classes declared in jstring.h,
// and initialization of static data.
//------------------------------------------------------------------------
// Copyright (C) 1997 by James MacKay
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// The author can be contacted via e-mail at: jmackay@sentex.net
//
//************************************************************************

#include "jstring.h"

#ifndef ENABLE_INLINES
   #pragma hdrstop         // prevent attempts to precompile .inl file
   #include "jstring.inl"
#endif

//************************************************************************
// Static data initialization.
//************************************************************************

const char      StringData::terminator          = '\0';
const StringInt StringBuffer::capacityBlockSize = 32;
const StringInt StringBuffer::minimumCapacity   = capacityBlockSize - 1;
const char      StringBuffer::defaultFillChar   = ' ';

//************************************************************************
// Class String method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Constructs a String containing numCopies copies of the fillChar.
//------------------------------------------------------------------------
String::String(char fillChar, StringInt numCopies)
   {
   require(numCopies >= 0);

   if (fillChar == terminator() || numCopies == 0)
      data = getStringData();
   else
      {
      data = getStringData(numCopies);
      memset(data->contents, fillChar, numCopies);
      }

   ensure(isValid());
   }

//------------------------------------------------------------------------
// Assignment operator.
//------------------------------------------------------------------------
String &String::operator =(const String &c)
   {
   require(c.isValid());

   if (this->data != c.data)
      {
      deleteStringData();
      data = c.data;
      data->refCount++;
      }

   ensure(isValid());
   return *this;
   }

//------------------------------------------------------------------------
// Returns a String that has str as its internal representation. Deleting
// this String will also delete str.
//------------------------------------------------------------------------
// Note that subsequent alteration or deletion of str could invalidate the
// String`s state, and so this method should be used carefully.
//------------------------------------------------------------------------
String String::valueOf(char *str)
   {
   require (str != 0);

   StringData *newData = new StringData;
   newData->refCount = 1;
   newData->contents = str;
   newData->length = newData->capacity = strlen(str);

   ensure(newData->isValid());
   return String(newData);
   }

//------------------------------------------------------------------------
// Returns the value of ch represented as a String.
//------------------------------------------------------------------------
String String::valueOf(char ch)
   {
   char temp[2] = {ch, terminator()};
   return String(temp);
   }

//------------------------------------------------------------------------
// Returns true iff *this is equal to the leading characters of fullStr
// and its length is at least minLength.
//------------------------------------------------------------------------
boolean String::isAbbreviationOf(const String &fullStr,
                                 StringInt minLength) const
   {
   require(isValid());
   require(fullStr.isValid());
   require(minLength >= 0);

   boolean isAbbrev = false;
   if (length() >= minLength)
      isAbbrev = (fullStr.indexOf(*this) == 0) ? true : false;
   return isAbbrev;
   }

//------------------------------------------------------------------------
// Returns the index of the first occurrence of ch at or after index
// fromIndex, or -1 if ch doesn`t occur at or after fromIndex.
//------------------------------------------------------------------------
StringInt String::indexOf(int ch, StringInt fromIndex) const
   {
   require(isValid());
   require(fromIndex >= 0);
   require(fromIndex <= length());

   if (ch == terminator() || fromIndex == length())
      return -1;
   char *matchPtr = strchr(data->contents + fromIndex, ch);
   return (matchPtr == 0) ? -1 : StringInt(matchPtr - data->contents);
   }

//------------------------------------------------------------------------
// Returns the index of the first occurrence of str at or after index
// fromIndex, or -1 if str doesn`t occur at or after fromIndex.
//------------------------------------------------------------------------
StringInt String::indexOf(const String &str, StringInt fromIndex) const
   {
   require(isValid());
   require(str.isValid());
   require(fromIndex >= 0);
   require(fromIndex <= length());

   if (fromIndex == length())
      return -1;
   char *matchPtr = strstr(data->contents + fromIndex, str.data->contents);
   return (matchPtr == 0) ? -1 : StringInt(matchPtr - data->contents);
   }

//------------------------------------------------------------------------
// Returns the index of the last occurrence of ch at or before index
// fromIndex, or -1 if ch doesn`t occur at or before fromIndex.
//------------------------------------------------------------------------
StringInt String::lastIndexOf(int ch, StringInt fromIndex) const
   {
   require(isValid());
   require(fromIndex >= 0);
   require(fromIndex <= length());

   fromIndex = min(fromIndex, length() - 1);
   for (StringInt index = fromIndex; index >= 0; index--)
      if (charAt(index) == ch)
         return index;
   return -1;
   }

//------------------------------------------------------------------------
// Returns the index of the last occurrence of str starting at or before
// index fromIndex, or -1 if str doesn`t occur at or before fromIndex.
//------------------------------------------------------------------------
StringInt String::lastIndexOf(const String &str, StringInt fromIndex) const
   {
   require(isValid());
   require(str.isValid());
   require(fromIndex >= 0);
   require(fromIndex <= length());

   StringInt strLength = str.length();
   fromIndex = min(fromIndex, length() - 1);
   for (StringInt index = fromIndex; index >= 0; index--)
      if (regionMatches(index, str, 0, strLength) == true)
         return index;
   return -1;
   }

//------------------------------------------------------------------------
// Returns the index of the first occurrence at or after startIndex of any
// character in charList, or -1 if no such character can be found.
//------------------------------------------------------------------------
StringInt String::indexOfCharIn(const String &charList,
                                StringInt startIndex) const
   {
   require(isValid());
   require(charList.isValid());
   require(startIndex >= 0);
   require(startIndex <= length());

   for (StringInt index = startIndex; index < length(); index++)
      if (charList.indexOf(charAt(index)) >= 0)
         return index;
   return -1;
   }

//------------------------------------------------------------------------
// Returns the index of the first occurrence at or after startIndex of any
// character NOT in charList, or -1 if no such character can be found.
//------------------------------------------------------------------------
StringInt String::indexOfCharNotIn(const String &charList,
                                   StringInt startIndex) const
   {
   require(isValid());
   require(charList.isValid());
   require(startIndex >= 0);
   require(startIndex <= length());

   for (StringInt index = startIndex; index < length(); index++)
      if (charList.indexOf(charAt(index)) < 0)
         return index;
   return -1;
   }

//------------------------------------------------------------------------
// Returns the index of the first occurrence at or after startIndex of any
// character NOT in the 1-character charList, or -1 if no such character
// can be found.
//------------------------------------------------------------------------
StringInt String::indexOfCharNotIn(char charList,
                                   StringInt startIndex) const
   {
   require(isValid());
   require(startIndex >= 0);
   require(startIndex <= length());

   for (StringInt index = startIndex; index < length(); index++)
      if (charAt(index) != charList)
         return index;
   return -1;
   }

//------------------------------------------------------------------------
// Returns the index of the last occurrence at or before startIndex of any
// character in charList, or -1 if no such character can be found.
//------------------------------------------------------------------------
StringInt String::lastIndexOfCharIn(const String &charList,
                                    StringInt startIndex) const
   {
   require(isValid());
   require(charList.isValid());
   require(startIndex >= 0);
   require(startIndex <= length());

   startIndex = min(startIndex, length() - 1);
   for (StringInt index = startIndex; index >= 0; index--)
      if (charList.indexOf(charAt(index)) >= 0)
         return index;
   return -1;
   }

//------------------------------------------------------------------------
// Returns the index of the last occurrence at or before startIndex of any
// character NOT in charList, or -1 if no such character can be found.
//------------------------------------------------------------------------
StringInt String::lastIndexOfCharNotIn(const String &charList,
                                       StringInt startIndex) const
   {
   require(isValid());
   require(charList.isValid());
   require(startIndex >= 0);
   require(startIndex <= length());

   startIndex = min(startIndex, length() - 1);
   for (StringInt index = startIndex; index >= 0; index--)
      if (charList.indexOf(charAt(index)) < 0)
         return index;
   return -1;
   }

//------------------------------------------------------------------------
// Returns the index of the last occurrence at or before startIndex of any
// character NOT in the 1-character charList, or -1 if no such character
// can be found.
//------------------------------------------------------------------------
StringInt String::lastIndexOfCharNotIn(char charList,
                                       StringInt startIndex) const
   {
   require(isValid());
   require(startIndex >= 0);
   require(startIndex <= length());

   startIndex = min(startIndex, length() - 1);
   for (StringInt index = startIndex; index >= 0; index--)
      if (charAt(index) != charList)
         return index;
   return -1;
   }

//------------------------------------------------------------------------
// Returns the index of the first whitespace character at or after index
// startIndex, or -1 if there is no whitespace character at or after
// startIndex.
//------------------------------------------------------------------------
StringInt String::whitespaceIndex(StringInt startIndex) const
   {
   require(isValid());
   require(startIndex >= 0);
   require(startIndex <= length());

   for (StringInt index = startIndex; index < length(); index++)
      if (isspace(charAt(index)))
         return index;
   return -1;
   }

//------------------------------------------------------------------------
// Returns the index of the first non-whitespace character at or after
// index startIndex, or -1 if there is no non-whitespace character at or
// after startIndex.
//------------------------------------------------------------------------
StringInt String::nonWhitespaceIndex(StringInt startIndex) const
   {
   require(isValid());
   require(startIndex >= 0);
   require(startIndex <= length());

   for (StringInt index = startIndex; index < length(); index++)
      if ( !isspace(charAt(index)) )
         return index;
   return -1;
   }

//------------------------------------------------------------------------
// Returns the index of the last whitespace character at or before index
// startIndex, or -1 if there is no whitespace character at or before
// startIndex.
//------------------------------------------------------------------------
StringInt String::lastWhitespaceIndex(StringInt startIndex) const
   {
   require(isValid());
   require(startIndex >= 0);
   require(startIndex <= length());

   startIndex = min(startIndex, length() - 1);
   for (StringInt index = startIndex; index >= 0; index--)
      if (isspace(charAt(index)))
         return index;
   return -1;
   }

//------------------------------------------------------------------------
// Returns the index of the last non-whitespace character at or before
// index startIndex, or -1 if there is no non-whitespace character at or
// before startIndex.
//------------------------------------------------------------------------
StringInt String::lastNonWhitespaceIndex(StringInt startIndex) const
   {
   require(isValid());
   require(startIndex >= 0);
   require(startIndex <= length());

   startIndex = min(startIndex, length() - 1);
   for (StringInt index = startIndex; index >= 0; index--)
      if ( !isspace(charAt(index)) )
         return index;
   return -1;
   }

//------------------------------------------------------------------------
// Returns the index of the first digit at or after index startIndex,
// or -1 if there is no digit at or after startIndex.
//------------------------------------------------------------------------
StringInt String::digitIndex(StringInt startIndex) const
   {
   require(isValid());
   require(startIndex >= 0);
   require(startIndex <= length());

   for (StringInt index = startIndex; index < length(); index++)
      if (isdigit(charAt(index)))
         return index;
   return -1;
   }

//------------------------------------------------------------------------
// Returns the index of the first non-digit at or after index startIndex,
// or -1 if there is no non-digit character at or after startIndex.
//------------------------------------------------------------------------
StringInt String::nonDigitIndex(StringInt startIndex) const
   {
   require(isValid());
   require(startIndex >= 0);
   require(startIndex <= length());

   for (StringInt index = startIndex; index < length(); index++)
      if ( !isdigit(charAt(index)))
         return index;
   return -1;
   }

//------------------------------------------------------------------------
// Returns a string containing the result of concatenating *this and str.
//------------------------------------------------------------------------
String String::concat(const String &str) const
   {
   require(isValid());

   String result;
   if (str.length() == 0)
      result = *this;
   else if (length() == 0)
      result = str;
   else
      {
      StringData *newData = getStringData(length() + str.length());
      strcpy(newData->contents, data->contents);
      strcpy(newData->contents + length(), str.data->contents);
      result = String(newData);
      }

   ensure(result.isValid());
   ensure(result.length() == length() + str.length());
   return result;
   }

//------------------------------------------------------------------------
// Returns a string made up of numCopies concatenated copies of *this.
//------------------------------------------------------------------------
String String::copies(StringInt numCopies) const
   {
   require(isValid());
   require(numCopies >= 0);

   String result;
   if (numCopies > 0 && length() > 0)
      {
      StringBuffer buf = *this;
      buf.ensureCapacity(numCopies * length());
      for (StringInt i = 1; i < numCopies; i++)
         buf.append(*this);
      result = buf.toString();
      }

   ensure(result.isValid());
   ensure(result.length() == numCopies * length());
   return result;
   }

//------------------------------------------------------------------------
// Returns a copy of *this with every occurrence of oldChar changed
// to newChar.
//------------------------------------------------------------------------
String String::replace(char oldChar, char newChar) const
   {
   require(isValid());
   require(newChar != terminator());

   String result;
   if (oldChar == terminator() || indexOf(oldChar) < 0)
      result = *this;
   else
      {
      StringData *newData = getStringData(length());
      for (StringInt i = 0; i < length(); i++)
         newData->contents[i] = (charAt(i) == oldChar) ? newChar : charAt(i);
      result = String(newData);
      }

   ensure(result.isValid());
   ensure(result.length() == length());
   return result;
   }

//------------------------------------------------------------------------
// Returns a copy of *this with every character in inCharList replaced by
// the character in the same position in outCharList. If the same
// character occurs more than once in inCharList, the leftmost one is used.
// inCharList and outCharList must have the same number of characters.
//------------------------------------------------------------------------
String String::translate(const String &inCharList,
                         const String &outCharList) const
   {
   require(isValid());
   require(inCharList.isValid());
   require(outCharList.isValid());
   require(inCharList.length() == outCharList.length());

   StringBuffer buf = *this;
   StringInt listIndex = -1;
   for (StringInt index = 0; index < length(); index++)
      {
      listIndex = inCharList.indexOf(charAt(index));
      if (listIndex >= 0)
         buf.setCharAt(index, outCharList.charAt(listIndex));
      }

   ensure(buf.isValid());
   ensure(buf.length() == length());
   return buf.toString();
   }

//------------------------------------------------------------------------
// Returns a string with the same contents as the substring of *this that
// starts at index fromIndex and continues to index toIndex. An empty
// string is returned if toIndex < fromIndex.
//------------------------------------------------------------------------
String String::substring(StringInt fromIndex, StringInt toIndex) const
   {
   require(isValid());
   require(fromIndex >= 0);
   require(fromIndex <= length());
   require(toIndex >= 0);
   require(toIndex <= length());

   StringInt newLength = min(toIndex, length() - 1) - fromIndex + 1;
   String result;
   if (newLength > 0)
      {
      if (fromIndex == 0 && toIndex >= length() - 1)
         result = *this;
      else
         {
         StringData *newData = getStringData(newLength);
         memcpy(newData->contents,
                data->contents + fromIndex, newData->length);
         result = String(newData);
         }
      }

   ensure(result.isValid());
   ensure(newLength < 0 || result.length() == newLength);
   ensure(newLength <= 0 || indexOf(result, fromIndex) == fromIndex);
   return result;

   }  // String::substring(StringInt, StringInt)

//------------------------------------------------------------------------
// Returns a string containing the leftmost numChars characters of *this,
// padding it on the right with spaces as necessary.
//------------------------------------------------------------------------
String String::left(StringInt numChars) const
   {
   require(isValid());
   require(numChars >= 0);

   StringBuffer buf = *this;
   buf.setLength(numChars);

   ensure(buf.isValid());
   ensure(buf.length() == numChars);
   return String(buf);
   }

//------------------------------------------------------------------------
// Returns a string containing the rightmost numChars characters of *this,
// padding it on the left with spaces as necessary.
//------------------------------------------------------------------------
String String::right(StringInt numChars) const
   {
   require(isValid());
   require(numChars >= 0);

   String result;
   if (numChars <= length())
      result = substring(length() - numChars);
   else
      result = String(' ', numChars - length()) + *this;

   ensure(result.isValid());
   ensure(result.length() == numChars);
   return result;
   }

//------------------------------------------------------------------------
// Returns a string of length numChars with *this centred in it, with
// spaces added as necessary to attain the required length. If an odd
// number of characters are to be added (or truncated), then the right
// side gains (loses) one more character than does the left.
//------------------------------------------------------------------------
String String::centre(StringInt numChars) const
   {
   require(isValid());
   require(numChars >= 0);

   String result;
   StringInt addLength = numChars - length();
   if (addLength > 0)
      {
      StringBuffer buf(numChars);
      result = buf.overlay(addLength / 2, *this).toString();
      }
   else if (addLength < 0)
      {
      result = substring(-addLength / 2,
                         (length() - 1) - (-addLength + 1) / 2);
      }
   else
      result = *this;

   ensure(result.isValid());
   ensure(result.length() == numChars);
   return result;
   }  // String::centre(StringInt)

//------------------------------------------------------------------------
// Returns a copy of *this with all leading and trailing whitespace
// removed.
//------------------------------------------------------------------------
String String::strip() const
   {
   require(isValid());

   String result;
   StringInt firstIndex = nonWhitespaceIndex();
   if (firstIndex >= 0)
      {
      StringInt lastIndex = lastNonWhitespaceIndex();
      check(lastIndex >= 0);
      result = substring(firstIndex, lastIndex);
      }

   ensure(result.isValid());
   return result;
   }

//------------------------------------------------------------------------
// Returns a copy of *this with all leading and trailing ch`s removed.
//------------------------------------------------------------------------
String String::strip(char ch) const
   {
   require(isValid());

   String result;
   StringInt firstIndex = indexOfCharNotIn(ch);
   if (firstIndex >= 0)
      {
      StringInt lastIndex = lastIndexOfCharNotIn(ch);
      check(lastIndex >= 0);
      result = substring(firstIndex, lastIndex);
      }

   ensure(result.isValid());
   return result;
   }

//------------------------------------------------------------------------
// Returns a copy of *this with all leading whitespace removed.
//------------------------------------------------------------------------
String String::stripLeading() const
   {
   require(isValid());

   String result;
   StringInt firstIndex = nonWhitespaceIndex();
   if (firstIndex >= 0)
      result = substring(firstIndex);

   ensure(result.isValid());
   return result;
   }

//------------------------------------------------------------------------
// Returns a copy of *this with all leading ch`s removed.
//------------------------------------------------------------------------
String String::stripLeading(char ch) const
   {
   require(isValid());

   String result;
   StringInt firstIndex = indexOfCharNotIn(ch);
   if (firstIndex >= 0)
      result = substring(firstIndex);

   ensure(result.isValid());
   return result;
   }

//------------------------------------------------------------------------
// Returns a copy of *this with all trailing whitespace removed.
//------------------------------------------------------------------------
String String::stripTrailing() const
   {
   require(isValid());

   String result;
   StringInt lastIndex = lastNonWhitespaceIndex();
   if (lastIndex >= 0)
      result = left(lastIndex + 1);

   ensure(result.isValid());
   return result;
   }

//------------------------------------------------------------------------
// Returns a copy of *this with all trailing ch`s removed.
//------------------------------------------------------------------------
String String::stripTrailing(char ch) const
   {
   require(isValid());

   String result;
   StringInt lastIndex = lastIndexOfCharNotIn(ch);
   if (lastIndex >= 0)
      result = left(lastIndex + 1);

   ensure(result.isValid());
   return result;
   }

//------------------------------------------------------------------------
// Returns a copy of *this, but all lower case.
//------------------------------------------------------------------------
String String::toLowerCase() const
   {
   require(isValid());

   StringBuffer buf = *this;
   for (StringInt i = 0; i < length(); i++)
      buf.setCharAt(i, tolower(charAt(i)));

   ensure(buf.isValid());
   ensure(buf.length() == length());
   ensure(equalsIgnoreCase(buf));
   return buf.toString();
   }

//------------------------------------------------------------------------
// Returns a copy of *this, but all upper case.
//------------------------------------------------------------------------
String String::toUpperCase() const
   {
   require(isValid());

   StringBuffer buf = *this;
   for (StringInt i = 0; i < length(); i++)
      buf.setCharAt(i, toupper(charAt(i)));

   ensure(buf.isValid());
   ensure(buf.length() == length());
   ensure(equalsIgnoreCase(buf));
   return buf.toString();
   }

//------------------------------------------------------------------------
// Returns the number of words in *this starting at or after index
// startIndex.
//------------------------------------------------------------------------
StringInt String::wordCount(StringInt startIndex) const
   {
   require(isValid());
   require(startIndex >= 0);
   require(startIndex <= length());

   StringInt count = 0;
   StringInt index = startIndex;
   for ( ; ; )
      {
      index = nonWhitespaceIndex(index);
      if (index < 0)
         break;
      count++;
      index = whitespaceIndex(index);
      if (index < 0)
         break;
      }

   ensure(count >= 0);
   return count;
   }

//------------------------------------------------------------------------
// Returns the length of the (wordIndex+1)th word after (character) index
// startIndex (which is 0 if *this doesn`t have that many words).
//------------------------------------------------------------------------
StringInt String::wordLength(StringInt wordIndex,
                             StringInt startIndex) const
   {
   require(isValid());
   require(wordIndex >= 0);

   StringInt wordLen = 0;
   StringInt index = indexOfWord(wordIndex, startIndex);
   if (index >= 0)
      {
      StringInt nextWhiteIndex = whitespaceIndex(index);
      wordLen = (nextWhiteIndex >= 0) ? nextWhiteIndex - index
                                      : length() - index;
      }

   ensure(wordLen >= 0);
   ensure(wordLen <= length());
   return wordLen;
   }

//------------------------------------------------------------------------
// Returns the (character) index of the (wordIndex+1)th word after
// (character) index startIndex, or -1 if there aren`t that many words in
// *this.
//------------------------------------------------------------------------
StringInt String::indexOfWord(StringInt wordIndex,
                              StringInt startIndex) const
   {
   require(isValid());
   require(wordIndex >= 0);

   // skip the first wordIndex words:
   StringInt index = startIndex;
   for (StringInt currWordInd = 0; currWordInd < wordIndex; currWordInd++)
      {
      index = nonWhitespaceIndex(index);
      if (index < 0)
         break;
      index = whitespaceIndex(index);
      if (index < 0)
         break;
      }

   if (index >= 0)
      index = nonWhitespaceIndex(index);

   ensure(index < length());
   return index;
   }

//------------------------------------------------------------------------
// Return the (wordIndex+1)th word after (character) index startIndex in
// *this, or an empty string if there aren`t that many words in *this.
//------------------------------------------------------------------------
String String::word(StringInt wordIndex, StringInt startIndex) const
   {
   require(isValid());
   require(wordIndex >= 0);

   String result;
   StringInt index = indexOfWord(wordIndex, startIndex);
   if (index >= 0)
      {
      StringInt endIndex = whitespaceIndex(index);
      endIndex = (endIndex >= 0) ? endIndex - 1 : length();
      result = substring(index, endIndex);
      }

   ensure(result.isValid());
   return result;
   }

//------------------------------------------------------------------------
// Returns a string containing the (startWordIndex+1)th through the last
// word in *this, or an empty string if there aren`t that many words in
// *this. The resulting string will have no leading or trailing whitespace,
// but will have all of the whitespace between the words.
//------------------------------------------------------------------------
String String::subword(StringInt startWordIndex) const
   {
   require(isValid());
   require(startWordIndex >= 0);

   String result;
   StringInt index = indexOfWord(startWordIndex);
   if (index >= 0)
      result = substring(index).trim();

   ensure(result.isValid());
   return result;
   }

//------------------------------------------------------------------------
// Returns a string containing the (startWordIndex+1)th through the
// (endIndex+1)th (or the last) word after (character) index startIndex in
// *this, or an empty string if there aren`t that many word in *this. The
// resulting string will have no leading or trailing whitespace, but will
// have all of the whitespace between the words.
//------------------------------------------------------------------------
String String::subword(StringInt startWordIndex,
                       StringInt endWordIndex, StringInt startIndex) const
   {
   require(isValid());
   require(startWordIndex >= 0);
   require(endWordIndex >= startWordIndex);

   String result;
   StringInt index = indexOfWord(startWordIndex, startIndex);
   if (index >= 0)
      {
      StringInt endIndex = indexOfWord(endWordIndex, startIndex);
      if (endIndex < 0)
         endIndex = length();
      else
         {
         endIndex = whitespaceIndex(endIndex);
         endIndex = (endIndex >= 0) ? endIndex - 1 : length();
         }
      result = substring(index, endIndex).trim();
      }

   ensure(result.isValid());
   return result;
   }  // String::subword(StringInt, StringInt, StringInt)

//------------------------------------------------------------------------
// Returns a pointer to a StringData object representing an empty string.
//------------------------------------------------------------------------
StringData *String::getStringData()
   {
   static StringData *emptyData = 0;

   if (emptyData == 0)
      {
      emptyData = new StringData;
      emptyData->refCount = 2;   // emptyData and the returned pointer
      emptyData->contents = new char[1];
      emptyData->contents[0] = terminator();
      emptyData->length   = 0;
      emptyData->capacity = 0;
      }
   else
      {
      emptyData->refCount++;
      }

   ensure(emptyData != 0);
   ensure(emptyData->isValid());
   ensure(strcmp(emptyData->contents, "") == 0);
   return emptyData;
   }

//------------------------------------------------------------------------
// Returns a pointer to a new StringData object of length len. Its
// contents are undefined, but null-terminated (see below).
//------------------------------------------------------------------------
StringData *String::getStringData(StringInt len)
   {
   require(len >= 0);

   StringData *sd = 0;
   if (len == 0)
      {
      sd = getStringData();
      }
   else
      {
      sd = new StringData;
      sd->refCount = 1;
      sd->length = sd->capacity = len;
      sd->contents = new char[sd->capacity + 1];
      sd->contents[len] = terminator();
      }

   // Note: because sd->contents is left uninitialized, sd->length may not
   //       equal strlen(sd->contents), so sd->isValid() isn`t checked here.
   ensure(sd != 0);
   ensure(sd->length == len);
   return sd;
   }

//------------------------------------------------------------------------
// Returns a pointer to a StringData object representing a boolean true.
//------------------------------------------------------------------------
StringData *String::getTrueStringData()
   {
   static StringData *trueData = 0;

   if (trueData == 0)
      {
      trueData = new StringData;
      trueData->refCount = 2;    // trueData and the returned pointer
      trueData->contents = new char[5];
      strcpy(trueData->contents, "true");
      trueData->length = trueData->capacity = 4;
      }
   else
      {
      trueData->refCount++;
      }

   ensure(trueData != 0);
   ensure(trueData->isValid());
   ensure(strcmp(trueData->contents, "true") == 0);
   return trueData;
   }

//------------------------------------------------------------------------
// Returns a pointer to a StringData object representing a boolean false.
//------------------------------------------------------------------------
StringData *String::getFalseStringData()
   {
   static StringData *falseData = 0;

   if (falseData == 0)
      {
      falseData = new StringData;
      falseData->refCount = 2;   // falseData and the returned pointer
      falseData->contents = new char[6];
      strcpy(falseData->contents, "false");
      falseData->length = falseData->capacity = 5;
      }
   else
      {
      falseData->refCount++;
      }

   ensure(falseData != 0);
   ensure(falseData->isValid());
   ensure(strcmp(falseData->contents, "false") == 0);
   return falseData;
   }

//------------------------------------------------------------------------
// Deletes the StringData object to which data points.
//------------------------------------------------------------------------
void String::deleteStringData()
   {
   require(isValid());
   require(data == 0 || data->refCount > 0);

   if (data != 0 && --data->refCount == 0)
      {
      delete[] data->contents;
      data->contents = 0;
      delete data;
      data = 0;
      }

   ensure(isValid());
   ensure(data == 0 || data->refCount > 0);
   }

#if CHECK_ASSERTIONS

//------------------------------------------------------------------------
// Returns true iff the state of *this is valid.
//------------------------------------------------------------------------
int String::isValid() const
   {
   int valid = 0;

   if (data == 0)
      valid = 1;
   else if (!data->isValid())
      valid = 0;
   else
      valid = 1;

   return valid;
   }

#endif  // #if CHECK_ASSERTIONS

//************************************************************************
// Class StringBuffer method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Assignment operator.
//------------------------------------------------------------------------
StringBuffer &StringBuffer::operator =(const StringBuffer &c)
   {
   require(c.isValid());

   if (this->data != c.data)
      {
      deleteStringData();
      data = c.data;
      data->refCount++;
      }

   ensure(isValid());
   return *this;
   }

//------------------------------------------------------------------------
// Makes sure the buffer can hold at least requiredCapacity characters.
//------------------------------------------------------------------------
void StringBuffer::ensureCapacity(StringInt requiredCapacity)
   {
   require(isValid());

   if (requiredCapacity > capacity())
      {
      // keep the same StringData object, just increase its capacity:
      char *oldContents = data->contents;
      data->capacity = capacityFor(requiredCapacity);
      data->contents = new char[data->capacity + 1];
      strcpy(data->contents, oldContents);
      delete oldContents;
      }

   ensure(isValid());
   ensure(capacity() >= requiredCapacity);
   }

//------------------------------------------------------------------------
// Sets the character at index index in the buffer to ch.
//------------------------------------------------------------------------
void StringBuffer::setCharAt(StringInt index, char ch)
   {
   require(isValid());
   require(index >= 0);
   require(index < length());

   if (ch == terminator())
      setLength(index);
   else if (data->contents[index] != ch)
      {
      if (referenceCount() > 1)
         copyBeforeWrite();
      data->contents[index] = ch;
      }

   ensure(isValid());
   }

//------------------------------------------------------------------------
// Sets the length of the buffer to len. If len is less than the current
// length, data is lost; if it is greater than length, then the added
// length is filled with spaces.
//------------------------------------------------------------------------
void StringBuffer::setLength(StringInt len)
   {
   require(isValid());
   require(len >= 0);

   StringInt addedLength = len - length();
   if (addedLength != 0)
      {
      if (referenceCount() > 1)
         copyBeforeWrite();
      if (addedLength > 0)
         {
         ensureCapacity(len);
         memset(data->contents + length(), defaultFillChar, addedLength);
         }
      data->length = len;
      data->contents[len] = terminator();
      }

   ensure(isValid());
   ensure(length() == len);
   return;
   }

//------------------------------------------------------------------------
// Clears the contents of the buffer without affecting its capacity.
// Equivalent to setLength(0).
//------------------------------------------------------------------------
void StringBuffer::clear()
   {
   require(isValid());

   if (length() > 0)
      {
      if (referenceCount() > 1)
         copyBeforeWrite();
      data->length = 0;
      data->contents[0] = terminator();
      }

   ensure(isValid());
   ensure(length() == 0);
   }

//------------------------------------------------------------------------
// Appends the contents of str to the buffer`s contents and returns a
// reference to the buffer.
//------------------------------------------------------------------------
StringBuffer &StringBuffer::append(const String &str)
   {
   require(isValid());
   require(str.isValid());

   if (str.length() > 0)
      {
      if (referenceCount() > 1)
         copyBeforeWrite();
      StringInt newLength = length() + str.length();
      ensureCapacity(newLength);
      strcpy(data->contents + length(), str.data->contents);
      data->length = newLength;
      }

   ensure(isValid());
   return *this;
   }

//------------------------------------------------------------------------
// Inserts the contents of str into the buffer`s contents starting at
// index offset and returns a reference to the buffer.
//------------------------------------------------------------------------
StringBuffer &StringBuffer::insert(StringInt offset, const String &str)
   {
   require(isValid());
   require(str.isValid());
   require(offset >= 0);
   require(offset <= length());

   if (str.length() == 0)
      return *this;
   else if (offset == length())
      return append(str);

   if (referenceCount() > 1)
      copyBeforeWrite();
   StringInt newLength = length() + str.length();
   ensureCapacity(newLength);

   // Move the last part of the contents of the buffer to the right far
   // enough so that there is space to insert str`s contents, then copy
   // str`s contents into that space:
   // check(data->contents and str.data->contents don`t overlap);
   memmove(data->contents + offset + str.length(),
           data->contents + offset, length() - offset);
   memcpy(data->contents + offset, str.data->contents, str.length());
   data->contents[newLength] = terminator();
   data->length = newLength;

   ensure(isValid());
   return *this;
   }  // StringBuffer::insert(StringInt, const String &)

//------------------------------------------------------------------------
// Overlays the contents of str onto the buffer`s contents starting at
// index offset and returns a reference to the buffer.
//------------------------------------------------------------------------
StringBuffer &StringBuffer::overlay(StringInt offset, const String &str)
   {
   require(isValid());
   require(str.isValid());
   require(offset >= 0);
   require(offset <= length());

   if (str.length() > 0)
      {
      if (referenceCount() > 1)
         copyBeforeWrite();
      if (offset + str.length() < length())
         {
         memcpy(data->contents + offset, str.data->contents, str.length());
         }
      else
         {
         ensureCapacity(offset + str.length());
         strcpy(data->contents + offset, str.data->contents);
         check(strlen(data->contents) == size_t(offset + str.length()));
         data->length = offset + str.length();
         }
      }

   ensure(isValid());
   return *this;
   }  // StringBuffer::overlay(StringInt, const String &)

//------------------------------------------------------------------------
// Returns a pointer to a StringData object representing an empty string.
//------------------------------------------------------------------------
StringData *StringBuffer::getStringData() const
   {
   static StringData *emptyData = 0;

   if (emptyData == 0)
      {
      emptyData = new StringData;
      emptyData->refCount = 2;   // emptyData and the returned pointer
      emptyData->length   = 0;
      emptyData->capacity = capacityFor(0);
      emptyData->contents = new char[emptyData->capacity + 1];
      emptyData->contents[0] = terminator();
      }
   else
      {
      emptyData->refCount++;
      }

   ensure(emptyData != 0);
   ensure(emptyData->isValid());
   ensure(strcmp(emptyData->contents, "") == 0);
   return emptyData;
   }

//------------------------------------------------------------------------
// Returns a pointer to a new StringData object of length len. Its
// contents are undefined, but are null-terminated (see below).
//------------------------------------------------------------------------
StringData *StringBuffer::getStringData(StringInt len) const
   {
   require(len >= 0);

   StringData *sd = 0;
   if (len == 0)
      {
      sd = getStringData();
      }
   else
      {
      sd = new StringData;
      sd->refCount = 1;
      sd->length = len;
      sd->capacity = capacityFor(len);
      sd->contents = new char[sd->capacity + 1];
      sd->contents[len] = terminator();
      }

   // Note: because sd->contents is left uninitialized, sd->length may not
   //       equal strlen(sd->contents), so sd->isValid() isn`t checked here.
   ensure(sd != 0);
   ensure(sd->length == len);
   return sd;
   }

//------------------------------------------------------------------------
// Returns what the capacity of a buffer that must have a capacity of at
// least requiredCapacity should be.
//------------------------------------------------------------------------
StringInt StringBuffer::capacityFor(StringInt requiredCapacity) const
   {
   if (requiredCapacity < 0)
      requiredCapacity = 0;

   StringInt newCapacity = (requiredCapacity / capacityBlockSize) + 1;
   newCapacity = (newCapacity * capacityBlockSize) - 1;

   ensure(newCapacity >= minimumCapacity);
   ensure(newCapacity > 0);
   ensure(newCapacity >= requiredCapacity);
   return newCapacity;
   }

//------------------------------------------------------------------------
// Makes a copy of the buffer`s StringData object before it changes it
// IF another string also references its StringData object.
//------------------------------------------------------------------------
void StringBuffer::copyBeforeWrite()
   {
   require(isValid());

   if (data->refCount > 1)
      {
      StringData *newData = new StringData;
      newData->refCount = 1;
      newData->length = length();
      newData->capacity = capacityFor(newData->length);
      newData->contents = new char[newData->capacity + 1];
      strcpy(newData->contents, data->contents);

      deleteStringData();
      data = newData;
      }

   ensure(isValid());
   ensure(data->refCount == 1);
   }

//------------------------------------------------------------------------
// Deletes the StringData object to which data points.
//------------------------------------------------------------------------
void StringBuffer::deleteStringData()
   {
   require(isValid());
   require(data == 0 || data->refCount > 0);

   if (data != 0 && --data->refCount == 0)
      {
      delete[] data->contents;
      data->contents = 0;
      delete data;
      data = 0;
      }

   ensure(isValid());
   ensure(data == 0 || data->refCount > 0);
   }

#if CHECK_ASSERTIONS

//------------------------------------------------------------------------
// Returns true iff the state of *this is valid.
//------------------------------------------------------------------------
int StringBuffer::isValid() const
   {
   int valid = 0;

   if (data == 0)
      valid = 1;
   else if (!data->isValid())
      valid = 0;
   else
      valid = 1;

   return valid;
   }

#endif  // #if CHECK_ASSERTIONS

//************************************************************************
// Class StringData method definitions.
//************************************************************************

#if CHECK_ASSERTIONS

//------------------------------------------------------------------------
// Returns true iff the state of *this is valid.
//------------------------------------------------------------------------
int StringData::isValid() const
   {
   int valid = 0;

   if (length < 0)
      {
      debugOutput("length is negative.\n");
      }
   else if (capacity < 0)
      {
      debugOutput("capacity is negative.\n");
      }
   else if (refCount < 1)
      {
      debugOutput("refCount is less than 1.\n");
      }
   else if (contents == 0)
      {
      debugOutput("contents is a NULL pointer, which shouldn\'t happen.\n");
      }
   else if (length > capacity)
      {
      debugOutput("length is larger than capacity.\n");
      }
   else if (size_t(length) != strlen(contents))
      {
      char buf[50];
      sprintf(buf, "length != strlen(contents) (== %i).\n", strlen(contents));
      debugOutput(buf);
      }
   else
      valid = 1;

   if (!valid)
      {
      outputState();
      }

   return valid;
   }

//------------------------------------------------------------------------
// Outputs the current state of a StringData object via debugOutput().
//------------------------------------------------------------------------
void StringData::outputState() const
   {
	const int bufLength = 512;
   char buf[bufLength];
   char format[] = "this = %p; refCount = "SIF"; length = "SIF"; "
                   "capacity = "SIF"; contents = %p\n*contents = ";

   sprintf(buf, format, this, refCount, length, capacity, contents);
   if ( (bufLength - strlen(buf) - 3) > 0)
      strncat(buf, contents, bufLength - strlen(buf) - 3);
   strcat(buf, "\n");
   debugOutput(buf);
   }

#endif  // #if CHECK_ASSERTIONS

//************************************************************************
// Class WordIterator method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Assignment operator.
//------------------------------------------------------------------------
WordIterator &WordIterator::operator =(const WordIterator &wi)
   {
   require(wi.isValid());

   if (this != &wi)
      {
      theString = wi.theString;
      indexOfNextWord = wi.indexOfNextWord;
      }

   ensure(isValid());
   return *this;
   }

//------------------------------------------------------------------------
// Returns the next word from the String (which is the empty String if
// there are no more words).
//------------------------------------------------------------------------
String WordIterator::nextWord()
   {
   require(isValid());

   String next;
   if (indexOfNextWord >= 0)
      {
      next = theString.word(0, indexOfNextWord);
      indexOfNextWord =
                   theString.indexOfWord(0, indexOfNextWord + next.length());
      }

   ensure(isValid());
   return next;
   }

