//************************************************************************
// Declaration of the String, StringBuffer and WordIterator classes, as 
// well as the StringData helper class.
//------------------------------------------------------------------------
// Copyright (C) 1997 by James MacKay
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// The author can be contacted via e-mail at: jmackay@sentex.net
//
//------------------------------------------------------------------------
// Notes:
// - Strings and StringBuffers behave as though their contents are
//   terminated with the character returned by the terminator() methods.
//    - inserting the terminator character into a string will cause the
//      string to end just before that character. It will still be valid,
//      just shorter
//    - the character past the last character in a string (i.e. the
//      length()`th character) can be referenced and accessed (e.g. by
//      (charAt()), but cannot be directly changed (e.g. by setCharAt()).
//    - methods that search a string for a character or a string (e.g.
//      indexOf()) will NOT include the terminator character in the search
//
// - an empty string contains only an empty string. However, methods that
//   return the index of the searched-for string will return -1 when
//   searching for an empty string in an empty string, since it isn`t at
//   an index in the string.
// - every string contains an empty string at the first index it is
//   searched for (provided the index is valid)
//
//************************************************************************
#ifndef JSTRING_H_
#define JSTRING_H_

#include <string.h>
#include <ctype.h>
#include <values.h>
#include "release.h"

//------------------------------------------------------------------------
// Define the StringInt type and related quantities.
//------------------------------------------------------------------------
typedef int StringInt;    // must be a signed type
const int maxStringInt = MAXINT;

// SIF stands for StringInt Format, and is the printf() format specifier
#define SIF "%i"
   
//------------------------------------------------------------------------
// if the compiler`s library doesn`t contain the stricmp() and or the
// strnicmp() functions (versions of strcmp() and strncmp(), respectively,
// that ignore case in the comparison), define them here:
//------------------------------------------------------------------------
#if TARGET_PLATFORM == LINUX_PLATFORM
   inline int stricmp(const char *s1, const char *s2)
      {
      return strcasecmp(s1, s2);
      }
   inline int strnicmp(const char *s1, const char *s2, size_t len)
      {
      return strncasecmp(s1, s2, len);
      }
#endif

//************************************************************************
// Class declarations.
//************************************************************************

class String;
class StringBuffer;
class StringData;
class WordIterator;

//------------------------------------------------------------------------
// Class String.
//------------------------------------------------------------------------
class String
   {
   public:
      String();
      String(const String &c);
      String(const char *str);
      String(const char *str, StringInt offset, StringInt count);
      String(const StringBuffer &buf);
      String(char fillChar, StringInt numCopies);
      virtual ~String();

      String &operator =(const String &c);
      operator const char *() const;
      void getChars(StringInt fromIndex, StringInt toIndex, 
                    char *dest, StringInt destBeginOffset) const;
      char *toCharArray() const;
      String toString() const;

      static char terminator();

      static String copyValueOf(const char *str);
      static String copyValueOf(const char *str, StringInt offset, 
                                StringInt count);
      static String valueOf(char *str);
      static String valueOf(boolean b);
      static String valueOf(char c);
      static String valueOf(short int si);
      static String valueOf(unsigned short int usi);
      static String valueOf(int i);
      static String valueOf(unsigned int ui);
      static String valueOf(long l);
      static String valueOf(unsigned long ul);

      static StringInt min(StringInt si1, StringInt si2);
      static StringInt max(StringInt si1, StringInt si2);

      char charAt(StringInt index) const;
      StringInt length() const;

      int compareTo(const String &str) const;
      int compareToIgnoreCase(const String &str) const;      
      boolean equals(const String &str) const;
      boolean equalsIgnoreCase(const String &str) const;

      boolean startsWith(const String &str, StringInt offset = 0) const;
      boolean endsWith(const String &suffix) const;
      boolean regionMatches(StringInt thisOffset, const String &str,
                            StringInt strOffset, StringInt len) const;
      boolean regionMatchesIgnoreCase(StringInt thisOffset, 
                                      const String &str,
                                      StringInt strOffset, 
                                      StringInt len) const;
      boolean isAbbreviationOf(const String &fullStr, 
                               StringInt minLength) const;

      StringInt indexOf(int ch, StringInt fromIndex = 0) const;
      StringInt indexOf(const String &str, StringInt fromIndex = 0) const;
      StringInt lastIndexOf(int ch) const;
      StringInt lastIndexOf(int ch, StringInt fromIndex) const;
      StringInt lastIndexOf(const String &str) const;
      StringInt lastIndexOf(const String &str, StringInt fromIndex) const;
 
      StringInt indexOfCharIn(const String &charList, 
                              StringInt startIndex = 0) const;
      StringInt indexOfCharNotIn(const String &charList,
                                 StringInt startIndex = 0) const;
      StringInt indexOfCharNotIn(char charList, 
                                 StringInt startIndex = 0) const;
      StringInt lastIndexOfCharIn(const String &charList,
                                  StringInt startIndex) const;
      StringInt lastIndexOfCharIn(const String &charList) const;
      StringInt lastIndexOfCharNotIn(const String &charList,
                                     StringInt startIndex) const;
      StringInt lastIndexOfCharNotIn(const String &charList) const;
      StringInt lastIndexOfCharNotIn(char charList, 
                                     StringInt startIndex) const;
      StringInt lastIndexOfCharNotIn(char charList) const;

      StringInt whitespaceIndex(StringInt startIndex = 0) const;
      StringInt nonWhitespaceIndex(StringInt startIndex = 0) const;
      StringInt lastWhitespaceIndex(StringInt startIndex) const;
      StringInt lastWhitespaceIndex() const;
      StringInt lastNonWhitespaceIndex(StringInt startIndex) const;
      StringInt lastNonWhitespaceIndex() const;

      StringInt digitIndex(StringInt startIndex = 0) const;
      StringInt nonDigitIndex(StringInt startIndex = 0) const;

      String concat(const String &str) const;
      String operator +(const String &str) const;
      String copies(StringInt numCopies) const;

      String replace(char oldChar, char newChar) const;
      String translate(const String &inCharList, 
                       const String &outCharList) const;

      String substring(StringInt fromIndex) const;
      String substring(StringInt fromIndex, StringInt toIndex) const;
      String left(StringInt numChars) const;
      String right(StringInt numChars) const;
      String centre(StringInt numChars) const;
      String center(StringInt numChars) const;

      String strip() const;
      String strip(char ch) const;
      String stripLeading() const;
      String stripLeading(char ch) const;
      String stripTrailing() const;
      String stripTrailing(char ch) const;
      String trim() const;

      String toLowerCase() const;
      String toUpperCase() const;

      StringInt wordCount(StringInt startIndex = 0) const;
      StringInt wordLength(StringInt wordIndex,
                           StringInt startIndex = 0) const;
      StringInt indexOfWord(StringInt wordIndex,
                            StringInt startIndex = 0) const;
      String    word(StringInt wordIndex, StringInt startIndex = 0) const;
      String    subword(StringInt startWordIndex) const;
      String    subword(StringInt startWordIndex,
                        StringInt endWordIndex, 
                        StringInt startIndex = 0) const;
   protected:
      #if CHECK_ASSERTIONS
         int isValid() const;   
      #endif
   private:
      friend class StringBuffer;

      StringInt referenceCount() const;

      String(StringData *sd);
      static StringData *getStringData();
      static StringData *getStringData(StringInt len);
      static StringData *getTrueStringData();
      static StringData *getFalseStringData();
      void deleteStringData();

      StringData *data;

   }; // class String

//------------------------------------------------------------------------
// Class StringBuffer.
//------------------------------------------------------------------------
class StringBuffer
   {
   public:
      StringBuffer();
      StringBuffer(const StringBuffer &c);
      StringBuffer(const String &str);
      StringBuffer(StringInt len);
      virtual ~StringBuffer();

      StringBuffer &operator =(const StringBuffer &c);
      operator const char *() const;
      void getChars(StringInt fromIndex, StringInt toIndex, 
                    char *dest, StringInt destBeginOffset) const;
      String toString() const;

      static char terminator();

      StringInt capacity() const;
      StringInt length() const;
      char charAt(StringInt index) const;

      void ensureCapacity(StringInt requiredCapacity);
      void setCharAt(StringInt index, char ch);
      void setLength(StringInt newLength);
      void clear();

      StringBuffer &append(const String &str);
      StringBuffer &append(const char *str);
      StringBuffer &append(const char *str, StringInt offset, 
                            StringInt count);
      StringBuffer &append(boolean b);
      StringBuffer &append(char c);
      StringBuffer &append(short si);
      StringBuffer &append(unsigned short usi);
      StringBuffer &append(int i);
      StringBuffer &append(unsigned int ui);
      StringBuffer &append(long l);
      StringBuffer &append(unsigned long ul);

      StringBuffer &insert(StringInt offset, const String &str);
      StringBuffer &insert(StringInt offset, const char *str);
      StringBuffer &insert(StringInt offset, boolean b);
      StringBuffer &insert(StringInt offset, char c);
      StringBuffer &insert(StringInt offset, short si);
      StringBuffer &insert(StringInt offset, unsigned short usi);
      StringBuffer &insert(StringInt offset, int i);
      StringBuffer &insert(StringInt offset, unsigned int ui);
      StringBuffer &insert(StringInt offset, long l);
      StringBuffer &insert(StringInt offset, unsigned long ul);

      StringBuffer &overlay(StringInt offset, const String &str);
   protected:
      #if CHECK_ASSERTIONS
         int isValid() const;  
      #endif
   private:
      friend class String;

      StringData *getStringData() const;
      StringData *getStringData(StringInt len) const;

      StringInt  capacityFor(StringInt requiredCapacity) const;
      StringInt referenceCount() const;

      void copyBeforeWrite();
      void deleteStringData();

      static const StringInt capacityBlockSize;
      static const StringInt minimumCapacity;
      static const char defaultFillChar;

      StringData *data;

   }; // class StringBuffer

//------------------------------------------------------------------------
// Class StringData.
//------------------------------------------------------------------------
class StringData
   {
   public:
      static const char terminator;

      char *contents;
      StringInt refCount;
      StringInt length;    // not including the terminator
      StringInt capacity;  // not including the terminator

      #if CHECK_ASSERTIONS
         int isValid() const;
         void outputState() const;
      #endif
   }; // class StringData

//------------------------------------------------------------------------
// Class WordIterator.
//------------------------------------------------------------------------
class WordIterator
   {
   public:
      WordIterator(const String &str);
      WordIterator(const WordIterator &wi);
      WordIterator &operator =(const WordIterator &wi);

      String nextWord();
      boolean hasMoreWords() const;
   private:
      #if CHECK_ASSERTIONS
         int isValid() const; 
      #endif

      String theString;
      StringInt indexOfNextWord;

   }; // class WordIndex

#ifdef ENABLE_INLINES
   #include "jstring.inl"
#endif

#endif   // #ifndef JSTRING_H_

