//************************************************************************
// Inline method definitions for all classes declared in jstring.h.
//------------------------------------------------------------------------
// Copyright (C) 1997 by James MacKay
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// The author can be contacted via e-mail at: jmackay@sentex.net
//
//************************************************************************

//************************************************************************
// String`s inline method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Returns the length of the String.
//------------------------------------------------------------------------
INLINE StringInt String::length() const
   {
   require(isValid());
   return data->length;
   }

//------------------------------------------------------------------------
// Returns the number of references to the String`s StringData.
//------------------------------------------------------------------------
INLINE StringInt String::referenceCount() const
   {
   require(isValid());
   return data->refCount;
   }

//------------------------------------------------------------------------
// Default constructor.
//------------------------------------------------------------------------
INLINE String::String()
   {
   data = getStringData();
   ensure(isValid());
   }

//------------------------------------------------------------------------
// Copy constructor.
//------------------------------------------------------------------------
INLINE String::String(const String &str)
   {
   require(str.isValid());

   data = str.data;
   data->refCount++;

   ensure(isValid());
   }

//------------------------------------------------------------------------
// Constructs a String from a null-terminated character array.
//------------------------------------------------------------------------
INLINE String::String(const char *str)
   {
   data = getStringData(strlen(str));
   memcpy(data->contents, str, data->length);

   ensure(isValid());
   }

//------------------------------------------------------------------------
// Constructs a String from a subarray of a null-terminated character 
// array.
//------------------------------------------------------------------------
INLINE String::String(const char *str, StringInt offset, StringInt count)
   {
   require(offset >= 0);   
   require(size_t(offset) < strlen(str));
   require(count >= 0);
   require(size_t(offset + count) <= strlen(str));

   data = getStringData(count);
   memcpy(data->contents, str + offset, count);

   ensure(isValid());
   ensure(length() == count);
   ensure(memcmp(data->contents, str + offset, length()) == 0);
   }

//------------------------------------------------------------------------
// Constructs a String from a StringBuffer.
//------------------------------------------------------------------------
INLINE String::String(const StringBuffer &buf)
   {
   require(buf.isValid());

   data = buf.data;
   data->refCount++;

   ensure(isValid());
   }

//------------------------------------------------------------------------
// Constructs a String from a pointer to a StringData object.
//------------------------------------------------------------------------
INLINE String::String(StringData *sd)
   {
   require(sd->isValid());

   data = sd;

   ensure(isValid());
   }

//------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------
INLINE String::~String()
   {
   deleteStringData();
   }

//------------------------------------------------------------------------
// Converts a String to a constant char pointer.
//------------------------------------------------------------------------
INLINE String::operator const char *() const
   {
   require(isValid());
   return data->contents;
   }

//------------------------------------------------------------------------
// Returns the character that terminates all Strings.
//------------------------------------------------------------------------
INLINE char String::terminator()
   {
   return StringData::terminator;
   }

//------------------------------------------------------------------------
// Returns a String containing a copy of the contents of str.
//------------------------------------------------------------------------
INLINE String String::copyValueOf(const char *str)
   {
   return String(str);
   }

//------------------------------------------------------------------------
// Returns a String containing a copy of the subarray of str starting at
// index offset that is count chcarcters long.
//------------------------------------------------------------------------
INLINE String String::copyValueOf(const char *str, StringInt offset,
                                    StringInt count)
   {
   return String(str, offset, count);
   }

//------------------------------------------------------------------------
// Returns the value of b represented as a String.
//------------------------------------------------------------------------
INLINE String String::valueOf(boolean b)
   {
   StringData *newData = (b == true) ? getTrueStringData()
                                     : getFalseStringData();
   return String(newData);
   }

//------------------------------------------------------------------------
// Returns the value of si represented as a String.
//------------------------------------------------------------------------
INLINE String String::valueOf(short int si)
   {
   char temp[80];
   sprintf(temp, "%hi", si);
   return String(temp);
   }

//------------------------------------------------------------------------
// Returns the value of usi represented as a String.
//------------------------------------------------------------------------
INLINE String String::valueOf(unsigned short int usi)
   {
   char temp[80];
   sprintf(temp, "%hu", usi);
   return String(temp);
   }

//------------------------------------------------------------------------
// Returns the value of i represented as a String.
//------------------------------------------------------------------------
INLINE String String::valueOf(int i)
   {
   char temp[80];
   sprintf(temp, "%i", i);
   return String(temp);
   }

//------------------------------------------------------------------------
// Returns the value of ui represented as a String.
//------------------------------------------------------------------------
INLINE String String::valueOf(unsigned ui)
   {
   char temp[80];
   sprintf(temp, "%u", ui);
   return String(temp);
   }

//------------------------------------------------------------------------
// Returns the value of l represented as a String.
//------------------------------------------------------------------------
INLINE String String::valueOf(long l)
   {
   char temp[80];
   sprintf(temp, "%li", l);
   return String(temp);
   }

//------------------------------------------------------------------------
// Returns the value of ul represented as a String.
//------------------------------------------------------------------------
INLINE String String::valueOf(unsigned long ul)
   {
   char temp[80];
   sprintf(temp, "%lu", ul);
   return String(temp);
   }

//------------------------------------------------------------------------
// Returns the smallest of two StringInts.
//------------------------------------------------------------------------
INLINE StringInt String::min(StringInt si1, StringInt si2)
   {
   return (si1 <= si2) ? si1 : si2;
   }

//------------------------------------------------------------------------
// Returns the largest of two StringInts.
//------------------------------------------------------------------------
INLINE StringInt String::max(StringInt si1, StringInt si2)
   {
   return (si1 >= si2) ? si1 : si2;
   }

//------------------------------------------------------------------------
// Returns the character in the String at index index.
//------------------------------------------------------------------------
INLINE char String::charAt(StringInt index) const
   {
   require(isValid());
   require(index >= 0);
   require(index <= length());

   return data->contents[index];
   }

//------------------------------------------------------------------------
// Compares *this and str and returns the following:
//
//     < 0  if *this is lexicographically less than str,
//    == 0  if *this is lexicographically equal to str,
//     > 0  if *this is lexicographically greater than str.
//------------------------------------------------------------------------
INLINE int String::compareTo(const String &str) const
   {
   require(isValid());
   require(str.isValid());

   if (data == str.data)
      return 0;
   return strcmp(data->contents, str.data->contents);
   }

//------------------------------------------------------------------------
// Compares *this and str and returns the following:
//
//     < 0  if *this is lexicographically less than str (ignoring case),
//    == 0  if *this is lexicographically equal to str (ignoring case),
//     > 0  if *this is lexicographically greater than str (ignoring case).
//------------------------------------------------------------------------
INLINE int String::compareToIgnoreCase(const String &str) const
   {
   require(isValid());
   require(str.isValid());

   if (data == str.data)
      return 0;
   return stricmp(data->contents, str.data->contents);
   }

//------------------------------------------------------------------------
// Returns true iff str is lexicographically equal to *this.
//------------------------------------------------------------------------
INLINE boolean String::equals(const String &str) const
   {
   return (compareTo(str) == 0) ? true : false;
   }

//------------------------------------------------------------------------
// Returns true iff str is lexicographically equal to *this, ignoring the
// case of the contents.
//------------------------------------------------------------------------
INLINE boolean String::equalsIgnoreCase(const String &str) const
   {
   return (compareToIgnoreCase(str) == 0) ? true : false;
   }

//------------------------------------------------------------------------
// Returns true iff the substring of *this starting at index thisOffset
// that is len characters long is equal to the substring of str that
// starts at index strOffset and is len characters long.
//------------------------------------------------------------------------
INLINE boolean String::regionMatches(StringInt thisOffset, 
                                      const String &str,
                                      StringInt strOffset,
                                      StringInt len) const
   {
   require(isValid());
   require(str.isValid());
   require(len >= 0);
   require(thisOffset >= 0);  
   require(thisOffset <= length());
   require(strOffset >= 0); 
   require(strOffset <= str.length());

   return (strncmp(data->contents + thisOffset,
                   str.data->contents + strOffset, len) == 0) ? true : false;
   }

//------------------------------------------------------------------------
// Returns true iff the substring of *this starting at index thisOffset
// that is len characters long is equal (ignoring case) to the substring
// of str that starts at index strOffset and is len characters long.
//   Note that both strings must be non-empty - otherwise one or both of
// the offsets is invalid.
//------------------------------------------------------------------------
INLINE boolean String::regionMatchesIgnoreCase(StringInt thisOffset,
                                               const String &str,
                                               StringInt strOffset,
                                               StringInt len) const
   {
   require(isValid());
   require(str.isValid());
   require(len >= 0);
   require(thisOffset >= 0);
   require(thisOffset <= length());
   require(strOffset >= 0); 
   require(strOffset <= str.length());

   return (strnicmp(data->contents + thisOffset,
                    str.data->contents + strOffset, len) == 0) ? true 
                                                               : false;
   }

//------------------------------------------------------------------------
// Returns true iff *this contains str at index offset. 
//------------------------------------------------------------------------
INLINE boolean String::startsWith(const String &str, StringInt offset) const
   {
   return regionMatches(offset, str, 0, str.length());
   }

//------------------------------------------------------------------------
// Returns true iff *this ends in suffix.
//------------------------------------------------------------------------
INLINE boolean String::endsWith(const String &suffix) const
   {
   if (suffix.length() > length())
      return false;
   return regionMatches(length() - suffix.length(), suffix, 0, 
                        suffix.length());
   }

//------------------------------------------------------------------------
// Returns the index of the last occurrence of ch, or -1 if ch doesn`t
// occur in *this.
//------------------------------------------------------------------------
INLINE StringInt String::lastIndexOf(int ch) const
   {
   return lastIndexOf(ch, length());
   }

//------------------------------------------------------------------------
// Returns the index of the last occurrence of str, or -1 if str doesn`t
// occur in *this.
//------------------------------------------------------------------------
INLINE StringInt String::lastIndexOf(const String &str) const
   {
   return lastIndexOf(str, length());
   }

//------------------------------------------------------------------------
// Returns the index of the last occurrence of any character in charList,
// or -1 if no such character can be found.
//------------------------------------------------------------------------
INLINE StringInt String::lastIndexOfCharIn(const String &charList) const
   {
   return lastIndexOfCharIn(charList, length());
   }

//------------------------------------------------------------------------
// Returns the index of the last occurrence of any character NOT in
// charList, or -1 if no such character can be found.
//------------------------------------------------------------------------
INLINE StringInt String::lastIndexOfCharNotIn(const String &charList) const
   {
   return lastIndexOfCharNotIn(charList, length());
   }

//------------------------------------------------------------------------
// Returns the index of the last occurrence of any character NOT in the
// 1-character charList, or -1 if no such character can be found.
//------------------------------------------------------------------------
INLINE StringInt String::lastIndexOfCharNotIn(char charList) const
   {
   return lastIndexOfCharNotIn(charList, length());
   }

//------------------------------------------------------------------------
// Returns the index of the last whitespace character in *this, or -1 if
// there is no whitespace character in *this.
//------------------------------------------------------------------------
INLINE StringInt String::lastWhitespaceIndex() const
   {
   return lastWhitespaceIndex(length());
   }

//------------------------------------------------------------------------
// Returns the index of the last non-whitespace character in *this, or -1
// if there is no non-whitespace character in *this.
//------------------------------------------------------------------------
INLINE StringInt String::lastNonWhitespaceIndex() const
   {
   return lastNonWhitespaceIndex(length());
   }

//------------------------------------------------------------------------
// Returns a string containing the results of concatenating *this and str.
//------------------------------------------------------------------------
INLINE String String::operator +(const String &str) const
   {
   return concat(str);
   }

//------------------------------------------------------------------------
// Returns a string with the same contents as the substring of *this that
// starts at index fromIndex and continues to the end.
//------------------------------------------------------------------------
INLINE String String::substring(StringInt fromIndex) const
   {
   return substring(fromIndex, length());
   }

//------------------------------------------------------------------------
// Returns a string of length numChars with *this centred in it, with
// spaces added as necessary to attain the required length. If an odd
// number of characters are to be added (or truncated), then the right
// side gains (loses) one more character than does the left.
//------------------------------------------------------------------------
INLINE String String::center(StringInt numChars) const
   {
   return centre(numChars);
   }

//------------------------------------------------------------------------
// Returns a copy of *this with all leading and trailing whitespace
// removed.
//------------------------------------------------------------------------
INLINE String String::trim() const
   {
   return strip();
   }

//------------------------------------------------------------------------
// Returns a copy of *this.
//------------------------------------------------------------------------
INLINE String String::toString() const
   {
   return String(*this);
   }

//------------------------------------------------------------------------
// Copies the characters at index fromIndex through toIndex into the
// character array pointed to by dest starting at index destBeginOffset.
//------------------------------------------------------------------------
INLINE void String::getChars(StringInt fromIndex, StringInt toIndex,
                              char *dest, StringInt destBeginOffset) const
   {
   require(isValid());
   require(fromIndex <= toIndex);
   require(fromIndex >= 0);
   require(toIndex <= length());
   require(destBeginOffset >= 0);

   StringInt copyLength = min(toIndex, length() - 1) - fromIndex + 1;
   memmove(dest + destBeginOffset, data->contents + fromIndex, copyLength);

   ensure(isValid());
   ensure(memcmp(dest + destBeginOffset, 
                 data->contents + fromIndex, copyLength) == 0);
   }

//------------------------------------------------------------------------
// Allocates and returns a pointer to a null-terminated character array
// containing a copy of the contents of *this.
//------------------------------------------------------------------------
INLINE char *String::toCharArray() const
   {
   require(isValid());

   char *array = new char[data->length + 1];
   strcpy(array, data->contents);

   ensure(isValid());
   ensure(array != 0);
   ensure(strlen(array) == size_t(length()));
   ensure(strcmp(array, data->contents) == 0);
   return array;
   }
 
//************************************************************************
// StringBuffer`s inline method definitions.
//************************************************************************

//------------------------------------------------------------------------
// Returns the length of the contents of the buffer.
//------------------------------------------------------------------------
INLINE StringInt StringBuffer::length() const
   {
   require(isValid());
   return data->length;
   }

//------------------------------------------------------------------------
// Returns the number of references to the StringBuffer`s StringData.
//------------------------------------------------------------------------
INLINE StringInt StringBuffer::referenceCount() const
   {
   require(isValid());
   return data->refCount;
   }

//------------------------------------------------------------------------
// Default constructor.
//------------------------------------------------------------------------
INLINE StringBuffer::StringBuffer()
   {
   data = getStringData();
   ensure(isValid());
   }

//------------------------------------------------------------------------
// Copy constructor.
//------------------------------------------------------------------------
INLINE StringBuffer::StringBuffer(const StringBuffer &buf)
   {
   require(buf.isValid());

   data = buf.data;
   data->refCount++;

   ensure(isValid());
   }

//------------------------------------------------------------------------
// Constructs a StringBuffer from a String.
//------------------------------------------------------------------------
INLINE StringBuffer::StringBuffer(const String &str)
   {
   require(str.isValid());

   data = str.data;
   data->refCount++;

   ensure(isValid());
   }

//------------------------------------------------------------------------
// Constructs a buffer of length len, initializing its contents to spaces.
//------------------------------------------------------------------------
INLINE StringBuffer::StringBuffer(StringInt len)
   {
   require(len >= 0);

   data = getStringData(len);
   memset(data->contents, defaultFillChar, data->length);

   ensure(isValid());
   ensure(length() == len);
   }

//------------------------------------------------------------------------
// Destructor.
//------------------------------------------------------------------------
INLINE StringBuffer::~StringBuffer()
   {
   deleteStringData();
   }

//------------------------------------------------------------------------
// Converts a StringBuffer to a constant char pointer.
//------------------------------------------------------------------------
INLINE StringBuffer::operator const char *() const
   {
   require(isValid());
   return data->contents;
   }

//------------------------------------------------------------------------
// Returns the character that terminates all StringBuffers.
//------------------------------------------------------------------------
INLINE char StringBuffer::terminator()
   {
   return StringData::terminator;
   }

//------------------------------------------------------------------------
// Returns the capacity of the buffer.
//------------------------------------------------------------------------
INLINE StringInt StringBuffer::capacity() const
   {
   require(isValid());
   return data->capacity;
   }

//------------------------------------------------------------------------
// Returns the character in the StringBuffer at index index.
//------------------------------------------------------------------------
INLINE char StringBuffer::charAt(StringInt index) const
   {
   require(isValid());
   require(index >= 0);
   require(index <= length());
      
   return data->contents[index];
   }

//------------------------------------------------------------------------
// Copies the characters at index fromIndex through toIndex into the
// character array pointed to by dest starting at index destBeginOffset.
//------------------------------------------------------------------------
INLINE void StringBuffer::getChars(StringInt fromIndex, StringInt toIndex,
                                    char *dest,
                                    StringInt destBeginOffset) const
   {
   require(isValid());
   require(fromIndex <= toIndex);
   require(fromIndex >= 0);
   require(toIndex <= length());
   require(destBeginOffset >= 0);

   StringInt copyLength = String::min(toIndex, length() - 1) - fromIndex + 1;
   memmove(dest + destBeginOffset, data->contents + fromIndex, copyLength);

   ensure(isValid());
   ensure(memcmp(dest + destBeginOffset, 
                 data->contents + fromIndex, copyLength) == 0);
   }

//------------------------------------------------------------------------
// Returns a String with the same contents as *this.
//------------------------------------------------------------------------
INLINE String StringBuffer::toString() const
   {
   return String(*this);
   }

//------------------------------------------------------------------------
// Appends the contents of str to the buffer`s contents and returns a
// reference to the buffer.
//------------------------------------------------------------------------
INLINE StringBuffer &StringBuffer::append(const char *str)
   {
   return append(String(str));
   }

//------------------------------------------------------------------------
// Appends the contents of the subarray of str (starting at index offset
// that is count characters long) to the buffer`s contents and returns a
// reference to the buffer.
//------------------------------------------------------------------------
INLINE StringBuffer &StringBuffer::append(const char *str, StringInt offset,
                                            StringInt count)
   {
   return append(String(str, offset, count));
   }

//------------------------------------------------------------------------
// Appends the value of b to the buffer`s contents and returns a
// reference to the buffer.
//------------------------------------------------------------------------
INLINE StringBuffer &StringBuffer::append(boolean b)
   {
   return append(String::valueOf(b));
   }

//------------------------------------------------------------------------
// Appends the value of c to the buffer`s contents and returns a
// reference to the buffer.
//------------------------------------------------------------------------
INLINE StringBuffer &StringBuffer::append(char c)
   {
   return append(String::valueOf(c));
   }

//------------------------------------------------------------------------
// Appends the value of si to the buffer`s contents and returns a
// reference to the buffer.
//------------------------------------------------------------------------
INLINE StringBuffer &StringBuffer::append(short si)
   {
   return append(String::valueOf(si));
   }

//------------------------------------------------------------------------
// Appends the value of usi to the buffer`s contents and returns a
// reference to the buffer.
//------------------------------------------------------------------------
INLINE StringBuffer &StringBuffer::append(unsigned short usi)
   {
   return append(String::valueOf(usi));
   }

//------------------------------------------------------------------------
// Appends the value of i to the buffer`s contents and returns a
// reference to the buffer.
//------------------------------------------------------------------------
INLINE StringBuffer &StringBuffer::append(int i)
   {
   return append(String::valueOf(i));
   }

//------------------------------------------------------------------------
// Appends the value of ui to the buffer`s contents and returns a
// reference to the buffer.
//------------------------------------------------------------------------
INLINE StringBuffer &StringBuffer::append(unsigned ui)
   {
   return append(String::valueOf(ui));
   }

//------------------------------------------------------------------------
// Appends the value of l to the buffer`s contents and returns a
// reference to the buffer.
//------------------------------------------------------------------------
INLINE StringBuffer &StringBuffer::append(long l)
   {
   return append(String::valueOf(l));
   }

//------------------------------------------------------------------------
// Appends the value of ul to the buffer`s contents and returns a
// reference to the buffer.
//------------------------------------------------------------------------
INLINE StringBuffer &StringBuffer::append(unsigned long ul)
   {
   return append(String::valueOf(ul));
   }

//------------------------------------------------------------------------
// Inserts the contents of str into the buffer`s contents starting at
// index offset and returns a reference to the buffer.
//------------------------------------------------------------------------
INLINE StringBuffer &StringBuffer::insert(StringInt offset,
                                          const char *str)
   {
   return insert(offset, String(str));
   }

//------------------------------------------------------------------------
// Inserts the value of b into the buffer`s contents starting at
// index offset and returns a reference to the buffer.
//------------------------------------------------------------------------
INLINE StringBuffer &StringBuffer::insert(StringInt offset, boolean b)
   {
   return insert(offset, String::valueOf(b));
   }

//------------------------------------------------------------------------
// Inserts the value of c into the buffer`s contents starting at
// index offset and returns a reference to the buffer.
//------------------------------------------------------------------------
INLINE StringBuffer &StringBuffer::insert(StringInt offset, char c)
   {
   return insert(offset, String::valueOf(c));
   }

//------------------------------------------------------------------------
// Inserts the value of si into the buffer`s contents starting at
// index offset and returns a reference to the buffer.
//------------------------------------------------------------------------
INLINE StringBuffer &StringBuffer::insert(StringInt offset, short si)
   {
   return insert(offset, String::valueOf(si));
   }

//------------------------------------------------------------------------
// Inserts the value of usi into the buffer`s contents starting at
// index offset and returns a reference to the buffer.
//------------------------------------------------------------------------
INLINE StringBuffer &StringBuffer::insert(StringInt offset,
                                          unsigned short usi)
   {
   return insert(offset, String::valueOf(usi));
   }

//------------------------------------------------------------------------
// Inserts the value of i into the buffer`s contents starting at
// index offset and returns a reference to the buffer.
//------------------------------------------------------------------------
INLINE StringBuffer &StringBuffer::insert(StringInt offset, int i)
   {
   return insert(offset, String::valueOf(i));
   }

//------------------------------------------------------------------------
// Inserts the value of ui into the buffer`s contents starting at
// index offset and returns a reference to the buffer.
//------------------------------------------------------------------------
INLINE StringBuffer &StringBuffer::insert(StringInt offset, unsigned ui)
   {
   return insert(offset, String::valueOf(ui));
   }

//------------------------------------------------------------------------
// Inserts the value of l into the buffer`s contents starting at
// index offset and returns a reference to the buffer.
//------------------------------------------------------------------------
INLINE StringBuffer &StringBuffer::insert(StringInt offset, long l)
   {
   return insert(offset, String::valueOf(l));
   }

//------------------------------------------------------------------------
// Inserts the value of ul into the buffer`s contents starting at
// index offset and returns a reference to the buffer.
//------------------------------------------------------------------------
INLINE StringBuffer &StringBuffer::insert(StringInt offset,
                                          unsigned long ul)
   {
   return insert(offset, String::valueOf(ul));
   }

//************************************************************************
// WordIterator`s inline method definitions.
//************************************************************************

#if CHECK_ASSERTIONS

//------------------------------------------------------------------------
// Returns nonzero iff the WordIterator`s state is valid.
//------------------------------------------------------------------------
INLINE int WordIterator::isValid() const
   {
   return (indexOfNextWord >= -1) && (indexOfNextWord <= theString.length());
   }

#endif  // #if CHECK_ASSERTIONS

//------------------------------------------------------------------------
// Constructs a WordIterator for the String str.
//------------------------------------------------------------------------
INLINE WordIterator::WordIterator(const String &str): theString(str)
   {
   indexOfNextWord = theString.indexOfWord(0);
   ensure(isValid());
   }

//------------------------------------------------------------------------
// Copy constructor.
//------------------------------------------------------------------------
INLINE WordIterator::WordIterator(const WordIterator &wi):
   theString(wi.theString)
   {
   require(wi.isValid());
   
   indexOfNextWord = wi.indexOfNextWord;

   ensure(isValid());
   }

//------------------------------------------------------------------------
// Returns true iff there is another word in the String.
//------------------------------------------------------------------------
INLINE boolean WordIterator::hasMoreWords() const
   {
   return (indexOfNextWord >= 0) ? true : false;
   }
