//************************************************************************
// Defines macros and classes for debugging, assertion checking, and
// platform- and compiler-dependent types, functions, etc.
//------------------------------------------------------------------------
// Copyright (C) 1997 by James MacKay
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
// The author can be contacted via e-mail at: jmackay@sentex.net
//
//************************************************************************
#ifndef RELEASE_H_
#define RELEASE_H_

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

//------------------------------------------------------------------------
// Define the string constants used in this file.
//------------------------------------------------------------------------
#define ASSERT_HEADING_MSG \
   "*** Assertion Violation ***"
#define DEBUG_HEADING_MSG \
   "*** Debugging Information ***"

#define PRECOND_FAILED_MSG \
   "Precondition failed in file %s, line %d: require(%s)\n"
#define CHECK_FAILED_MSG \
   "Assertion failed in file %s, line %d: check(%s)\n"
#define POSTCOND_FAILED_MSG \
   "Postcondition failed in file %s, line %d: ensure(%s)\n"

//------------------------------------------------------------------------
// Set the maximum length of an assertion violation message (= the max.
// length of any of the previous 3 strings ater parameter substitution).
//------------------------------------------------------------------------
#define MAX_ASSERT_MSG_LENGTH  256

//************************************************************************
// Define constants describing various target environment information.
//************************************************************************

//------------------------------------------------------------------------
// define target platform constants:
//------------------------------------------------------------------------
#define DOS_PLATFORM     1
#define WIN_PLATFORM     2
#define LINUX_PLATFORM   3

//------------------------------------------------------------------------
// define I/O type constants:
//
// STD_IO - use standard input/output
// WIN_IO - use Windows input/output
//------------------------------------------------------------------------
#define STD_IO           1 
#define WIN_IO           2


//             ----=====   Changeable Section   =====----


//------------------------------------------------------------------------
// Author- and copyright-related strings.
//------------------------------------------------------------------------
#define AUTHOR_NAME        "James MacKay"
#define COPYRIGHT_NOTICE   "Copyright (C) 1997 James MacKay"

//------------------------------------------------------------------------
// INLINE functions are inline functions iff ENABLE_INLINES is defined.
//------------------------------------------------------------------------
// The same setting should be used for ALL modules in a program. 
//------------------------------------------------------------------------
#define ENABLE_INLINES       1

//------------------------------------------------------------------------
// Define CHECK_ASSERTIONS as 1 to have assertions checked and debugging
// information output, and as 0 to turn off all assertion checking and
// debugging output.
//------------------------------------------------------------------------
// The same setting should be used for ALL modules in a program. 
//------------------------------------------------------------------------
#define CHECK_ASSERTIONS     0

//------------------------------------------------------------------------
// When an assertion is violated, that assertion is usually included in
// the message indicating the violation. However, this adds a stringized
// version of each assertion to the code, enlarging the size of the program
// (though only when assertion checking is on). If COMPACT_ASSERT_MSGS is
// defined as 1, a generic string (e.g. 'require(?)' for precondition
// violations) is used to describe an assertion violation of a given type;
// if it is defined as 0, the stringized version of each assertion is
// used (and included in the code).
//------------------------------------------------------------------------
// The same setting should be used for ALL modules in a program. 
//------------------------------------------------------------------------
#define COMPACT_ASSERT_MSGS  0

//------------------------------------------------------------------------
// Define the constant indicating the target platform. Or comment all of
// them out and define it on compiler`s command line, as in
//
// cc -DTARGET_PLATFORM=LINUX_PLATFORM ...
//------------------------------------------------------------------------
// The same setting should be used for ALL modules in a program. 
//------------------------------------------------------------------------
#define TARGET_PLATFORM    LINUX_PLATFORM
//#define TARGET_PLATFORM    WIN_PLATFORM
//#define TARGET_PLATFORM    DOS_PLATFORM

//------------------------------------------------------------------------
// Define the following constants,  whose values depend on the target 
// platform and/or the compiler being used:
//
// IO_TYPE - the type of I/O to use on the target platform, e.g. when 
//           outputting debugging messages
//
// IS_BOOL_TYPE - define this as 1 if the C++ 'bool' type is supported by
//                the compiler; define it as 0 if it isn`t
//
// ASSERTS_THROW_EXCEPTS - define this as 1 if an assertion violation
//                         should throw an exception; define it as 0 if
//                         the program should exit instead.
//------------------------------------------------------------------------
#if TARGET_PLATFORM == DOS_PLATFORM
   #define IO_TYPE               STD_IO
   #define IS_BOOL_TYPE          0
   #define ASSERTS_THROW_EXCEPTS 0
#elif TARGET_PLATFORM == WIN_PLATFORM
   #define IO_TYPE               WIN_IO
   #define IS_BOOL_TYPE          0
   #define ASSERTS_THROW_EXCEPTS 0
#elif TARGET_PLATFORM == LINUX_PLATFORM
   #define IO_TYPE               STD_IO
   #define IS_BOOL_TYPE          1
   #define ASSERTS_THROW_EXCEPTS 0
#else
   #error TARGET_PLATFORM is undefined or set to an invalid value
#endif


//         ----=====   End of Changeable Section   =====----


//------------------------------------------------------------------------
// Define the boolean type.
//------------------------------------------------------------------------
#if IS_BOOL_TYPE
   typedef bool boolean;
#else
   enum boolean
      {
      false = 0,
      true  = 1
      };
#endif

//------------------------------------------------------------------------
// Define INLINE depending on whether INLINE functions are to be inlined.
//------------------------------------------------------------------------
#ifdef ENABLE_INLINES
   #define INLINE inline
#else
   #define INLINE
#endif

//------------------------------------------------------------------------
// Declare the exception classes that are thrown when a precondition,
// assertion or postcondition is violated.
//------------------------------------------------------------------------
// Note: these classes should be defined even if assertions aren`t
//       checked and/or assertion violations don`t throw exceptions.  
//------------------------------------------------------------------------
class AnyAssertionException
   {
   // empty
   };

class PreconditionViolatedException: public AnyAssertionException
   {
   // empty
   };

class CheckViolatedException: public AnyAssertionException
   {
   // empty
   };

class PostconditionViolatedException: public AnyAssertionException
   {
   // empty
   };

#if CHECK_ASSERTIONS
   //------------------------------------------------------------------------
   // include headers, etc. needed for debugging/assertion output:
   //------------------------------------------------------------------------
   #if IO_TYPE == WIN_IO
      #include <windows.h>
   #endif

   //------------------------------------------------------------------------
   // Outputs info as debugging information.
   //------------------------------------------------------------------------
   inline void debugOutput(const char *info)
      {
      #if IO_TYPE == STD_IO
         fflush(stdout);     // in case stdout & stderr write to the same device
         fputs(info, stderr);
         fflush(stderr);
      #elif IO_TYPE == WIN_IO
         ::MessageBox(0, info, DEBUG_HEADING_MSG,
                      MB_OK | MB_ICONEXCLAMATION);
      #else
         #error No debugOutput() function defined for the current IO_TYPE
      #endif
      }   

   //------------------------------------------------------------------------
   // define macros for placing assertion checks in the program.
   //------------------------------------------------------------------------
   // These macros expand to statements to help prevent their being necessary
   // parts of the program, and they have else clauses so as not to introduce
   // and 'dangling-else' problems.
   //------------------------------------------------------------------------
   #if COMPACT_ASSERT_MSGS
      #define require(cond) \
         if ( !ProgramContract::checkPreconds || (cond) ) \
            ;                                             \
         else                                             \
            ProgramContract::onPrecondFail("...", __FILE__, __LINE__)

      #define check(cond) \
         if ( !ProgramContract::checkChecks || (cond) ) \
            ;                                           \
         else                                           \
            ProgramContract::onCheckFail("...", __FILE__, __LINE__)

      #define ensure(cond) \
         if ( !ProgramContract::checkPostconds || (cond) ) \
            ;                                              \
         else                                              \
            ProgramContract::onPostcondFail("...", __FILE__, __LINE__)
   #else
      #define require(cond) \
         if ( !ProgramContract::checkPreconds || (cond) ) \
            ;                                             \
         else                                             \
            ProgramContract::onPrecondFail(#cond, __FILE__, __LINE__)

      #define check(cond) \
         if ( !ProgramContract::checkChecks || (cond) ) \
            ;                                           \
         else                                           \
            ProgramContract::onCheckFail(#cond, __FILE__, __LINE__)

      #define ensure(cond) \
         if ( !ProgramContract::checkPostconds || (cond) ) \
            ;                                              \
         else                                              \
            ProgramContract::onPostcondFail(#cond, __FILE__, __LINE__)
   #endif

   //------------------------------------------------------------------------
   // Class that control which assertions are checked and what happens when
   // an assertion is violated.
   //------------------------------------------------------------------------
   class ProgramContract
      {
      public:
         enum
            {
            checkPreconds  = 1,             // preconditions are checked
            checkChecks    = 1,             // checks are checked
            checkPostconds = 1              // postconditions are checked
            };

			static void onPrecondFail(const char *cond, const char *fileName, 
                                   int lineNum);
         static void onCheckFail(const char *cond, const char *fileName,
                                   int lineNum);
         static void onPostcondFail(const char *cond, const char *fileName,
                                   int lineNum);
      private:
         static void outputAssertInfo(const char *format, const char *cond,
                                      const char *fileName, int lineNum);
      }; // class ProgramContract

   //------------------------------------------------------------------------
   // Outputs the specified information about an assertion violation. It
   // assumes that the format string, after cond, fileName and lineNum have
   // been inserted into it, will be at most MAX_ASSERT_MSG_LENGTH
   // characters long.
   //------------------------------------------------------------------------
   inline void ProgramContract::outputAssertInfo(const char *format, 
                                                 const char *cond,
                                                 const char *fileName, int lineNum)
      {
      char buf[MAX_ASSERT_MSG_LENGTH];
      sprintf(buf, format, fileName, lineNum, cond);   

      #if IO_TYPE == STD_IO
         fflush(stdout);     // in case stdout & stderr write to the same device
         fputs(buf, stderr);
         fflush(stderr);
      #elif IO_TYPE == WIN_IO
         ::MessageBox(0, buf, ASSERT_HEADING_MSG,
                      MB_OK | MB_ICONEXCLAMATION);
      #else
          #error No ProgramContract::outputAssertInfo() function defined \
                 for the current IO_TYPE
      #endif
      }

   //------------------------------------------------------------------------
   // Called when a precondition is violated.
   //------------------------------------------------------------------------
   inline void ProgramContract::onPrecondFail(const char *cond,
                                              const char *fileName, int lineNum)
      {
      outputAssertInfo(PRECOND_FAILED_MSG, cond, fileName, lineNum);
      #if ASSERTS_THROW_EXCEPTS
         throw PreconditionViolatedException();
      #else
         exit(1);
      #endif
      }

   //------------------------------------------------------------------------
   // Called when an assertion is violated.
   //------------------------------------------------------------------------      
   inline void ProgramContract::onCheckFail(const char *cond,
                                            const char *fileName, int lineNum)
      {
      outputAssertInfo(CHECK_FAILED_MSG, cond, fileName, lineNum);
      #if ASSERTS_THROW_EXCEPTS
         throw CheckViolatedException();
      #else
         exit(1);
      #endif
      }

   //------------------------------------------------------------------------
   // Called when a postcondition is violated.
   //------------------------------------------------------------------------
   inline void ProgramContract::onPostcondFail(const char *cond,
                                               const char *fileName, int lineNum)
      {
      outputAssertInfo(POSTCOND_FAILED_MSG, cond, fileName, lineNum);
      #if ASSERTS_THROW_EXCEPTS
         throw PostconditionViolatedException();
      #else
         exit(1);
      #endif
      }

#else  // assertions aren`t being checked
   #define require(cond)
   #define check(cond)
   #define ensure(cond)
   #define debugOutput(info)
#endif  // #if CHECK_ASSERTIONS

//------------------------------------------------------------------------
// Undefine macros used only in this file
//------------------------------------------------------------------------
#undef ASSERT_HEADING_MSG
#undef DEBUG_HEADING_MSG

#undef PRECOND_FAILED_MSG
#undef CHECK_FAILED_MSG
#undef POSTCOND_FAILED_MSG

#undef MAX_ASSERT_MSG_LENGTH
#undef COMPACT_ASSERT_MSGS

#endif   // #ifndef RELEASE_H_
